#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <linux/gamma/libgamma.h>

#define NUMARG 6
#define RECV_COUNTER 13


void usage(char *prog) {
  fprintf(stderr,
"Usage: %s <# of comm (< # of processes)> <max msg len (bytes)> <max iters> <bool: barrier sync at each iter?> <bool: correctness test at each iter?>\n",prog);
  return;
}


int main(int argc, char *argv[]) {

int MonNumero,n,i,j,ii,iinorm,ntimes,nmax,proc,K,barrflag,correctflag;
unsigned char *message,*message_rec;
time_586 time1, time2;
double MaxT,time,time_final;
unsigned char par_pid;


if(argc < NUMARG) {
   usage(argv[0]);
   exit(-1);
}

gamma_init_argv(&argc,&argv);
MonNumero = gamma_my_node();
proc = gamma_how_many_nodes();
par_pid = gamma_my_par_pid();

K=atoi(argv[1]);
if(K>(proc-1)) {
   if(!MonNumero) fprintf(stderr,"Too many comm: %d is the max\n",proc-1);
   gamma_abort(); exit(-1);
}
nmax=atoi(argv[2]);
if(nmax < 1) {
   if(!MonNumero) fprintf(stderr,"Wrong msg size\n");
   gamma_abort(); exit(-1);
}
ntimes=atoi(argv[3]);
if(ntimes<1) {
   if(!MonNumero) fprintf(stderr,"Wrong no. of iterations\n");
   gamma_abort(); exit(-1);
}
barrflag=atoi(argv[4]);
correctflag=atoi(argv[5]);

if(!MonNumero) {
fprintf(stdout,"#============================================\n");
fprintf(stdout,"#   KTK for GAMMA with %d PROCESSES\n",proc);
fprintf(stdout,"#============================================\n");
}

if(!(message=(unsigned char *)malloc(nmax))) {
   fprintf(stderr,"Could not allocate memory %d\n", nmax);
   gamma_abort(); exit(-1);
}

if(!(message_rec=(unsigned char *)malloc(nmax*proc))) {
   fprintf(stderr,"Could not allocate memory %d\n", nmax*proc);
   gamma_abort(); exit(-1);
}

/*  Set input ports  */
for(ii = proc-1+MonNumero; ii > (proc-1-K)+MonNumero; ii--) {
   iinorm=ii%proc;
   gamma_set_passive_port(iinorm,DONT_MIND,DONT_MIND,DONT_MIND,
     RECV_COUNTER,GO_BACK,NULL,0);
}

/*  Set output ports  */
for(ii = MonNumero+1; ii < K+MonNumero+1; ii++) {
   iinorm=ii%proc;
   gamma_set_passive_port(proc+iinorm,iinorm,par_pid,MonNumero,
     DONT_MIND,DONT_MIND,NULL,0);
}

gamma_sync();

if (!MonNumero) {
  fprintf(stdout,"%10s  %13s  %18s\n",
    "msg size","end2end delay","aggreg. throughput");
  fprintf(stdout,"%10s  %13s  %18s\n",
    "","","(bidirectional)");
  fprintf(stdout,"%10s  %13s  %18s\n",
    "","(usec)","(MByte/s)");
}

gamma_time(time1);
n = 64;
while(n <= nmax)
 {
  time=0;
  for(i = 0; i < ntimes; i++) {

    if(barrflag) gamma_sync();

    if (correctflag) {
      memset(message_rec,0,n);
      for (j=0;j<n;j++)
        message[j] = (unsigned char)(j*(1+MonNumero));
    }

    for(ii = proc-1+MonNumero; ii > (proc-1-K)+MonNumero; ii--) {
      iinorm=ii%proc;
      gamma_post_recv(iinorm,&message_rec[n*iinorm],n);
    }

    gamma_sync();

    gamma_time(time1);

    for(ii = MonNumero+1; ii < K+MonNumero+1; ii++) {
      iinorm=ii%proc;
      gamma_send_flowctl(proc+iinorm,message,n);
    }
     
    gamma_wait(RECV_COUNTER,K);

    gamma_time(time2);
    MaxT = gamma_time_diff(time2,time1);

    if(barrflag) gamma_sync();

//    MPI_Reduce(&MaxT,&T,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
//    if(!MonNumero) time = time + T;
    if(!MonNumero) time += MaxT;

    if (!correctflag) continue;

    for(ii = proc-1+MonNumero; ii > (proc-1-K)+MonNumero; ii--) {
      iinorm=ii%proc;
      for (j=0;j<n;j++)
        if (message_rec[n*iinorm+j] != (unsigned char)(j*(1+iinorm))) {
	  fprintf(stderr,"[%d] err\n",MonNumero);
	  break;
	}
    }

  }

  if(!MonNumero) {
     time_final = time/ntimes;
 
     fprintf(stdout,"%10d  %13.6f  %18f\n",
       n,(time_final/2.),(double)(n*K*2)/time_final);
  }

  n = n*2;
}


gamma_sync();
gamma_exit();
return 0;

}
