/*  GAMMA version  */
/*  Giuseppe Ciaccio 20 Aug 1998 */

#define MEASURE_LOOP
#define NMAX 1500
#define REAL double
#define gamma_SEND(a,b,c) gamma_send_fast(a,b,c)

#ifdef MEASURE_LOOP
	#define START 200
	#define STOP 1400
	#define STEP 200
#else
	#define N 100
	#define START N
	#define STOP N
	#define STEP 100
#endif

#include <stdio.h>
#include <stdlib.h>
#include <linux/gamma/libgamma.h>
#include "measure.h"

#define MAX_NODES 20	/*  No more than M1_PORT!  */
#define MASTER_NODE 0
#undef PRINT_RESULT
#define PERFORMANCE
#define PERF_SHORT
#undef TIME
#undef FREQUENCIES

#define M1_PORT 64
#define M2_PORT 65
#define RES_SEM 66
#define EVERYBODY 67
#define TOKEN_PORT 68


extern int dgemm_();


void rec_handler(void) {
	gamma_signal(gamma_active_port);
}
void err_handler(void) {
	gamma_sigerr(gamma_active_port);
}
void rec_res_handler(void) {
	gamma_signal(RES_SEM);
}
void err_res_handler(void) {
	gamma_sigerr(RES_SEM);
}


int main(int argc, char **argv) {

REAL		*m1,*m2,*m3;
double		zero,one;
int		z,p,q,whoami,n_nodes,n,m,master_cols;
char		*trans = "No transpose";
double		record_mm[NPROVE];
time_586	t_s,t_e;


if (argc < 2) {
	printf("Usage: mat_mul <No_Of_Processors>\n");
	exit(0);
}

zero = 0.0;
one = 1.0;

n_nodes = atoi(argv[1]) > MAX_NODES ? MAX_NODES : atoi(argv[1]);
gamma_init(n_nodes,argc,argv);
whoami = gamma_my_node();

/*  Allocate operands.  */
if (whoami == MASTER_NODE) {
	m1 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));
	m2 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));
	m3 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));
}
else {
	m1 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));
	m2 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL)/n_nodes);
	m3 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL)/n_nodes);
}

/*  Lock operands into RAM.  */
if (whoami == MASTER_NODE)
	gamma_mlock(m3,NMAX*NMAX*sizeof(REAL));
else {
	gamma_mlock(m1,NMAX*NMAX*sizeof(REAL));
	gamma_mlock(m2,NMAX*NMAX*sizeof(REAL)/n_nodes);
}


for (n=START; n<=STOP; n+=STEP) {

m = n/n_nodes;
master_cols = m+(n%n_nodes);

if (whoami == MASTER_NODE) {

	/*  Init operands (transposed, FORTRAN layout).  */
	for (p=0;p<n;p++)
		for (q=0;q<n;q++) {
			m1[n*q+p] = p;
			m2[n*q+p] = 1;
			m3[n*q+p] = 0;
	}

	/*
	 *  Port #j of MASTER_NODE will send rows of m2 to port M2_PORT of
	 *  slave #j,
	 *  and will receive rows of m3 from port M2_PORT of slave #j.
	 */
	for (p=1;p<n_nodes;p++) {
		gamma_setport(
			p,
			p,gamma_my_par_pid(),M2_PORT,
			rec_res_handler,err_res_handler,
			&(m3[(p-1)*m*n]),m*n*sizeof(REAL)
		);
	}

	/*  Output port M1_PORT will bcast the whole m1 to each slave.  */
	gamma_setport(
		M1_PORT,
		BROADCAST,gamma_my_par_pid(),M1_PORT,
		NULL,NULL,NULL,0
	);

	/*  Token output port  */
	gamma_setport(
		TOKEN_PORT,
		1,gamma_my_par_pid(),TOKEN_PORT,
		NULL,NULL,NULL,0
	);

	gamma_sync();

	/*  START THE GAME!!  */

	for (z=0;z<NPROVE;z++) {

#ifdef PRINT_RESULT
		/*
		 *  Init operands again to show the correct result
		 *  (transposed, FORTRAN layout).
		 */
		if (z == NPROVE-1)
		for (p=0;p<n;p++)
			for (q=0;q<n;q++) {
				m1[n*q+p] = p;
				m2[n*q+p] = 1;
				m3[n*q+p] = 0;
		}
#endif

gamma_time(t_s);
		/*  Bcast the whole m1 to each slave.  */
		gamma_SEND(M1_PORT,m1,n*n*sizeof(REAL));

		/*  Distribute strips of m2 to slaves.  */
		/*  Master keeps the last (possibly unbalanced) strip.  */
		for (p=1;p<n_nodes;p++)
			gamma_SEND(p,&(m2[(p-1)*m*n]),m*n*sizeof(REAL));

		gamma_SEND(TOKEN_PORT,NULL,0);

		/*  Perform the local part of the multiplication.  */
		dgemm_(trans,trans,&n,&master_cols,&n,&one,m1,&n,
		&(m2[(n_nodes-1)*m*n]),
		&n,&zero,
		&(m3[(n_nodes-1)*m*n]),
		&n);

		/*
		 *  Wait for the other n_nodes-1 chunks of the result
		 *  from slaves.
		 */
		if (gamma_wait(RES_SEM, n_nodes-1)) {
			printf("MASTER:  error receiving m3\n");
			fflush(stdout);
			gamma_exit();
			exit(-1);
		}
gamma_time(t_e);

		/*  Record the experiment  */
record_mm[z] = gamma_time_diff(t_e,t_s);

	}

#ifdef PRINT_RESULT
	printf("\n");
	printf("M1:\n");
	for (p=0;p<n;p++) {
		for (q=0;q<n;q++)
			printf("%g ",m1[n*q+p]);
		printf("\n");
	}
	printf("\n");
	printf("M2:\n");
	for (p=0;p<n;p++) {
		for (q=0;q<n;q++)
			printf("%g ",m2[n*q+p]);
		printf("\n");
	}
	printf("\n");
	printf("M1 * M2:\n");
	for (p=0;p<n;p++) {
		for (q=0;q<n;q++)
			printf("%g ",m3[n*q+p]);
		printf("\n");
	}
	printf("\n");
#endif

#ifdef PERFORMANCE
#ifdef PERF_SHORT
	{
		double mmm,Mmm,medmm,varmm;
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef TIME
		printf("%d %g\n",n,medmm);
#else
		printf("%d %g\n",n,(double)n*(double)n*(double)(2*n-1)/medmm);
#endif
		fflush(stdout);
	}
#else
	{
		double mmm,Mmm,medmm,varmm;
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
		printf("\n\n");
		printf("PERFORMANCE (microsec) (best < average < worst):\n\n");
	        printf("execution time %g < %g < %g\n", mmm,medmm,Mmm);
		printf("\n");
	        printf("variance\t%g\n", varmm);
		printf("\n\n");
#ifdef FREQUENCIES
	        printf("FREQUENCIES (%):\n");
		frequencies(record_mm,mmm,Mmm);
#endif
		fflush(stdout);

	}
#endif
#endif
}


else {

	/*
	 *  Port M2_PORT is used for getting strips of m2 from MASTER_NODE
	 *  as well as sending strips of the result back.
	 */
	gamma_setport(
		M2_PORT,
		MASTER_NODE,gamma_my_par_pid(),whoami,
		rec_handler, err_handler,
		m2,m*n*sizeof(REAL)
	);

	gamma_setport(
		M1_PORT,
		DONT_MIND,DONT_MIND,DONT_MIND,
		rec_handler, err_handler,
		m1,n*n*sizeof(REAL)
	);

	gamma_setport(
		TOKEN_PORT,
		whoami+1,gamma_my_par_pid(),TOKEN_PORT,
		rec_handler,err_handler,NULL,0
	);

	gamma_sync();

	for (z=0;z<NPROVE;z++) {

		/*  Wait for the work from MASTER  */
		if (gamma_wait(M1_PORT,1)) {
			printf("SLAVE%d:  error receiving m1\n",whoami);
			fflush(stdout);
			gamma_exit();
			exit(-1);
		}
		if (gamma_wait(M2_PORT,1)) {
			printf("SLAVE%d:  error receiving m2\n",whoami);
			fflush(stdout);
			gamma_exit();
			exit(-1);
		}

		/*  Perform the local part of the multiplication.  */
		dgemm_(trans,trans,&n,&m,&n,&one,m1,&n,m2,&n,&zero,m3,&n);

		/*  Send the partial result back to MASTER.  */
		if (gamma_wait(TOKEN_PORT,1)) {
			printf("SLAVE%d:  error receiving token\n",
				whoami);
			fflush(stdout);
			gamma_exit();
			exit(-1);
		}
		gamma_SEND(M2_PORT,m3,m*n*sizeof(REAL));
		gamma_SEND(TOKEN_PORT,NULL,0);

	}
}

}
gamma_exit();
return 0;
}
