/*  Giuseppe Ciaccio 20 Aug 1998  */

#define MEASURE_LOOP
#define NMAX 1500
#define REAL double

#ifdef MEASURE_LOOP
	#define START 16
	#define STOP 1000
	#define STEP 4
#else
	#define N 1500
	#define START N
	#define STOP N
	#define STEP 100
#endif

#include <stdio.h>
#include <stdlib.h>
#include <linux/gamma/libgamma.h>
#include "measure.h"

#undef PRINT_RESULT
#define PERFORMANCE
#undef TIME
#define PERF_SHORT
#undef FREQUENCIES


extern int dgemm_();


int main(int argc, char **argv) {

REAL		*m1,*m2,*m3;
double		zero,one;
int		z,p,q,n;
char		*trans = "No transpose";
double		record_mm[NPROVE];
time_586	t_s,t_e;


zero = 0.0;
one = 1.0;

/*  Allocate operands.  */
m1 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));
m2 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));
m3 = (REAL *)malloc(NMAX*NMAX*sizeof(REAL));

for (n=START; n<=STOP; n+=STEP) {

	/*  Init operands (transposed, FORTRAN layout).  */
	for (p=0;p<n;p++)
		for (q=0;q<n;q++) {
			m1[n*q+p] = p;
			m2[n*q+p] = 1;
			m3[n*q+p] = 0;
	}

	
	/*  START THE GAME!!  */

	for (z=0;z<NPROVE;z++) {

#ifdef PRINT_RESULT
		/*
		 *  Init operands again to show the correct result
		 *  (transposed, FORTRAN layout).
		 */
		if (z==NPROVE-1)
		for (p=0;p<n;p++)
			for (q=0;q<n;q++) {
				m1[n*q+p] = p;
				m2[n*q+p] = 1;
				m3[n*q+p] = 0;
		}
#endif

gamma_time(t_s);
		dgemm_(trans,trans,&n,&n,&n,&one,m1,&n,m2,&n,&zero,m3,&n);
gamma_time(t_e);
		/*  Record the experiment  */
record_mm[z] = gamma_time_diff(t_e,t_s);

	}

#ifdef PRINT_RESULT
	printf("\n");
	printf("M1:\n");
	for (p=0;p<n;p++) {
		for (q=0;q<n;q++)
			printf("%g ",m1[n*q+p]);
		printf("\n");
	}
	printf("\n");
	printf("M2:\n");
	for (p=0;p<n;p++) {
		for (q=0;q<n;q++)
			printf("%g ",m2[n*q+p]);
		printf("\n");
	}
	printf("\n");
	printf("M1 * M2:\n");
	for (p=0;p<n;p++) {
		for (q=0;q<n;q++)
			printf("%g ",m3[n*q+p]);
		printf("\n");
	}
	printf("\n");
#endif

#ifdef PERFORMANCE
#ifdef PERF_SHORT
	{
		double mmm,Mmm,medmm,varmm;
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef TIME
		printf("%d %g\n",n,medmm);
#else
		printf("%d %g\n",n,(double)(2*n-1)*(double)n*(double)n/medmm);
#endif
		fflush(stdout);
	}
#else
	{
		double mmm,Mmm,medmm,varmm;
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
		printf("\n\n");
		printf("PERFORMANCE (microsec) (best < average < worst):\n\n");
	        printf("execution time %g < %g < %g\n", mmm,medmm,Mmm);
		printf("\n");
	        printf("variance\t%g\n", varmm);
		printf("\n\n");
#ifdef FREQUENCIES
	        printf("FREQUENCIES (%):\n");
		frequencies(record_mm,mmm,Mmm);
#endif
		fflush(stdout);

	}
#endif
#endif

}

return 0;

}
