#undef USE_2P
#undef USE_ISEND
#define USE_FLOWCTL
#undef USE_BCAST

#undef SEND_TIME
#undef DELAY

#undef STREAM
#ifdef STREAM
#define SEND_TIME
#define max(a,b) ((a) > (b) ? (a) : (b))
#define NSTREAM max(10,(len ? (10000000/len) : 10000))
#else
#define NSTREAM 1
#endif

#define MAX_LEN 1000000

#ifdef USE_FLOWCTL
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p_flowctl(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend_flowctl(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p_flowctl(a,b,c,d,e)
#else
#define gamma_SEND(a,b,c)  gamma_send(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p(a,b,c,d,e)
#endif

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure_curva.h"

#define MASTER_NODE 0
#define SLAVE_NODE 1
#define PORT 1


unsigned char *data,*data1;
int len;

#ifdef USE_2P
void handler(void) {
	gamma_switch_to_buffer(PORT,data+12,len);
	return;
}
#endif


int main(int argc, char **argv) {

	double record_mm[NPROVE];

	int i,j;
	time_586
		t_send_s,
		t_rec;
#ifdef SEND_TIME
	time_586
		t_send_e;
#endif
	double time_overhead;

	typedef struct {
		int x;
		double y;
	} tr;
	tr *total_record;
	int i_len = 0;

	gamma_init(2,argc,argv);
	/*  Allocate data space (+4096 to allow page alignment)  */
#ifdef USE_2P
	data1 = (unsigned char *)malloc(4096+MAX_LEN*sizeof(unsigned char)+12);
#else
	data1 = (unsigned char *)malloc(4096+MAX_LEN*sizeof(unsigned char));
#endif
	/*  Align to page boundary  */
	data = (unsigned char *)(((unsigned long)data1+4096) & (~4095));

	total_record = (tr *)malloc(10000*sizeof(tr));

	/*  Pin communication data structures in RAM.  */
#ifdef USE_2P
	gamma_mlock(data,MAX_LEN+12);
#else
	gamma_mlock(data,MAX_LEN);
#endif

	len = 0;
	while (len <= MAX_LEN) {

	/*  Init the message to be exchanged.  */
#ifdef USE_2P
	for (j=0;j<len+12;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#else
	for (j=0;j<len;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif

	/*  Compute the gamma_time() overhead...  */
	for (i=0;i<NPROVE;i++) {
		gamma_time(t_send_s);
 		gamma_time(t_rec);
		record_mm[i] = gamma_time_diff(t_rec, t_send_s);
	}
	{
	double Mmm,medmm,varmm;
	mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
	}

	if (gamma_my_node() == MASTER_NODE) {

		/*  this is PING  */

#ifdef STREAM
		fprintf(stderr,"%d \n",len);fflush(stderr);
#endif

#ifdef USE_2P
		gamma_set_ap_port_2p(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
#ifndef USE_BCAST
		gamma_set_passive_port(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#else
		gamma_set_passive_port(
			PORT,
			BROADCAST,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
#endif

		total_record[i_len].x = len;
		total_record[i_len].y = 0;

		gamma_sync();

		for (i=0;i<NPROVE;i++) {
			int status;
#ifdef STREAM
			int stream;
#endif

			gamma_time(t_send_s);

#ifdef STREAM
			for (stream = NSTREAM; stream; stream--) {
#endif

#ifdef USE_ISEND
#ifdef USE_2P
			j = gammaISEND2P(PORT,data,12,data+12,len);
#else
			j = gamma_ISEND(PORT,data,len);
#endif
#else
#ifdef USE_2P
			gammaSEND2P(PORT,data,12,data+12,len);
#else
			gamma_SEND(PORT,data,len);
#endif
#endif

#ifdef STREAM
			}
#endif
#ifdef SEND_TIME
			gamma_time(t_send_e);
#endif
			status = gamma_wait(PORT,1);

 			gamma_time(t_rec);

			if (status) {
				printf("PING:  ERROR\n");
				fflush(stdout);
				free(data1);
				gamma_exit();
				exit(-1);
			}
#ifdef USE_ISEND
			gamma_wsend(j);
#endif

#ifdef SEND_TIME
			total_record[i_len].y +=
		(gamma_time_diff(t_send_e, t_send_s)-time_overhead)/NSTREAM;
#else
#ifdef STREAM
			total_record[i_len].y +=
		(gamma_time_diff(t_rec, t_send_s)-time_overhead)/NSTREAM;
#else
			total_record[i_len].y +=
		(gamma_time_diff(t_rec, t_send_s)-time_overhead)/2;
#endif
#endif

		}

		total_record[i_len].y /= NPROVE;
#ifndef DELAY
		if (len)
			total_record[i_len].y = len/total_record[i_len].y;
#endif
	}


	else {
		/*  this is PONG  */
#ifdef USE_2P
		gamma_set_ap_port_2p(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
#ifndef USE_BCAST
		gamma_set_passive_port(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#else
		gamma_set_passive_port(
			PORT,
			BROADCAST,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
#endif
		gamma_sync();

		for (i=0;i<NPROVE;i++) {

			if (!gamma_wait(PORT,NSTREAM)) {
#ifdef STREAM
				gamma_send_fast(PORT,NULL,0);
#else
#ifdef USE_ISEND
#ifdef USE_2P
				j = gammaISEND2P(PORT,data,12,data+12,len);
				gamma_wsend(j);
#else
				j = gamma_ISEND(PORT,data,len);
				gamma_wsend(j);
#endif
#else
#ifdef USE_2P
				gammaSEND2P(PORT,data,12,data+12,len);
#else
				gamma_SEND(PORT,data,len);
#endif
#endif
#endif
				continue;
			}
			printf("PONG:  ERROR\n");
			fflush(stdout);
			free(data1);
			gamma_exit();
			exit(-1);
		}
	}

	i_len++;
	if (len<9000) len += 4;
	else if (len <100000) len += 1000;
	else if (len <1000000) len += 10000;
	else len += 100000;

	}

	if (gamma_my_node() == MASTER_NODE) {
		for (i=0; i<i_len; i++)
			printf("%d %g\n",total_record[i].x,total_record[i].y);
		fflush(stdout);
	}
	free(data1);
	gamma_exit();
	return 0;
}
