#define USE_FLOWCTL

#define NSTREAM 1

#ifdef USE_FLOWCTL
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#else
#define gamma_SEND(a,b,c)  gamma_send(a,b,c)
#endif

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure.h"

#define MASTER_NODE 0
#define SLAVE_NODE 1
#define PORT 1


unsigned char *sdata,*sdata1;
unsigned char *rdata,*rdata1;
int len;

int main(int argc, char **argv) {

	double record_mc[NPROVE];
	double record_mm[NPROVE];
	int i,me;
	int j;
	time_586 t_send_s, t_rec;
	time_586 t_send_e;
	double time_overhead;

	gamma_init(2,argc,argv);
   	len = atoi(argv[1]);
	me = gamma_my_node();

	/*  Allocate data space (+4096 to allow page alignment)  */
	sdata1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
	rdata1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
	/*  Align to page boundary  */
	sdata = (unsigned char *)(((unsigned long)sdata1+4096) & (~4095));
	rdata = (unsigned char *)(((unsigned long)rdata1+4096) & (~4095));

	/*  Init the message.  This also brings buffer in RAM  */
	for (i=0;i<len;i++) {
		sdata[i] = (unsigned char)(i%128+me*128);
		rdata[i] = (unsigned char)0;
	}

	/*  Computing the gamma_time() overhead...  */
	for (i=0;i<NPROVE;i++) {
		gamma_time(t_send_s);
 		gamma_time(t_rec);
		record_mm[i] = gamma_time_diff(t_rec, t_send_s);
	}
	{
	double Mmm,medmm,varmm;
	mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
	}

	gamma_set_passive_port(
		PORT,
		(me+1)%2,gamma_my_par_pid(),PORT,
		PORT,GO_BACK,rdata,len
		);

	gamma_sync();

	for (i=0;i<NPROVE;i++) {

		for (j=0;j<len;j++) {
			sdata[j] = (unsigned char)(j%128+me*128);
			rdata[j] = (unsigned char)0;
		}

		gamma_sync();
		gamma_time(t_send_s);
		gamma_SEND(PORT,sdata,len);
		gamma_time(t_send_e);
		gamma_wait(PORT,1);
 		gamma_time(t_rec);

		/*  Check data correctness.  */
		for (j=0;j<len;j++)
         if (rdata[j] != (unsigned char)(j%128+((me+1)%2)*128)) {
			int k;
            printf("PING:  data error\n");
            for (k=j-5;k<j+10;k++)
				printf("%d:%d ",k%128+((me+1)%2)*128,rdata[k]);
            printf("\n");
			break;
            }

		record_mc[i] =
		gamma_time_diff(t_send_e, t_send_s)-time_overhead;
		record_mm[i] =
		gamma_time_diff(t_rec, t_send_s)-time_overhead;

	}


	gamma_timers(t_send_s,t_rec);

	free(sdata1);
	free(rdata1);
	gamma_exit();

        if (me == MASTER_NODE) {
                double mmc,Mmc,medmc,varmc;
                double mmm,Mmm,medmm,varmm;
                printf("\n\n");
                printf("PERFORMANCE (microsec, MB/sec) (lowest < average < highest):\n\n");
                mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
                printf("total send time        %g < %g < %g\n",
                        mmc,medmc,Mmc);
                if (len)
                printf("send throughput        %g < %g < %g\n",
                        NSTREAM*len/Mmc,NSTREAM*len/medmc,
                        NSTREAM*len/mmc);
                mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
                printf("total delay            %g < %g < %g\n",
                        mmm,medmm,Mmm);
                if (len)
                printf("end-to-end throughput  %g < %g < %g\n",
                        NSTREAM*len/Mmm,NSTREAM*len/medmm,
                        NSTREAM*len/mmm);
                printf("\n\n");

                printf("VARIANCE:  \n");
                printf("send time        %g\n", varmc);
                printf("end-to-end delay %g\n", varmm);
                printf("\n\n");

                fflush(stdout);
        }
	return 0;
}
