#undef USE_2P
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p_flowctl(a,b,c,d,e)

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include "measure.h"

#undef USE_BCAST
#ifdef USE_2P
	#undef USE_BCAST
#endif
#define MASTER_NODE 0
#define SLAVE_NODE 1
#define PORT 1
#define CORRECTNESS
#undef SYSCALL_TIME
#define PERFORMANCE
#define PERF_SHORT
#undef DELAY
#undef FREQUENCIES


unsigned char *data;
int len;

#ifdef USE_2P
void handler(void) {
	gamma_switch_to_buffer(PORT,data+12,len);
	return;
}
#endif


int main(int argc, char **argv) {

#ifdef PERFORMANCE
#ifdef SYSCALL_TIME
	double record_mc[NPROVE];
#endif
	double record_mm[NPROVE];
#endif
	int i,j;
	time_586
		t_send_s,
		t_rec;
#ifdef SYSCALL_TIME
	time_586
		t_send_e;
#endif
#ifdef PERFORMANCE
	double time_overhead;
#endif

	gamma_init(2,argc,argv);
   	len = atoi(argv[1]);
#ifdef USE_2P
	data = (unsigned char *)malloc(32+len*sizeof(unsigned char)+12);
#else
	data = (unsigned char *)malloc(32+len*sizeof(unsigned char));
#endif

	/*  Init the message to be exchanged.  */
#ifdef USE_2P
	for (j=0;j<len+12;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#else
	for (j=0;j<len;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif


	if (gamma_my_node() == MASTER_NODE) {

		/*  this is PING  */

#ifdef USE_2P
		gamma_set_ap_port_2p(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
#ifndef USE_BCAST
		gamma_set_passive_port(
			PORT,
			SLAVE_NODE,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#else
		gamma_set_passive_port(
			PORT,
			BROADCAST,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
#endif

#ifdef PERFORMANCE
		/*  Computing the gamma_time() overhead...  */
		for (i=0;i<NPROVE;i++) {
			gamma_time(t_send_s);
 			gamma_time(t_rec);
			record_mm[i] = gamma_time_diff(t_rec, t_send_s);
		}
		{
		double Mmm,medmm,varmm;
		mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
		}
#endif

		gamma_sync();

		for (i=0;i<NPROVE;i++) {
			int status;
#ifdef USE_2P
	for (j=0;j<len+12;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#else
	for (j=0;j<len;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif
gamma_sync();
#ifdef PERFORMANCE
			gamma_time(t_send_s);
#endif
#ifdef USE_2P
			gammaSEND2P(PORT,data,12,data+12,len);
#else
			gamma_SEND(PORT,data,len);
#endif
#ifdef PERFORMANCE
#ifdef SYSCALL_TIME
			gamma_time(t_send_e);
#endif
#endif
			status = gamma_wait(PORT,1);
#ifdef PERFORMANCE
 			gamma_time(t_rec);
#endif
			if (status) {
				printf("PING:  ERROR\n");
				fflush(stdout);
				gamma_exit();
				exit(-1);
			}

			/*
			 *  Force printout of kernel-level time measurements,
			 *  if enabled.  Otherwise nothing happens.
			 */
			gamma_timers(t_send_s,t_rec);

#ifdef CORRECTNESS
			/*  Check data correctness.  */
#ifdef USE_2P
                        for (j=0;j<len+12;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("PING:  data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
                                }
#else
                        for (j=0;j<len;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("PING:  data error\n");
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
                                }
#endif
#endif

#ifdef PERFORMANCE
#ifdef SYSCALL_TIME
			record_mc[i] =
			gamma_time_diff(t_send_e, t_send_s)-time_overhead;
#endif
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)/2-time_overhead;
#endif

		}

		gamma_exit();

#ifdef PERFORMANCE
#ifdef PERF_SHORT
		{
			double mmm,Mmm,medmm,varmm;
			mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef DELAY
		        printf("%d %g\n",len,medmm);
#else
			if (!len)
			        printf("%d %g\n",len,medmm);
			else
			        printf("%d %g\n",len,len/medmm);
#endif
#ifdef FREQUENCIES
		        printf("variance: %g\n", varmm);
			frequencies(record_mm,mmm,Mmm);
#endif
		}
#else
		{
		printf("\n\n");
		printf("PERFORMANCE (microsec, MB/sec) (best < average < worst):\n\n");
		{
#ifdef SYSCALL_TIME
			double mmc,Mmc,medmc,varmc;
#endif
			double mmm,Mmm,medmm,varmm;
#ifdef SYSCALL_TIME
			mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
		        printf("sys call   %g < %g < %g\n", mmc,medmc,Mmc);
#endif
			mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
		        printf("delay       %g < %g < %g\n", mmm,medmm,Mmm);
		        printf("throughput  %g < %g < %g\n", len/mmm,len/medmm,len/Mmm);
			printf("\n\n");

		        printf("VARIANCE:  \n");
#ifdef SYSCALL_TIME
		        printf("sys call %g\n", varmc);
#endif
		        printf("delay %g\n", varmm);
			printf("\n\n");

#ifdef FREQUENCIES
		        printf("FREQUENCIES:\n");
#ifdef SYSCALL_TIME
			printf("sys call:\n");
			frequencies(record_mc,mmc,Mmc);
#endif
			printf("delay:\n");
			frequencies(record_mm,mmm,Mmm);
#endif
			fflush(stdout);

		}
		}
#endif
#endif
	}


	else {
		/*  this is PONG  */
#ifdef USE_2P
		gamma_set_ap_port_2p(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			handler,handler,
			PORT,GO_BACK,data,len+12,
			12
		);
#else
#ifndef USE_BCAST
		gamma_set_passive_port(
			PORT,
			MASTER_NODE,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#else
		gamma_set_passive_port(
			PORT,
			BROADCAST,gamma_my_par_pid(),PORT,
			PORT,GO_BACK,data,len
		);
#endif
#endif
		gamma_sync();
		for (i=0;i<NPROVE;i++) {
#ifdef USE_2P
	for (j=0;j<len+12;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#else
	for (j=0;j<len;j++) {
		if (gamma_my_node() == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif
gamma_sync();
			if (!gamma_wait(PORT,1)) {
#ifdef USE_2P
				gammaSEND2P(PORT,data,12,data+12,len);
#else
				gamma_SEND(PORT,data,len);
#endif
				continue;
			}
			printf("PONG:  ERROR\n");
			fflush(stdout);
			gamma_exit();
			exit(-1);
		}
		gamma_exit();
	}
return 0;
}
