#define NT 8


#undef USE_ISEND
#define USE_FLOWCTL

#undef SEND_TIME
#define PERFORMANCE
#define PERF_SHORT
#undef DELAY
#undef FREQUENCIES

#undef USE_GAMMA_TEST

#undef STREAM
#ifdef STREAM
#undef SEND_TIME
#define max(a,b) ((a) > (b) ? (a) : (b))
#define NSTREAM max(10,(len ? (10000000/len) : 10000))
#else
#define NSTREAM 1
#endif

#define CORRECTNESS

#ifdef USE_FLOWCTL
#define gamma_SEND(a,b,c)  gamma_send_flowctl(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p_flowctl(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend_flowctl(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p_flowctl(a,b,c,d,e)
#else
#define gamma_SEND(a,b,c)  gamma_send(a,b,c)
#define gammaSEND2P(a,b,c,d,e)  gamma_send_2p(a,b,c,d,e)
#define gamma_ISEND(a,b,c)  gamma_isend(a,b,c)
#define gammaISEND2P(a,b,c,d,e)  gamma_isend_2p(a,b,c,d,e)
#endif

#include <linux/gamma/libgamma.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "measure.h"

#undef TRACE
#define MASTER_NODE 0
#define SLAVE_NODE 1


int len;
int me;
pthread_cond_t          sem,tsem;
pthread_mutex_t         lock,tlock;


void gamma_sync_t_init
(unsigned int port)
{

	int i;


	if (!gamma_my_node())
		gamma_set_passive_port(SYNC_PORT-port,
			BROADCAST,gamma_my_par_pid(),SYNC_PORT-port,
			SYNC_PORT-port,GO_AHEAD,NULL,0
		);
	else {
	        for (i=1;i<=8;i++)  /* Max 256 nodes allowed */
        	        if (((gamma_my_node() >> i) << i) != gamma_my_node())
                	        break;
	        gamma_set_passive_port(SYNC_PORT-port,
        	        gamma_my_node() - (1<<(i-1)),
			gamma_my_par_pid(), SYNC_PORT-port,
                	SYNC_PORT-port,GO_AHEAD,NULL,0
      		);
	}
	return;
}


int gamma_sync_t
(unsigned long port)
{
        int i;

        for (i=1;i<256;i<<=1) {  /* Max 256 nodes allowed */
                if (gamma_my_node() & i)
                        break;
                if ((gamma_my_node()+i) >= gamma_how_many_nodes())
                        break;
                if (gamma_wait(SYNC_PORT-port,1))
                        return -GAMMA_ERR_GENERIC;
        }
        if (gamma_SEND(SYNC_PORT-port,NULL,0) < 0)
                return -GAMMA_ERR_GENERIC;
        if (gamma_my_node())
                if (gamma_wait(SYNC_PORT-port,1))
                        return -GAMMA_ERR_GENERIC;
        return 0;
}


void *pp(void *portnumberptr) {

#ifdef PERFORMANCE
#ifdef SEND_TIME
	double record_mc[NPROVE];
#endif
	double record_mm[NPROVE];
#endif
	unsigned char *data,*data1;
	unsigned long port;
	int i;
#ifdef CORRECTNESS
	int j;
#endif
#ifdef USE_ISEND
	int hndl;
#endif
#ifdef USE_GAMMA_TEST
	int prev_val;
#endif
	time_586
		t_send_s,
		t_rec;
#ifdef SEND_TIME
	time_586
		t_send_e;
#endif
#ifdef PERFORMANCE
	double time_overhead;
#endif


	pthread_mutex_lock(&tlock);
	pthread_cond_wait(&tsem,&tlock);
	pthread_mutex_unlock(&tlock);

	port = *(unsigned long *)portnumberptr;
#ifdef USE_GAMMA_TEST
	prev_val = 0;
#endif

//fprintf(stderr,"ciao: [%d %ld] start\n",me,port);

	/*  Allocate data space (+4096 to allow page alignment)  */
	pthread_mutex_lock(&tlock);
	data1 = (unsigned char *)malloc(4096+len*sizeof(unsigned char));
	pthread_mutex_unlock(&tlock);

	/*  Align to page boundary  */
	data = (unsigned char *)(((unsigned long)data1+4096) & (~4095));

	/*  Pin communication data structures in RAM.  */
#if 1
	gamma_mlock(data,len);
#endif

#ifdef CORRECTNESS
	/*  Init the message to be exchanged.  */
	for (j=0;j<len;j++) {
		if (me == MASTER_NODE)
			data[j] = (unsigned char)j;
		else
			data[j] = (unsigned char)0;
	}
#endif


	if (me == MASTER_NODE) {
		gamma_set_passive_port(
			port,
			SLAVE_NODE,gamma_my_par_pid(),port,
			port,GO_BACK,data,len
		);
	}
	else {
		gamma_set_passive_port(
			port,
			MASTER_NODE,gamma_my_par_pid(),port,
			port,GO_BACK,data,len
		);
	}

#ifdef PERFORMANCE
	/*  Computing the gamma_time() overhead...  */
	for (i=0;i<NPROVE;i++) {
		gamma_time(t_send_s);
 		gamma_time(t_rec);
		record_mm[i] = gamma_time_diff(t_rec, t_send_s);
	}
	{
	double Mmm,medmm,varmm;
	mean_variance(record_mm,&time_overhead,&Mmm,&medmm,&varmm);
	}
#endif

	gamma_sync_t(port);

	for (i=0;i<NPROVE;i++) {
#ifdef STREAM
		int stream;
#endif

		if (me == MASTER_NODE) {
#ifdef CORRECTNESS
			for (j=0;j<len;j++) data[j] = (unsigned char)j;
			gamma_sync_t(port);
#endif

#ifdef PERFORMANCE
			gamma_time(t_send_s);
#endif

#ifdef STREAM
			for (stream = NSTREAM; stream; stream--) {
#endif

#ifdef USE_ISEND
			hndl = gamma_ISEND(port,data,len);
#else
			gamma_SEND(port,data,len);
#endif

#ifdef STREAM
			}
#endif
#ifdef PERFORMANCE
#ifdef SEND_TIME
			gamma_time(t_send_e);
#endif
#endif
#ifdef TRACE
printf("[%d]: master sent\n",i);fflush(stdout);
#endif
#ifdef USE_GAMMA_TEST
			while ((gamma_test(port) - prev_val) < 1);
			prev_val = gamma_test(port);
#else
			gamma_wait(port,1);
#endif
#ifdef TRACE
printf("[%d]: master recvd\n",i);fflush(stdout);
#endif
#ifdef PERFORMANCE
 			gamma_time(t_rec);
#endif
#ifdef USE_ISEND
			gamma_wsend(hndl);
#endif

#ifdef CORRECTNESS
			/*  Check data correctness.  */
                        for (j=0;j<len;j++)
                                if (data[j] != (unsigned char)j) {
                                        int k;
                                        printf("[%ld] PING:  data error\n",port);
                                        for (k=j-5;k<j+10;k++)
						printf("%d:%d ",k,data[k]);
                                        printf("\n");
					break;
                                }
#endif

#ifdef PERFORMANCE
#ifdef SEND_TIME
			record_mc[i] =
			gamma_time_diff(t_send_e, t_send_s)-time_overhead;
#endif
#ifdef STREAM
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)-time_overhead;
#else
			record_mm[i] =
			gamma_time_diff(t_rec, t_send_s)/2-time_overhead;
#endif
#endif

		}
		else {
#ifdef CORRECTNESS
			for (j=0;j<len;j++) data[j] = 0;
			gamma_sync_t(port);
#endif

#ifdef USE_GAMMA_TEST
			while ((gamma_test(port) - prev_val) < NSTREAM);
			prev_val = gamma_test(port);
#else
			gamma_wait(port,NSTREAM);
#endif
#ifdef TRACE
printf("[%d]: slave recvd\n",i);fflush(stdout);
#endif
#ifdef STREAM
			gamma_send_flowctl(port,NULL,0);
#else
#ifdef USE_ISEND
			hndl = gamma_ISEND(port,data,len);
			gamma_wsend(hndl);
#else
			gamma_SEND(port,data,len);
#endif
#endif
#ifdef TRACE
printf("[%d]: slave sent\n",i);fflush(stdout);
#endif
		}

	}

	/*
	 *  Force printout of kernel-level time measurements,
	 *  if enabled.  Otherwise nothing happens.
	 */
	gamma_timers(t_send_s,t_rec);

#if 1
	gamma_munlockall();
#endif
	pthread_mutex_lock(&tlock);
	free(data1);
	pthread_mutex_unlock(&tlock);

#ifdef PERFORMANCE
	if (me == MASTER_NODE) {
#ifdef PERF_SHORT
#ifdef SEND_TIME
		double mmc,Mmc,medmc,varmc;
		mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
#ifdef DELAY
	        printf("%d %g\n",len,medmc/NSTREAM);
#else
		if (!len)
		        printf("%d %g\n",len,medmc/NSTREAM);
		else
		        printf("%d %g\n",len,NSTREAM*len/medmc);
#endif
#else
		double mmm,Mmm,medmm,varmm;
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
#ifdef DELAY
	        printf("%d %g\n",len,medmm/NSTREAM);
#else
		if (!len)
		        printf("%d %g\n",len,medmm/NSTREAM);
		else
		        printf("%d %g\n",len,NSTREAM*len/medmm);
#endif
#endif
#else
#ifdef SEND_TIME
		double mmc,Mmc,medmc,varmc;
#endif
#ifndef STREAM
		double mmm,Mmm,medmm,varmm;
#endif
		printf("\n\n");
		printf("PERFORMANCE (microsec, MB/sec) (lowest < average < highest):\n\n");
#ifdef SEND_TIME
		mean_variance(record_mc,&mmc,&Mmc,&medmc,&varmc);
	        printf("total send time        %g < %g < %g\n",
			mmc,medmc,Mmc);
#ifdef STREAM
		if (!len)
	        printf("avg. send time         %g < %g < %g\n",
			mmc/NSTREAM,medmc/NSTREAM,
			Mmc/NSTREAM);
		else
#else
		if (len)
#endif
	        printf("send throughput        %g < %g < %g\n",
			NSTREAM*len/Mmc,NSTREAM*len/medmc,
			NSTREAM*len/mmc);
#endif
#ifndef STREAM
		mean_variance(record_mm,&mmm,&Mmm,&medmm,&varmm);
	        printf("total delay            %g < %g < %g\n",
			mmm,medmm,Mmm);
#ifdef STREAM
		if (!len)
	        printf("avg. delay             %g < %g < %g\n",
			mmm/NSTREAM,medmm/NSTREAM,
			Mmm/NSTREAM);
		else
#else
		if (len)
#endif
	        printf("end-to-end throughput  %g < %g < %g\n",
			NSTREAM*len/Mmm,NSTREAM*len/medmm,
			NSTREAM*len/mmm);
#endif
		printf("\n\n");

	        printf("VARIANCE:  \n");
#ifdef SEND_TIME
	        printf("send time        %g\n", varmc);
#endif
#ifndef STREAM
	        printf("end-to-end delay %g\n", varmm);
#endif
		printf("\n\n");

#ifdef FREQUENCIES
	        printf("FREQUENCIES:\n");
#ifdef SEND_TIME
		printf("send time:\n");
		frequencies(record_mc,mmc,Mmc);
#endif
#ifndef STREAM
		printf("end-to-end delay:\n");
#endif
		frequencies(record_mm,mmm,Mmm);
#endif
		fflush(stdout);

#endif
	}
#endif

//fprintf(stderr,"ciao: [%d %ld] quitting\n",me,port);
	pthread_cond_signal(&sem);
	return 0;

}


#define PTHMUTEXINIT33(mutx)\
do {\
  (mutx).__m_reserved = 0;\
  (mutx).__m_count = 0;\
  (mutx).__m_owner = 0;\
  (mutx).__m_kind = PTHREAD_MUTEX_TIMED_NP;\
  (mutx).__m_lock.__status = 0;\
  (mutx).__m_lock.__spinlock = __LT_SPINLOCK_INIT;\
} while (0)

#define PTHCONDINIT33(sem)\
do {\
  (sem).__c_lock.__status = 0;\
  (sem).__c_lock.__spinlock = __LT_SPINLOCK_INIT;\
  (sem).__c_waiting = 0;\
} while (0)

int main(int argc, char **argv) {

	int i;
	pthread_t p;
	unsigned long ports[NT];


	gamma_init(2,argc,argv);
   	len = atoi(argv[1]);
	me = gamma_my_node();
	PTHMUTEXINIT33(lock);
	PTHMUTEXINIT33(tlock);
	PTHCONDINIT33(sem);
	PTHCONDINIT33(tsem);
	pthread_mutex_lock(&lock);
	for (i=0;i<NT;i++) {
		ports[i] = i+1;
		gamma_sync_t_init(i+1);
	}
	gamma_sync();
	for (i=0;i<NT;i++)
		pthread_create(&p,NULL,pp,&ports[i]);  /*  will "immediately" block  */
	/*  Make "really sure" they have been spawned and are blocked...  */
	sleep(4);
	/*  unblock the threads  */
	pthread_cond_broadcast(&tsem);
/*  Wait for the threads to quit...  */
	for (i=0;i<NT;i++)
		pthread_cond_wait(&sem,&lock);
	pthread_mutex_unlock(&lock);
	/*  Make "really sure" they have quit...  */
	sleep(4);
	gamma_exit();
	return 0;
}
