/***************************************************************************
    COPYRIGHT NOTE

    GAMMA:		Genoa Active Message MAchine
			BIND2 layer:
			Basic Interface for Network Device Drivers.
			Derived from the Linux network driver skge.c
			and the NICs developed By Giuseppe Ciaccio.

    Copyright		(C) 2007 Kiara Ottogalli, Daniel Rosquete
    e-mail:		kottogal@uc.edu.ve, dhrosquete@uc.edu.ve

    This program  or include  file  is free software,  either  original  or
    derived from other  software made available  under the terms of the GNU
    General Public License. You can redistribute and/or modify this program
    or include  file under the  terms of the GNU  General Public License as
    published  by the  Free Software Foundation;  either version 2  of  the
    License, or (at your option) any later version.

    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR   IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.

    END OF COPYRIGHT NOTE
****************************************************************************/

#define MAX_FLUSH_WAIT  100000  /*
				 * Max waiting (usec) for each pending Tx ring
				 * entry to be marked as "done", while waiting
				 * for Tx ring to flush
				 */
#define GAMMA_TX_RING_SIZE 1024	/*
				 *  Size of Tx descriptor ring.
				 *  skge.c: #define MAX_TX_RING_SIZE 1024
				 */
#define GAMMA_RX_RING_SIZE 4096	/*
				 *  Size of Rx descriptor ring.
				 *  skge.c: #define MAX_RX_RING_SIZE
				 */

#define MAX_CREDIT_SIZE 32
#define LOW_WATERMARK 28	/*  Must be > 0 and < MAX_CREDIT_SIZE  */

#ifdef USE_JUMBO_FRAMES
#define FRAME_SIZ 9000		/*  Max payload chunk = 9 Kbytes  */
#else
#define FRAME_SIZ 1500		/*  Max payload chunk = 1476 bytes  */
#endif


#ifdef MORE_NICS_SAMENODE_ROUTING
/*======================================================================*/
/*  GLOBAL MACROS                                                       */
/*======================================================================*/

/* these global macros are currently used only by rx macros */

#define bind2_global_decl \
struct skge_port *global_adapters[MAX_NUM_NICS]; \
struct skge_ring *global_rx_rings[MAX_NUM_NICS]

#define bind2_global_init \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) {\
		global_adapters[i] = gamma_devices[i]->priv; \
		global_rx_rings[i] = global_adapters[i]->rx_ring; \
	} \
} while (0)

#endif


/*======================================================================*/
/*  TX MACROS                                                           */
/*======================================================================*/

#define tx_optimal_frag(l) FRAME_SIZ

#define tx_common_decl \
	struct skge_port *sk; \
	struct skge_ring *tx_ring; \
	struct skge_tx_desc *txd; \
	dma_addr_t addr; \
	int ttail

#define tx_common_init(nic_idx) \
do { \
  sk = netdev_priv(gamma_devices[nic_idx]); \
  tx_ring = &sk->tx_ring; \
} while (0)

#define tx_ring_head (tx_ring->to_clean_integer)
#define tx_ring_tail (tx_ring->to_use_integer)
#define tx_ring_occupancy \
((tx_ring->to_clean_integer <= tx_ring->to_use_integer) ? \
sk->tx_ring.to_use_integer - sk->tx_ring.to_clean_integer : \
GAMMA_TX_RING_SIZE - \
(sk->tx_ring.to_clean_integer - sk->tx_ring.to_use_integer))

/* peppe da fare? */
#define tx_reset_and_restart() \
do { skge_down(sk->netdev); skge_up(sk->netdev); busywait(7000000); } while (0)

#define tx_enq_pkt1(pkt_ptr,pkt_size) \
do { \
  txd = &sk->tx_ring[tx_ring_tail]; \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  addr = virt_to_phys(pkt_ptr); \
  txd->dma_hi = (u64)addr >> 32; \
  txd->dma_lo = (u64)addr & 0xffffffff; \
  incr_modulo(skge_entry,GAMMA_TX_RING_SIZE); \
  wmb(); \
} while (0)

#define tx_enq_pkt(header_ptr,header_size,payload_ptr,payload_size) \
for (;;) { /*  Allows easy escapes in fall-through programming style  */ \
\
  if (header_size+payload_size <= HEADER_SIZE+MAX_1ST_PIECE_SIZE) { \
    memcpy(&(((unsigned char *)header_ptr)[header_size]), \
            (void *)phys_to_virt((unsigned long)(payload_ptr)),payload_size); \
    tx_enq_pkt1(header_ptr,(header_size)+(payload_size)); \
    break; \
  } \
\
  skge_entry = sk->tx_ring->to_use_integer; \
  txd = &sk->tx_ring[skge_entry]; \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  addr = virt_to_phys(header_ptr); \
  txd->dma_hi = (u64)addr >> 32; \
  txd->dma_lo = (u64)addr & 0xffffffff; \
  incr_modulo(skge_entry,GAMMA_TX_RING_SIZE); \
\
  txd = &sk->tx_ring[skge_entry]; \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  txd->dma_hi = (u64)(unsigned long)payload_ptr >> 32; \
  txd->dma_lo = (u64)(unsigned long)payload_ptr & 0xffffffff; \
  incr_modulo(skge_entry,GAMMA_TX_RING_SIZE); \
\
  wmb(); \
  break; \
}\
ttail=(int)sk->tx_ring->to_use

/*Might be a mistake*/\
#define tx_start() pci_read_config_dword(&sk->hw,ttail,skge_entry)

#define tx_forw_tail() sk->tx_ring->to_use = skge_entry

#define tx_flush_ring(tx_desc_idx,n,res) \
do { \
  time_586 a,b; \
  int r; \
\
  res = r = 1; \
  for (;;) { \
    if (tx_ring_head == tx_ring_tail) /* empty ring */ break; \
    txd = &sk->tx_ring[skge_entry]; /*Might be a mistake */\
    gamma_time(a); \
    for (;;) { \
    if (tx_ring_head != sk->hw->idx[0].tx_consumer) { \
	    incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
        break; \
    } \
      gamma_time(b); \
      if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
        r = 0; /* tx desc timed out */ \
        break; \
      } \
    } \
    if (!r) { \
/*      printg( \
         "tx_flush_ring(): failure at pos %d, status %08x (tail at %d).\n", \
          tx_ring_head,tp->hw_status->status,tx_ring_tail); \*/\
      res = 0; /*  report the failure */ \
      do { skge_down(sk->netdev); skge_up(sk->netdev); } while (0) /*Might be wrong */\
      incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
      r = 1; \
    } \
    if (tx_ring_head == (tx_desc_idx)) break; \
  }\
} while (0)

/*  similar to tx_flush_ring(), different guard; assumes non-empty ring  */
#define tx_wait_ring_room(n,res) \
do { \
  time_586 a,b; \
  int r; \
\
  res = r = 1; \
  for (;;) { \
    gamma_time(a); \
    for (;;) { \
    if (tx_ring_head != sk->hw->idx[0].tx_consumer) { \
	    incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
        break; \
    } \
      gamma_time(b); \
      if ((unsigned long)gamma_time_diff_ticks(b,a) > (n)*CLOCK) { \
        r = 0; /* tx desc timed out */ \
        break; \
      } \
    } \
    if (!r) { \
/*      printg( \
         "tx_flush_ring(): failure at pos %d, status %08x (tail at %d).\n", \
          tx_ring_head,tp->hw_status->status,tx_ring_tail); \*/\
      res = 0; /*  report the failure */ \
      do { skge_down(sk->netdev); skge_up(sk->netdev); } while (0) /*Might be wrong */\
      incr_modulo(tx_ring_head,GAMMA_TX_RING_SIZE); \
      r = 1; \
    } \
    /* if (tx_ring_has_room()) */ if (tx_ring_head == tx_ring_tail) break; \
  }\
} while (0)


/*======================================================================*/
/*  RX MACROS                                             	        */
/*======================================================================*/

#define rx_common_decl \
	struct skge_port *sk; \
	struct skge_rx_desc *desc=NULL,*rdesc=NULL; \
	u32 opaque_key = 0, desc_idx, *post_ptr = NULL; \
	u16 hw_idx; \
	int rtail

#define rx_common_init(dev) \
	sk = netdev_priv(dev)

#ifdef MORE_NICS_SAMENODE_ROUTING
#define poll_common_decl \
	?????
#else
#define poll_common_decl_and_init \
	struct skge_port *sk = netdev_priv(gamma_devices[0]); \
	struct skge_rx_desc *desc,*rdesc; \
	u32 opaque_key, desc_idx, *post_ptr; \
	u16 hw_idx
#endif

#define rx_ring_tail hw_idx
#define rx_ring_head (sk->rx_ring->to_clean)
#define read_rx_ring_head(head) (head) = sk->rx_ring->to_clean
#define write_rx_ring_head(head)  sk->rx_ring->to_clean = (head)
#define forward_rx_ring_head() \
do { \
     incr_modulo(sk->rx_ring->to_clean,GAMMA_RX_RING_SIZE);\
     incr_modulo(sk->rx_ring->to_use,GAMMA_RX_RING_SIZE);\
} while (0)

#ifdef MORE_NICS_SAMENODE_ROUTING

#define rx_ring_head_arg(idx) global_rx_rings[idx]->to_clean ???
#define rx_read_pkt_info_arg(idx) ????? \
do { \
} while (0)

#endif

#define clear_rx_irqs()
//tw32_mailbox(MAILBOX_INTERRUPT_0 + TG3_64BIT_REG_LOW, 0x00000000)

#define rx_read_pkt_info_light() \
do { \
  hw_idx = sk->hw->idx[0].rx_producer; \
  rmb(); \
} while (0)

#define rx_read_pkt_info() \
do { \
   rdesc=&sk->rx_ring[rx_ring_head];\
  } \
} while (0)

#define rx_status() (rdesc ? rdesc->status & RXD_ERR_MASK: 0)
#define rx_err_pktnotavail() (rx_ring_head == hw_idx)

#define rx_header(h) \
do { \
/* peppe: on x86, pci_map_single() is equivalent to virt_to_phys() */ \
  (h) = (l2s *)phys_to_virt(((u64)desc->addr_hi << 32) \
				+ ((u64)desc->addr_lo & 0xffffffff)); \
} while (0)

#define discard_pkt() \
do { \
  rdesc->status = 0; \
  wmb(); \
  rtail=(int)sk->rx_ring->to_clean;\
  pci_read_config_dword(&sk->hw,rtail,rx_ring_head);\
  forward_rx_ring_head(); \
} while (0)

#define rx_suspend_irq() \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) \
		skge_suspend(netdev_priv(gamma_devices[i]),0); \
} while (0)

#define rx_resume_irq() \
do { \
	int i; \
	for (i=0; i<num_gamma_devices; i++ ) \
		skge_resume(netdev_priv(gamma_devices[i]),0); \
} while (0)
