/*======================================================================*/
/*  GAMMA GENERIC INCLUDE FILE                                          */
/*======================================================================*/

/* Cannot include asm/desc.h due to conflicting types for idt_table */
#define __ARCH_DESC_H		/*  avoid to include asm/desc.h         */

#include <linux/spinlock.h>     /*  "spin_lock()", etc.                 */
#include <linux/sched.h>        /*  "current->pid", "EXIT_ZOMBIE"       */
#include <asm/io.h>             /*  "outl", etc.                        */
#include <asm/system.h>         /*  "local_save_flags", "wmb()", etc.   */
#include <linux/mm.h>           /*  "GFP_KERNEL", etc.                  */
#include <asm/pgtable.h>        /*  "pte_page", "pte_offset", etc.      */
#include <asm/uaccess.h>        /*  "copy_from_user", "copy_to_user"    */
#include <linux/linkage.h>	/*  "asmlinkage"                        */
#include <linux/netdevice.h>	/*  "struct device", "netif_running()"  */
#include <linux/preempt.h>	/*  "preempt_disable()", etc.           */
#include <linux/pci.h>		/*  "struct pci_driver"                 */
#include <asm/atomic.h>
#ifdef MODULE
#include <linux/module.h>
#endif

typedef struct task_struct task_t;

/*  DO NOT CHANGE THE ORDER OF WHAT FOLLOWS!!  */

#include "lib/gamma_userlev.h"
#include "gamma_def.h"

#if CPU == 32
#include <asm/processor.h>      /*  "struct desc_struct"  */
#elif CPU == 64
#include <asm/desc_defs.h>      /*  "struct gate_struct"  */
#else
#warning CPU is neither 32- nor 64-bit ?!?
#endif

#include "gamma_nic_dependent.h"
#include "gamma_macros.h"
#include "gamma_header.h"
#include "gamma_types.h"

asmlinkage int gamma_system_call(void);         /* Used by gamma_entry.S */
asmlinkage int gamma_ret_from_handler(void);    /* Used by gamma_entry.S */

/*  These are used by the device driver:  */
int gamma_register_driver(struct pci_driver *);
void gamma_unregister_driver(struct pci_driver *);
int rx_gamma(unsigned int);
extern unsigned long		num_gamma_devices;
extern struct net_device	*gamma_devices[MAX_NUM_NICS];
extern int			gamma_initialized;
extern spinlock_t		gamma_susp_irq_count_lock;
extern spinlock_t		gamma_rx_lock;
extern unsigned long		gamma_susp_irq_count;

void gamma_kernel_send                          /* Used by sys_gamma_vm.c */
( unsigned long, unsigned char, unsigned char, void *, unsigned short, int );
long sys_gamma_vm_destroy(u_int32_t);           /* sys_gamma_vm_create()  */
long __sys_gamma_vm_abort(u_int32_t,int,int);   /* gamma_kernel_rec()  */


#ifdef USE_DAMNICK
int damnick_icopy_k2u(void *, void *, unsigned long);
int damnick_icopy_k2u_nu
	(void *, void *, unsigned long, unsigned long *, unsigned long);
int damnick_writel_u(unsigned long *, unsigned long);
#endif
