/***************************************************************************
	Read COPYRIGHT file in this directory
****************************************************************************/


#include "gamma.h"
#include "gamma_leading.h"
static char *gamma_authors = "ciaccio@disi.unige.it";


/************************************************************************
 *  GAMMA Data structures						*
 ************************************************************************/

/*
 * No. and array of NICs probed by the GAMMA driver.
 * Initialized by the driver probe function.
 */
	unsigned long		num_gamma_devices;
	struct net_device	*gamma_devices[MAX_NUM_NICS];
/*
 * Local ethernet addresses, initialized by gamma_start(). 
 * Used by make_send_header() macro to fill the sender field
 * in eth frames header.
 */
	eth_addr		local_hw_addr[MAX_NUM_NICS];

#ifdef MORE_NICS_SAMENODE_ROUTING
/* 
 * Global arrays for adapters and rx_rings.
 * Initialized by gamma_start(). See under bind2/ for details. 
 */
	bind2_global_decl;
#endif

	struct {
		task_t			*task;
		struct net_device	*dev;
	}			gamma_run_handler;

	unsigned long		gamma_irq_disabled;
	unsigned long		total_num_nodes = 0;
	unsigned long		phys_whoami;

	typedef struct {
		volatile task_t		*task;
		int			 pid;
		unsigned int		 dev_idx;
		time_586		 time;
		char			 dummy[12]; /* align */
	} yield_t;
	yield_t			gamma_must_yield;

/*  sys_gamma_suspend_irq()/sys_gamma_resume_irq() balance, and mutex  */
	spinlock_t		gamma_susp_irq_count_lock = SPIN_LOCK_UNLOCKED;
	unsigned long		gamma_susp_irq_count;

/*
 *  Table of (ptrs to) active parallel applications (one entry for each).
 *  Initialized by gamma_start(); entries allocated by sys_gamma_vm_create()
 *  and gamma_kernel_rec().
 */
	par_entry		*par_proc_table[MAX_PAR_PID+1];

/*
 *  Credit-based flow ctl machinery, with congestion control.
 *  Initialized by sys_gamma_getconfig() and gamma_start().
 */
	atomic_t		local_credit[MAX_NUM_NODES];
	unsigned long		initial_credit[MAX_NUM_NODES];
	unsigned long		credit_ceil[MAX_NUM_NODES];
	unsigned long		credit_growth_cnt[MAX_NUM_NODES];
	unsigned long		credit_size;

/*
 *  Per-destination rings of (ptrs to) per-packet tx temp buffers.
 *  Bcast pkts use entry #phys_whoami ("self", which would otherwise be
 *  unused).
 *  Alloc and init by gamma_start(), init by gamma_getconfig().
 */

typedef struct {
	void *b;		/*  Ptr to temp buffer  */
	long l;			/*
				 *  pkt from blocking send routines:
				 *    pkt size (>=0);
				 *  pkt from non-blocking send routines:
				 *    1 plus the NIC Tx ring tail at the time
				 *    when the pkt was enqueued,
				 *    then change sign (<0 therefore).
				 */
	unsigned long hlen;	/*
				 *  non-blocking send only:
				 *  size of header (+ possible 1st chunk,
				 *  in case 2-way gather send was used).
				 */
	unsigned long plen;	/*
				 *  non-blocking send only:
				 *  size of of payload (2nd chunk only,
				 *  in case 2-way gather send was used).
				 */
} tx_tmpbuf_t;

typedef struct {
	tx_tmpbuf_t buf[GAMMA_TX_WIN_SIZE];
	volatile u16 tail;
#ifdef SEQCHECK
/*
 *  Pkt numbering.  Each pkt is stamped by a sequence number.
 *  The current pkt number for destination node #N is stored here;
 *  the expected pkt number from source node #N is stored at rx_seq[N].
 */
	volatile u8 tx_seq;
/*  Pad to doubleword.  */
	u8 dummy;
#else
	u16 dummy;
#endif
} tx_tmp_ring_t;

tx_tmp_ring_t		tx_tmp_ring[MAX_NUM_NODES];
tx_tmp_ring_t		kernel_tx_tmp_ring;

#ifdef SEQCHECK
/*
 *  Table of pkt numbering.  Each pkt is stamped by a sequence number.
 *  The current pkt number for destination node #N is stored at tx_seq[N];
 *  the expected pkt number from source node #N is stored at rx_seq[N].
 *  Initialized by gamma_getconfig().
 *  Bcast pkts have distinct numbering.
 *  DO NOT CHANGE the type, unless also changing in the GAMMA header.
 */
	u8			rx_seq[MAX_NUM_NODES];
	u8			rx_seq_bcast[MAX_NUM_NODES];

/*
 *  Per-node table of flags: 1 iff pkts from given node must be rejected;
 *  1 iff given node has experienced pkt losses (congestion?).
 *  Initialized by sys_gamma_getconfig()
 */
	unsigned long		rx_discard[MAX_NUM_NODES];
	unsigned long		congested[MAX_NUM_NODES];
#endif

/*
 *  Counter of invocations to sys_gamma_poll(), plus a threshold.
 *  The counter is incremented each time a polling phase ends without getting
 *  any new pkt.
 *  When its value get higher than the threshold, the waiting node
 *  may suppose there is a dealock due to one or more missing pkts;
 *  it then asks everybody for their current frame number (by sending TEST
 *  pkt).
 *  This helps detecting missing pkts and thus keeps the GAMMA protocol alive.
 */

#define REPEAT_POLL 200                  /*  200 usec poll phase duration  */
#define MAX_POLLS (1000000/REPEAT_POLL)  /*  1 sec timeout                 */
#define SIGURG_TIMEOUT (100000*CLOCK)    /*  0.1 sec timeout before SIGURG */
#define CEILING_MAX_POLLS 1*MAX_POLLS    /*  Can be increased              */

	int			how_many_polls = 0;
	int			max_polls = MAX_POLLS;
	int			base_max_polls = MAX_POLLS;

/*
 *  spin/lock semaphores for mutual excl on Tx/Rx.
 *  Initialized by gamma_start().
 */
	spinlock_t		gamma_tx_lock;
	spinlock_t		gamma_rx_lock;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	rwlock_t		gamma_rxintranode_lock;
#endif

/*
 *  Order of magnitude of sizeof(par_entry), needed for get_free_pages().
 *  Initialized by gamma_start().
 */
	unsigned long		par_entry_magnitude;

/*  Routing stuff  */
	eth_addr		eth_table[MAX_NUM_NODES];
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned short		routing_table[MAX_NUM_NODES];
#endif

	int			gamma_initialized = 0;

#ifdef MEASURE_OUT_TH
unsigned long out_counter = 0;
unsigned long out_total_amount = 0;
#endif

#ifdef MEASURE_OUT_TH_MEM
unsigned long out_counter_mem = 0;
unsigned long out_total_amount_mem_delay = 0;
unsigned long out_total_amount_mem_bytes = 0;
#endif

#ifdef MEASURE_IN_TH
unsigned long in_counter = 0;
unsigned long in_total_amount = 0;
#endif

#ifdef MEASURE_IN_TH_MEM
unsigned long in_counter_mem = 0;
unsigned long in_total_amount_mem_delay = 0;
unsigned long in_total_amount_mem_bytes = 0;
#endif


#ifdef MODULE
/************************************************************************
 *	Module support							*
 ************************************************************************/

/*
 * GAMMA module needs a table and some functions of Linux that are unexported.
 * Hereafter is a Very Dirty hacking for catching them.
 */

/* Table idt_table[] declared in linux/arch/[ i386 | x86_64 ]/kernel/ */

long idt_table_gammaparam = 0;

module_param(idt_table_gammaparam,long,0);

#if CPU == 32
struct desc_struct *idt_table;
#elif CPU == 64
struct gate_struct *idt_table;
#else
#warning CPU is neither 32- nor 64-bit ?!?
#endif

/* idt_table is initialized in gamma_start() */


/* Function split_vma() defined in linux/mm/mmap.c */

long split_vma_gammaparam = 0;

module_param(split_vma_gammaparam,long,0);

int (*split_vma_ptr)
(struct mm_struct *, struct vm_area_struct *, unsigned long, int);

int split_vma
(struct mm_struct * mm, struct vm_area_struct * vma, unsigned long addr,
int new_below)
{
	split_vma_ptr = (void *)split_vma_gammaparam;
	return split_vma_ptr(mm,vma,addr,new_below);
}


/* Function find_vma_prev() defined in linux/mm/mmap.c */

long find_vma_prev_gammaparam = 0;

module_param(find_vma_prev_gammaparam,long,0);

struct vm_area_struct *(*find_vma_prev_ptr)
(struct mm_struct *, unsigned long, struct vm_area_struct **);

struct vm_area_struct *find_vma_prev
(struct mm_struct *mm, unsigned long addr, struct vm_area_struct **pprev)
{
	find_vma_prev_ptr = (void *)find_vma_prev_gammaparam;
	return find_vma_prev_ptr(mm,addr,pprev);
}


/* Function vma_merge() defined in linux/mm/mmap.c */

long vma_merge_gammaparam = 0;

module_param(vma_merge_gammaparam,long,0);

struct vm_area_struct *(*vma_merge_ptr)
(struct mm_struct *, struct vm_area_struct *, unsigned long,
unsigned long, unsigned long, struct anon_vma *,
struct file *, pgoff_t, struct mempolicy *);

struct vm_area_struct *vma_merge
(struct mm_struct *mm, struct vm_area_struct *prev, unsigned long addr,
unsigned long end, unsigned long vm_flags, struct anon_vma *anon_vma,
struct file *file, pgoff_t pgoff, struct mempolicy *policy)
{
	vma_merge_ptr = (void *)vma_merge_gammaparam;
	return vma_merge_ptr(mm,prev,addr,end,vm_flags,anon_vma,file,
		pgoff,policy);
}


/* Function make_pages_present() defined in linux/mm/memory.c */

long make_pages_present_gammaparam = 0;

module_param(make_pages_present_gammaparam,long,0);

int (*make_pages_present_ptr)
(unsigned long, unsigned long);

int make_pages_present
(unsigned long addr, unsigned long end)
{
	make_pages_present_ptr = (void *)make_pages_present_gammaparam;
	return make_pages_present_ptr(addr,end);
}


/* Function free_uid() defined in linux/kernel/user.c */

long free_uid_gammaparam = 0;

module_param(free_uid_gammaparam,long,0);

void (*free_uid_ptr)
(struct user_struct *);

void free_uid
(struct user_struct *up)
{
	free_uid_ptr = (void *)free_uid_gammaparam;
	return free_uid_ptr(up);
}


/* Function pci_match_device() defined in linux/drivers/pci/pci-driver.c */

long pci_match_device_gammaparam = 0;

module_param(pci_match_device_gammaparam,long,0);

struct pci_device_id * (*pci_match_device_ptr)
(struct pci_driver *, struct pci_dev *);

struct pci_device_id *pci_match_device
(struct pci_driver *drv, struct pci_dev *dev)
{
	pci_match_device_ptr = (void *)pci_match_device_gammaparam;
	return pci_match_device_ptr(drv,dev);
}

#endif  /* MODULE */


/************************************************************************
 *	Some function protos						*
 ************************************************************************/


asmlinkage long sys_gamma_poll(unsigned long,unsigned long);
static inline int is_low_watermark(unsigned long,unsigned long,unsigned long);


/************************************************************************
 *	Init/shutdown routines						*
 ************************************************************************/


static void gamma_start
(void)
{

	int i,j;


#ifdef MODULE
	if (!idt_table_gammaparam) {
printk("GAMMA module error:  missing or incorrect \"idt_table_gammaparam\"\n");
		return;
	}
#if CPU == 32
	idt_table = (struct desc_struct *)idt_table_gammaparam;
#elif CPU == 64
	idt_table = (struct gate_struct *)idt_table_gammaparam;
#else
#warning CPU is neither 32- nor 64-bit ?!?
#endif
#endif

/*
 *  Allocate per-packet tx tmp bufs (one for each Tx ring entry).
 *  We don't yet know how many nodes are actually present in the cluster;
 *  in doubt, we must allocate as many Tx rings as MAX_NUM_NODES.
 *  Later, we de-allocate those one unneeded (gamma_getconfig()).
 */
	for (i=0; i < MAX_NUM_NODES; i++) {
		for (j=0; j < GAMMA_TX_WIN_SIZE; j++) {
			tx_tmp_ring[i].buf[j].b =
				kmalloc(FRAME_SIZ, GFP_KERNEL);
			if (tx_tmp_ring[i].buf[j].b == NULL) {
				printg("gamma_start():  kmalloc() failed\n");
				return;
			}
			tx_tmp_ring[i].buf[j].l = 0;
		}
		tx_tmp_ring[i].tail = 0;
#ifdef SEQCHECK
		tx_tmp_ring[i].tx_seq = 0;
#endif
	}
	for (j=0; j < GAMMA_TX_WIN_SIZE; j++) {
		kernel_tx_tmp_ring.buf[j].b =
			kmalloc(FRAME_SIZ, GFP_KERNEL);
		if (kernel_tx_tmp_ring.buf[j].b == NULL) {
			printg("gamma_start():  kmalloc() failed\n");
			return;
		}
		kernel_tx_tmp_ring.buf[j].l = 0;
	}
	kernel_tx_tmp_ring.tail = 0;
#ifdef SEQCHECK
	kernel_tx_tmp_ring.tx_seq = 0;
#endif
	gamma_run_handler.task = NULL;
	gamma_must_yield.task = NULL;
	gamma_must_yield.pid = 0;
	gamma_must_yield.time.hi = 0;
	gamma_must_yield.time.lo = 0;
	gamma_irq_disabled = 0;
	gamma_susp_irq_count = 0;

/*  The table of parallel processes is empty initially  */
	for (i=0; i <= MAX_PAR_PID; i++)
		par_proc_table[i] = NULL;

/*  Initial credit size is max (as defined by constant MAX_CREDIT_SIZE).  */
	credit_size = MAX_CREDIT_SIZE;

/*  Compute the magnitude of sizeof(par_entry), needed to allocate GAMMA vm  */
	for (i=1, j=0; i*PAGE_SIZE<sizeof(par_entry); i<<=1, j++);
	par_entry_magnitude = j;

	spin_lock_init(&gamma_tx_lock);
	spin_lock_init(&gamma_rx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	rwlock_init(&gamma_rxintranode_lock);
#endif

/*  Set the entry points for the GAMMA primitives (trap 0x21)  */
	gamma_set_system_gate(0x21, &gamma_system_call);

#ifdef MORE_NICS_SAMENODE_ROUTING
/*  Initialize global arrays of adapters and rx_rings */
	bind2_global_init;
#endif

/*
 *  Initialize array of local mac addresses from probed nics. 
 *  If multiple nics support is disabled, only first entry in this array is
 *  significant.
 */
	for (i=0; i<num_gamma_devices; i++) {
		for (j=0; j<6; j++)
			local_hw_addr[i].addr[j] = gamma_devices[i]->dev_addr[j];
		local_hw_addr[i].addrl =
			((unsigned long)local_hw_addr[i].addr[0]) +
			((unsigned long)local_hw_addr[i].addr[1] << 8) +
			((unsigned long)local_hw_addr[i].addr[2] << 16) +
			((unsigned long)local_hw_addr[i].addr[3] << 24);
		local_hw_addr[i].addrw =
			((unsigned long)local_hw_addr[i].addr[4]) +
			((unsigned long)local_hw_addr[i].addr[5] << 8);
	}


	printg("\n");
	printg("GAMMA  Genoa Active Message MAchine (Linux 2.6) on %s  "
		"15 Apr 2009\n", gamma_devices[0]->name);

#ifdef DRIVER_DEBUG_EVERYTHING
	printg("       Debug level: Very High\n");
#elif defined DRIVER_DEBUG
	printg("       Debug level: High\n");
#elif defined DRIVER_DEBUG_LIGHT
	printg("       Debug level: Low\n");
#else
	printg("       Debug level: Null\n");
#endif
#ifdef SAFE_TX
	printg("       safe tx enabled\n");
#endif
#ifdef USE_JUMBO_FRAMES
	printg("       jumbo frame support enabled:  MTU %d\n",FRAME_SIZ);
#else
	printg("       jumbo frame support disabled\n");
#endif
	for(i=0; i < num_gamma_devices; i++)
		printg("       Registered with NIC %s\n",
			gamma_devices[i]->name);
#ifdef MORE_NICS_SAMENODE_ROUTING
	printg("       Multiple NICs support enabled\n");
        printg("       Found %ld devices:\n", num_gamma_devices);
	for(i=0; i < num_gamma_devices; i++) {
		printg("              Device %d: MAC Address %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n", i,
		       gamma_devices[i]->dev_addr[0],
		       gamma_devices[i]->dev_addr[1],
		       gamma_devices[i]->dev_addr[2],
		       gamma_devices[i]->dev_addr[3],
		       gamma_devices[i]->dev_addr[4],
		       gamma_devices[i]->dev_addr[5]); 
	}
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	printg("       multiple process instances per node: enabled\n");
	printg("               (max %d instances/node)\n", MAX_NUM_LOCAL_INST);
#else
	printg("       multiple process instances per node: disabled\n");
#endif
#ifdef NIC_IS_TG3
	printg("       Size of Tx ring %d\n", GAMMA_TX_RING_SIZE);
#else
	printg("       Size of Tx ring %d     Size of Rx ring %d\n",
		GAMMA_TX_RING_SIZE, GAMMA_RX_RING_SIZE);
#endif
	printg("       GAMMA protocol: %s\n",gamma_authors);
	printg("       BIND2 layer   : %s\n\n",bind2_authors);

	return;
}


void gamma_shutdown
(void)
{

	int i,j;


	for (i=0; i < MAX_NUM_NODES; i++)
		for (j=0; j < GAMMA_TX_WIN_SIZE; j++)
			if (tx_tmp_ring[i].buf[j].b != NULL)
				kfree(tx_tmp_ring[i].buf[j].b);
	for (j=0; j < GAMMA_TX_WIN_SIZE; j++)
		if (kernel_tx_tmp_ring.buf[j].b != NULL)
			kfree(kernel_tx_tmp_ring.buf[j].b);
	gamma_unset_system_gate(0x21);
	gamma_initialized = 0;
	printg("GAMMA driver now inactive\n");
	return;

}


/* Very Dirty hacking to allow pci device predation by the GAMMA driver */

#define MAX_NUM_TOTAL_NICS 8
struct pci_dev *gamma_candidate[MAX_NUM_TOTAL_NICS];
struct device_driver *gamma_abusive_driver;
int gamma_cand_last;


static int __gamma_pci_scan_dev
( struct device *dev, void *data )
{
	struct pci_dev *pdev = to_pci_dev(dev);
	struct pci_driver *drv = (struct pci_driver *)data;

	if (gamma_cand_last < MAX_NUM_TOTAL_NICS) {
		if (pci_match_device(drv,pdev) != NULL) {

DEBUG(printk("GAMMA:  found device %p\n", pdev));

			gamma_candidate[gamma_cand_last++] = pdev;
		}
	}
	return 0;
}


static void gamma_build_candidate_list
(struct pci_driver *drv)
{

DEBUG(printk("GAMMA:  scanning for potential GAMMA devices...\n"));

	gamma_cand_last = 0;
	bus_for_each_dev(&pci_bus_type,NULL,drv,__gamma_pci_scan_dev);
/* the above call will possibly change gamma_cand_last */
}


static void gamma_free_dev
(int n)
{
	struct pci_dev *pdev;
	int i,j;

	gamma_abusive_driver = NULL;
	for (i=0,j=gamma_cand_last-1; i<n && j>=0; i++,j--) {
		pdev = gamma_candidate[j];
/* take note of the original driver, will try to restore it when rmmod */
		gamma_abusive_driver = &pdev->driver->driver;

DEBUG(printk(
"GAMMA:  freeing pdev %p from driver %s...\n",pdev,pdev->driver->name
));

		device_release_driver(&pdev->dev);
	}
}


int gamma_register_driver
(struct pci_driver *drv)
{
	int ret;

	num_gamma_devices = 0;
	ret = pci_register_driver(drv);
	if (ret >= 0)
		if (num_gamma_devices >= MAX_NUM_NICS) {
/* All required GAMMA devices now registered with the GAMMA driver */
			gamma_start();
			return ret;
		}
/*
 * Failed to register the required no. of GAMMA NICs.
 * Try to predate the required GAMMA NICs from possible "abusive" drivers.
 * This includes the devices that have just been recognized by this driver.
 */

/* Workaround for bypassing already-registered GAMMA driver: change name */
	drv->name[0]++;

	gamma_build_candidate_list(drv);
	gamma_free_dev(MAX_NUM_NICS);
/*
 * All required devices should now be free, including those ones previously
 * recognized by this driver itself.  Now try to re-register with all
 * devices again
 */

DEBUG(printk("GAMMA:  try to re-register\n"));

	num_gamma_devices = 0;
	ret = pci_register_driver(drv);
	if (ret >= 0)
		if (num_gamma_devices > 0) {
/* We did our best to find the GAMMA devices, and this is what we got */
			gamma_start();
			return ret;
		}
/* Failed to detect any GAMMA devices.  Bail out. */
	return ret;
}


void gamma_unregister_driver
(struct pci_driver *drv)
{
	int retval;

	pci_unregister_driver(drv);
	if (num_gamma_devices > 0) {
		gamma_shutdown();
		if (gamma_abusive_driver != NULL) {
/* try to restore the original driver for the now free devices */
			retval = driver_attach(gamma_abusive_driver);

printk(KERN_EMERG "GAMMA: original driver %s reattached to network devices\n",
gamma_abusive_driver->name);

		}
	}
}


/************************************************************************
 *	Service send/receive routines					*
 ************************************************************************/


/*  CAUTION:  do not use flow ctl here!  */
void gamma_kernel_send
(
	unsigned long phys_dest,
	unsigned char par_pid,
	unsigned char kind,
	void *data,
	unsigned short len,
	int use_lock
)
{

	tx_tmp_ring_t			*win;
	unsigned long			entry;
	l2s				*h;
	int				i;
	tx_common_decl;

	if (use_lock) spin_lock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	if (phys_dest != BROADCAST)
		tx_common_init( routing_table[phys_dest] );
	else {
		/* luigi: gamma_kernel_send broadcast is actually a software implementation. 
		 * should be changed in next releases. */
DEBUG_E(printg("gamma_kernel_send: entering software broadcast.\n"));

		for (i=0; i<total_num_nodes; i++) {
			if (i != phys_whoami)
				gamma_kernel_send(i,par_pid,kind,data,len,0);
		}
		if (use_lock) spin_unlock(&gamma_tx_lock);
		return;
	}
#else
	tx_common_init(0);
#endif

DEBUG_E(printg("gamma_kernel_send: entering...\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(printg("     Send to physical node %lu through NIC %d\n",
	       phys_dest,
	       routing_table[phys_dest]));
#endif

/*
 *  Pkts of the flow-ctl protocol must be kept into the
 *  Tx tmp ring (for possible retransmission, and to keep the numbering
 *  consistent); other kernel pkts need not.
 */
	if (kind == FILL_CREDIT || kind == CRED_REQ)
		if (phys_dest != BROADCAST)
			win = &tx_tmp_ring[phys_dest];
		else
			win = &tx_tmp_ring[phys_whoami];
	else
		win = &kernel_tx_tmp_ring;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;

	if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:
#ifdef MORE_NICS_SAMENODE_ROUTING
	/* Now broadcast is a sw one, but this code is ready for hw implementation. */
	make_send_header(phys_dest,par_pid,kind,0,ETH_GAMMA_KERNEL,1,len,0,h,
			 (phys_dest == BROADCAST)? 0 : routing_table[phys_dest]);
#else
	make_send_header(phys_dest,par_pid,kind,0,ETH_GAMMA_KERNEL,1,len,0,h,0);
#endif

#ifdef SEQCHECK
	if (kind == FILL_CREDIT || kind == CRED_REQ) {
/*
 *  Pkts of the flow-ctl protocol must be kept numbered,
 *  as they are kept into the Tx tmp ring for possible retransmission.
 */
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
	}

DEBUG_F(
if (kind == FILL_CREDIT)
	printg("send FILL_CREDIT to node %ld, frame_no %d\n",
		phys_dest,read_send_header_frameno(h));
if (kind == CRED_REQ)
	printg("send CRED_REQ to node %ld, frame_no %d\n",
		phys_dest,read_send_header_frameno(h));
if (kind == ACK)
	printg("send ACK to node %ld, frame_no %d\n",
		phys_dest,*(unsigned char *)data);
if (kind == ACKB)
	printg("send ACKB to node %ld, frame_no %d\n",
		phys_dest,*(unsigned char *)data);
)

#endif

	memcpy((u8 *)h+HEADER_SIZE,data,len);

DEBUG_E(printg(
"gamma_kernel_send(): %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll
));

DEBUG(printg(
"gamma_kernel_send(): sending to node %ld, kind %d\n",
phys_dest,kind
));

	tx_enq_pkt1(h,HEADER_SIZE+len);
	tx_start();
	tx_forw_tail();

/*
 * this is a workaround to a hw bug of e1000: the tx engine occasionally hangs.
 * Flushing the tx ring can detect and solve this in a few seconds, and also
 * helps the tx engine not to hang.  For other NICs, this is just harmless.
 */
	tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);

#ifdef SAFE_TX
	if (tx_ring_occupancy >= STEPS_THEN_STOP)
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
	win->buf[entry].l = HEADER_SIZE+len;

end:

	if (use_lock) spin_unlock(&gamma_tx_lock);
	return;

ring_full:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (i) goto room_in_tx_ring;

DEBUG_L(printg("gamma_kernel_send(): Transmit failure before send.\n"));

	goto end;

}


#ifdef SEQCHECK
/*
 *  Scan a given tx tmp ring to retrieve pkts which are to be retransmitted.
 */
static int gamma_scan_tx_tmp_ring
(unsigned long phys_dest, u32 frameno, int is_bcast)
{

	tx_tmp_ring_t			*win;
	int				i;
	l2s				*h;


	if (!is_bcast)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

/*  scan the tx tmp ring backward (modulo ring size)  */
	for (i = win->tail-1; i != win->tail; i--) {

		if (i < 0) i += GAMMA_TX_WIN_SIZE;
		h = win->buf[i].b;

DEBUG(printg(
"gamma_scan_tx_tmp_ring(): (search %u, pos %d) %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
frameno, i, h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll
));

		if (read_send_header_frameno(h) != frameno)
			continue;
/*  Pkt found.  */
		return i;

	}

/*  Pkt not found.  */
	return -1;

}


/*  Re-enqueue pkts to be retransmitted.  */
/*  CAUTION:  do not use flow ctl here!  */
static int gamma_retx
(unsigned long phys_dest, unsigned long pos, int is_bcast)
{

	tx_tmp_ring_t			*win;
	int				ii;
	int				tx_handle;
	l2s				*h;
	int				i;
	tx_common_decl;


#ifdef MORE_NICS_SAMENODE_ROUTING
	tx_common_init( routing_table[phys_dest] );
#else
	tx_common_init(0);
#endif

	if (!is_bcast)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

/*  loop over the tx tmp ring from "pos" to end.  */
	for (ii = pos; ; ) {

		h = win->buf[ii].b;
		tx_handle = -(win->buf[ii].l) - 1;
		if (tx_handle >= 0) {
#if 0
/*
 *  Pkt is from a non-blocking send.  We must check it is still pending in
 *  the Tx ring so that we can still retransmit it
 *  (according to the non-blocking semantics of isend() routines).
 */
			if ((!(tx_ring_occupancy))
			|| (tx_handle > tx_ring_tail)
			|| ((tx_ring_head < tx_ring_tail)
				&& (tx_handle <= tx_ring_head))
		) {
#endif

DEBUG_L(printg(
"Could not retransmit a flushed pkt originated from a non-blocking send()\n"
));

				goto skip;
#if 0
			}
#endif
		}
/*
 *  This pkt is OK for retransmission.  But we have to change its type
 *  to "retransmitted".
 *  Rationale:  pkts in the Tx tmp ring can be bcast; if we retransmit them,
 *  we must prevent non-involved nodes to catch them twice!
 */
		if (rec_header_frmtype(h) == ETH_GAMMA)
			set_send_header_frmtype(h,ETH_GAMMA_RETX);
		else if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL)
			set_send_header_frmtype(h,ETH_GAMMA_KERNEL_RETX);
/*  Enqueue pkt for retransmit.  */
#if 0
		if (tx_handle >= 0) {
printg("ciao retx isend pkt, addr %8.8lx\n",
tx_entry_2_payloadptr((tx_handle+GAMMA_TX_RING_SIZE-1) % GAMMA_TX_RING_SIZE));
/*  Pkt from non-blocking send:  copy payload from user to kernel space.  */
			memcpy(
((u8 *)h) + win->buf[i].hlen,
tx_entry_2_payloadptr((tx_handle+GAMMA_TX_RING_SIZE-1) % GAMMA_TX_RING_SIZE),
win->buf[i].plen
			);
printg("ciao fatto\n");
		}
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

DEBUG_F(printg(
"gamma_retx(): -> %ld, frm %d, pos %d: %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
phys_dest,read_send_header_frameno(h),ii,
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll
));

		tx_enq_pkt1(h,win->buf[ii].l);
		tx_start();
		tx_forw_tail();
/*
 * this is a workaround to a hw bug of e1000: the tx engine occasionally hangs.
 * Flushing the tx ring can detect and solve this in a few seconds, and also
 * helps the tx engine not to hang.  For other NICs, this is just harmless.
 */
		tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif

skip:

		incr_modulo(ii,GAMMA_TX_WIN_SIZE);
		if (ii == win->tail)
			break;

	}

	return 0;

ring_full:

	tx_wait_ring_room(MAX_FLUSH_WAIT,i);
	if (i) goto room_in_tx_ring;

DEBUG_L(printg("gamma_retx(): Transmit failure before send.\n"));

	return -1;

}
#endif


static void gamma_kernel_rec
(
	void *addr,
	unsigned char par_pid,
	unsigned short len,
	unsigned char kind,
	unsigned long phys_sender
)
{

	par_entry		*p;
	port_entry		*pp;
	u32			frame_no;
	int			i,j,pos,procnum,found;
	u_int32_t		rnd_key,pid;


	if (phys_sender >= MAX_NUM_NODES) return;

	p = par_proc_table[par_pid];

	switch (kind) {

	case FILL_CREDIT:

#ifdef SEQCHECK
		if (!congested[phys_sender]) {
			if (initial_credit[phys_sender]
				>= credit_ceil[phys_sender]
			) {
				/*  Credit size raises the allowed ceiling.  */
				if (credit_ceil[phys_sender] >= credit_size)
					/*  Credit ceiling is max.  */
					goto out;
				/*
				 *  Credit ceiling can grow.
				 *  We gradually (and slowly) push up the
				 *  ceiling until max is reached.
				 */
				if (++credit_growth_cnt[phys_sender] >= 1000) {
					credit_growth_cnt[phys_sender] = 0;
					credit_ceil[phys_sender]++;
				}
				goto out;
			}
			/*
			 *  Push up the credit size till current ceiling.
			 *  Do not push up the ceiling yet.
			 */
			initial_credit[phys_sender]++;
			goto out;
		}
/*  A possible congestion on node "phys_sender".  Slow down...  */
		congested[phys_sender] = 0;
/*  Shrink down the credit ceiling (just a bit)  */
		credit_ceil[phys_sender]
			= gmax(2,initial_credit[phys_sender]-1);
		credit_growth_cnt[phys_sender] = 0;
/*  Shrink down the credit size (quite a lot)  */
		if (initial_credit[phys_sender] > 2)
			initial_credit[phys_sender]
				= gmax(2,initial_credit[phys_sender]>>1);
/*  This is a less aggressive shrink rule...I could not do any meaningful
comparison with the previous one yet :(
				= gmax(2,initial_credit[phys_sender]-1);
*/
out:

#endif
/*  Refill credit, but do not increase past the max limit  */
		if (atomic_read(&local_credit[phys_sender])
		  + initial_credit[phys_sender] <= 2 * MAX_CREDIT_SIZE)
			atomic_add(
				initial_credit[phys_sender],
				&local_credit[phys_sender]);
		else
			atomic_set(
				&local_credit[phys_sender],
				2 * MAX_CREDIT_SIZE);
		return;

	case CRED_REQ:

	 	gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
		return;

#ifdef SEQCHECK
	case ACK:

/*  Check the frame numbering of "phys_sender" (unicast).  */
		frame_no = *(u32 *)addr;
		if (frame_no == rx_seq[phys_sender]) {

DEBUG(printg("ACK from node %ld, frame_no %u\n",phys_sender,frame_no))

/*  No missing pkt was detected. Increase the time interval to next check.  */
			if (max_polls < CEILING_MAX_POLLS)
				max_polls +=
				base_max_polls/(2*(total_num_nodes-1));
			return;
		}
/*  Missing pkts from "phys_sender". See rx_gamma(), label "err_frame_no"  */

DEBUG_L(printg(
"gamma_kernel_rec(): ACK #%u, expected #%u, phys sender %ld\n",
frame_no,rx_seq[phys_sender],phys_sender
))

		if (((frame_no-rx_seq[phys_sender]+256)%256)
			> GAMMA_TX_WIN_SIZE) {

DEBUG_L(printg("Unfeasible ack?  Discard...\n"));

			return;
		}

		max_polls = base_max_polls;
		gamma_kernel_send(BROADCAST,DONT_MIND,CONGESTION,NULL,0,1);
		frame_no = rx_seq[phys_sender];
		gamma_kernel_send(
			phys_sender,DONT_MIND,NACK,&frame_no,
			sizeof(frame_no),1);
		rx_discard[phys_sender] = 1;
		return;

	case ACKB:

/*  Check the frame numbering of "phys_sender" (bcast).  */

		frame_no = *(u32 *)addr;
		if (frame_no == rx_seq_bcast[phys_sender]) {

DEBUG(printg("ACKB from node %ld, frame_no %u\n",phys_sender,frame_no))

/*  No missing pkt was detected. Increase the time interval to next check.  */
			if (max_polls < CEILING_MAX_POLLS)
				max_polls +=
				base_max_polls/(2*(total_num_nodes-1));
			return;
		}
/*  Missing pkts from "phys_sender". See rx_gamma(), label "err_frame_no"  */

DEBUG_L(printg(
"gamma_kernel_rec(): ACKB #%u, expected #%u, phys sender %ld\n",
frame_no,rx_seq_bcast[phys_sender],phys_sender
))

		if (((frame_no-rx_seq_bcast[phys_sender]+256)%256)
			> GAMMA_TX_WIN_SIZE) {

DEBUG_L(printg("Unfeasible bcast ack?  Discard...\n"));

			return;
		}

		frame_no = rx_seq_bcast[phys_sender];
		gamma_kernel_send(
			phys_sender,DONT_MIND,NACKB,&frame_no,
			sizeof(frame_no),1);
		rx_discard[phys_sender] = 1;
		return;

	case CONGESTION:

/*
 *  A possible congestion on node "phys_sender".  Slow down at next credit
 *  renewal.
 */
		congested[phys_sender] = 1;
		return;

	case NACK:

/*  NACK for missing unicast pkts.  */

                frame_no = *(u32 *)addr;

DEBUG_L(printg("NACK from node %ld, frame %u\n",phys_sender,frame_no));

		spin_lock(&gamma_tx_lock);
		pos = gamma_scan_tx_tmp_ring(phys_sender,frame_no,0);
		if (pos >= 0) {
/*  Go-back-N and retransmit.  */
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_START,NULL,0,0);
			gamma_retx(phys_sender,pos,0);
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_END,NULL,0,0);
			spin_unlock(&gamma_tx_lock);
			return;
		}

DEBUG_L(printg(
"Missing pkt %u from node %ld:  not found. Cannot retransmit!\n",
frame_no,phys_sender
));

		gamma_kernel_send(phys_sender,DONT_MIND,ERR_NO_RETX,addr,len,0);
		spin_unlock(&gamma_tx_lock);
		return;

	case NACKB:

/*  NACK for missing bcast pkts.  */

                frame_no = *(u32 *)addr;

DEBUG_L(printg("NACKB from node %ld, bcast frame %u\n",phys_sender,frame_no));

		spin_lock(&gamma_tx_lock);
		pos = gamma_scan_tx_tmp_ring(phys_sender,frame_no,1);
		if (pos >= 0) {
/*  Go-back-N and retransmit.  */
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_START,NULL,0,0);
			gamma_retx(phys_sender,pos,1);
			gamma_kernel_send(
				phys_sender,DONT_MIND,RETX_END,NULL,0,0);
			spin_unlock(&gamma_tx_lock);
			return;
		}

DEBUG_L(printg(
"Missing bcast pkt %u from node %ld:  not found. Cannot retransmit!\n",
frame_no,phys_sender
));

		gamma_kernel_send(phys_sender,DONT_MIND,ERR_NO_RETX,addr,len,0);
		spin_unlock(&gamma_tx_lock);
		return;

	case TEST:

/*
 *  This node is asked to spread its table of frame numbering (unicast
 *  as well as bcast);
 *  this can help recovering from a deadlock in case of pkt losses.
 */

DEBUG(printg("TEST from node %ld\n",phys_sender))

		if (total_num_nodes == 0)
/*
 *  Maybe this machine has rebooted and has not yet called
 *  sys_gamma_getconfig(), while other ones are waiting for pkts.
 *  Discard this request.
 */
			return;

		spin_lock(&gamma_tx_lock);

		frame_no = tx_tmp_ring[phys_sender].tx_seq;
		gamma_kernel_send(
			phys_sender,DONT_MIND,ACK,
			&frame_no,sizeof(frame_no),0);
		frame_no = tx_tmp_ring[phys_whoami].tx_seq;
		gamma_kernel_send(
			phys_sender,DONT_MIND,ACKB,
			&frame_no,sizeof(frame_no),0);

		spin_unlock(&gamma_tx_lock);

		return;

	case RETX_START:

/*
 *  A sequence of retransmitted pkts begins.
 *  Since this may occur only after detecting a missing pkt in the
 *  sequence from node "phys_sender", we should have set
 *  rx_discard[phys_sender] = 1.
 */
		if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg("Start of retx sequence, sender %ld\n",phys_sender));

			rx_discard[phys_sender] = 2;
			return;
		}

DEBUG_F(printg(
"gamma_kernel_rec():  Start of double retransmission from node %ld(!!!)\n",
phys_sender
));

		return;

	case RETX_END:

/*
 *  A sequence of retransmitted pkts ends.
 *  Since this may occur only after a sequence of retransmitted pkts
 *  from node "phys_sender", we should have set
 *  rx_discard[phys_sender] = 2.
 */
		if (rx_discard[phys_sender] == 2) {

DEBUG_F(printg("End of retx sequence, sender %ld\n",phys_sender));

			rx_discard[phys_sender] = 0;
			return;
		}
		if (rx_discard[phys_sender] == 1) {
/*
 *  Asked for retransmission but did not catch the RETX_START pkt;
 *  or, detected missing pkt during a retransmission sequence and
 *  therefore refusing remaining part of sequence, but did not yet get
 *  RETX_START.
 */

DEBUG_F(printg(
"gamma_kernel_rec():  unexpected RETX_END (missing RETX_START?)\n"
));

			return;
		}

DEBUG_F(printg(
"gamma_kernel_rec():  End of double retransmission from node %ld(!!!)\n",
phys_sender
));

		return;
#endif  /*  SEQCHECK  */

	case VM_WISH:

/*  A remote proc wishes to join a virtual GAMMA identified by key  */

/*  find the virtual GAMMA by key (par_pid #0 is reserved) */
		rnd_key = (*(wish_t *)addr).key;
		pid = (*(wish_t *)addr).pid;
		p = NULL;
		found = 0;
		for (i=1; i <= MAX_PAR_PID; i++) {
			p = par_proc_table[i];
			if (p == NULL) continue;
			if (p->vm.random_key != rnd_key) continue;
			found = 1;
			break;
		}
		if (!found)
/* maybe data struct for virtual GAMMA not created yet.  ignore... */
			return;

/*  Catch the wish if not already done so.  */
/*  Mutex mandatory on p->wish_pid[], reuse gamma_tx_lock for this.  */
		spin_lock(&gamma_tx_lock);
		for (i=0;i<MAX_NUM_LOCAL_INST;i++) {
			if (p->wish_pid[phys_sender*MAX_NUM_LOCAL_INST+i]
			== pid) {
				/* wish already taken */
				spin_unlock(&gamma_tx_lock);
				return;
			}
			if (p->wish_pid[phys_sender*MAX_NUM_LOCAL_INST+i]
			!= -1)
				continue;

DEBUG_L(printg(
"%s: got wish from node %ld pid %u for virtual GAMMA key %u\n",
__FUNCTION__,phys_sender,pid,rnd_key
));

			p->vm.num_instances++;
			p->wish_pid[phys_sender*MAX_NUM_LOCAL_INST+i] = pid;
			spin_unlock(&gamma_tx_lock);
			return;
		}

printg(
"%s: reject wish from node %ld pid %u for virtual GAMMA key %u: no free slot\n",
__FUNCTION__,phys_sender,pid,rnd_key
);

		spin_unlock(&gamma_tx_lock);
		return;

	case VM_ABORT:

		if (p == NULL) {

DEBUG(printg(
"gamma_kernel_rec(): trying to abort non-existant virtual GAMMA #%d from remote node\n",
par_pid
));

			return;
		}

		if (p->vm.random_key != *(typeof(p->vm.random_key) *)addr) {

DEBUG(printg(
"gamma_kernel_rec(): trying to abort wrong virtual GAMMA #%d from remote node\n",
par_pid
));

			return;
		}

		if (p->quitting) {

DEBUG(printg(
"gamma_kernel_rec(): virtual GAMMA #%d already quitting\n",par_pid
));

			return;
		}

		p->quitting = 1;

/*  Ask all local instances (if any) to abort (SIGTERM sighandler)  */
		for (i=0; i < MAX_NUM_LOCAL_INST; i++)
			if (
				p->inst_table[i].whoami >= 0 &&
				p->inst_table[i].pid > 0 &&
				p->inst_table[i].task != NULL &&
				! (p->inst_table[i].task->state &
					(EXIT_ZOMBIE | EXIT_DEAD | TASK_DEAD)
				)
			) {
				force_sig(SIGTERM, p->inst_table[i].task);

DEBUG_L(printg("gamma_kernel_rec(): aborting par_pid %d procnum %d...\n",
par_pid,i));

			}

		return;


	case VM_CREATE:

		if (p != NULL) {

DEBUG_E(printg(
"gamma_kernel_rec(): virtual GAMMA #%d already exists, ignore\n",par_pid
));

			return;
		}

/*  Allocate a descriptor for a new virtual GAMMA  */
		p = (par_entry *)__get_free_pages(GFP_ATOMIC,
					par_entry_magnitude);
		if (p == NULL) {
printg("gamma_kernel_rec(): kmalloc() failed creating virtual GAMMA #%d\n",
par_pid);
			return;
		}
		memset(p,0,sizeof(par_entry));
		for (i=0; i < MAX_NUM_LOCAL_INST; i++) {
			p->inst_table[i].pid = -1;
			p->inst_table[i].task = NULL;
			p->inst_table[i].whoami = -1;
		}

/*  Get and store the virtual GAMMA descriptor  */
                memcpy(&(p->vm),addr,len);
/*
 *  Compute the table of ethernet addresses of the new virtual GAMMA, and
 *  instance # ("whoami") of all the "procnum" processes to be run on the
 *  local node
 *  (-1 iff local node is not involved at all in the virtual GAMMA).
 */
		procnum = 0;
		for (i=0; i < p->vm.num_instances; i++) {
			if (p->vm.node_list[i] != phys_whoami) continue;
/*
 *  Involved node: set "whoami" to the instance # and allocate room for
 *  per-port pre-computed pkt headers
 */
			procnum++;
			if (procnum > MAX_NUM_LOCAL_INST) {
printg("gamma_kernel_rec():  too many proc instances on this node when creating virtual GAMMA %d.  Aborting...\n",par_pid);
/*  Abort virtual GAMMA locally and globally.  */
				__sys_gamma_vm_abort(par_pid,1,0);
				return;
			}
			p->inst_table[procnum-1].whoami = i;
			for (j=0; j <= MAX_PORT_NUMBER; j++) {
				pp = &p->inst_table[procnum-1].port_table[j];
				pp->semaphore =
				pp->boundary_1p =
				pp->in_size =
				pp->in_room = 0;
				pp->in_buffer =
				pp->in_progress = NULL;
				pp->buffer_kind = GO_BACK;
#ifdef MORE_INSTANCES_ON_SAME_NODE
				pp->partner_is_intra_node =
				pp->bcast_was_susp = 0;
#endif
				pp->header = kmalloc(
					HEADER_SIZE+MAX_1ST_PIECE_SIZE,
					GFP_ATOMIC);
				if (pp->header == NULL) {
printg("gamma_kernel_rec(): kmalloc() failed when creating virtual GAMMA #%d\n",par_pid);
					return;
				}
			}
		}

		if (!procnum) {

DEBUG(printg(
"gamma_kernel_rec():  Virtual GAMMA %d (key %u):  this node not involved\n",
par_pid,p->vm.random_key
));

			free_pages((unsigned long)p,par_entry_magnitude);
			return;
		}

/*  This node is part of the virtual GAMMA being created  */
		p->num_locinst = procnum;
		par_proc_table[par_pid] = p;

{
DEBUG(
int k;
if (procnum == 1) {
printg("gamma_kernel_rec():  Virtual GAMMA %d (key %u): num_instances %d\n",
		par_pid,p->vm.random_key,p->vm.num_instances);
	printg("node list: \n");
	for (k=0; k<p->vm.num_instances; k++)
		printg("%d - ",p->vm.node_list[k]);
	printg("--\n");
	printg("phys_whoami %ld\n",phys_whoami);
	for (k=0; k<p->vm.num_instances; k++)
		printg(
			"%d: %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
			k,
			eth_table[p->vm.node_list[k]].addr[0],
			eth_table[p->vm.node_list[k]].addr[1],
			eth_table[p->vm.node_list[k]].addr[2],
			eth_table[p->vm.node_list[k]].addr[3],
			eth_table[p->vm.node_list[k]].addr[4],
			eth_table[p->vm.node_list[k]].addr[5]
		);
}
if (procnum > 0)
	for (k=1; k<=procnum; k++)
		printg("Local instance %d\n",p->inst_table[k-1].whoami);
);
}

		return;

	} /* switch */

	return;

}


#ifdef MORE_INSTANCES_ON_SAME_NODE
/************************************************************************
 *	Intra-node send/receive routines				*
 ************************************************************************/


/*
 *  Compute min shared ring room avail in a range of local process instances,
 *  all belonging to the given par_pid.
 *  If available room is larger than SHARED_RING_LOW, reserve one entry
 *  in each shared ring.
 */

static inline unsigned int
shared_ring_room
(unsigned long par_pid, unsigned int min, unsigned int max, int excl)
{
	int room,r,rc;
	par_entry *p;
	shared_ring_t *shared_ring;

	p = par_proc_table[par_pid];

	room=SHARED_RING_SIZE+1;
	for (r=min;r<=max;r++) {
		if (r == excl) continue;
		shared_ring = &p->inst_table[r].shared_ring;
		rc = atomic_dec_return(&shared_ring->room) + 1;
		if (rc < room) room = rc;
	}
	if (room > SHARED_RING_LOW)
		return room;
/*  Not enough room -- give the reserved room back  */
	for (r=min;r<=max;r++) {
		if (r == excl) continue;
		shared_ring = &p->inst_table[r].shared_ring;
		atomic_inc(&shared_ring->room);
	}
	return room;
}


static int gamma_send_2p_intranode
(
	unsigned long my_par_pid,
	unsigned long my_procnum,
	unsigned long out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2,
	unsigned long is_bcast
)
{

	port_entry			*out_p, *dest_p;
	unsigned long			dest_par_pid;
	unsigned long			min_procnum;
	unsigned long			max_procnum;
	long				excl_procnum; /* -1 == no exclusion */
	unsigned long			dest_port;
	unsigned long			len;
	unsigned long			cnk_len;
	long				dp;
	shared_ring_t			*shared_ring;
	shared_ring_entry		*desc;
	task_t				*receiver;
	unsigned int			dirty;
	unsigned long			num_locdest = 1;


	out_p = &par_proc_table[my_par_pid]->inst_table[my_procnum]
						.port_table[out_port];
	dest_par_pid = out_p->partner_par_pid;
	min_procnum = out_p->partner_procnum;
	max_procnum = out_p->partner_procnum;
	excl_procnum = -1;
	dest_port = out_p->partner_port;

	if (is_bcast) goto send_bcast;

DEBUG_E(printg(
"gamma_send_2p_intranode(%ld): enter: len1 %ld, len2 %ld, my_procnum %ld, num_locdest %ld\n",
my_procnum,len1,len2,my_procnum,num_locdest
));

send_bcast_back:

	if (shared_ring_room(dest_par_pid,min_procnum,max_procnum,excl_procnum)
	<= SHARED_RING_LOW)
		goto ring_full;

/*  First chunk.  */
	len = len1+len2;
	cnk_len = gmin(len,PAGE_SIZE);

/*  for all destination shared rings.  */
	for (dp = min_procnum; dp <= max_procnum; dp++) {

/*  omit myself if in destination group.  */
	if (dp == excl_procnum) continue;

	dest_p = &par_proc_table[dest_par_pid]->inst_table[dp]
						.port_table[dest_port];
	receiver = dest_p->task;
	if (receiver == NULL) goto no_partner;

	shared_ring = &par_proc_table[dest_par_pid]->inst_table[dp]
		.shared_ring;

/* next avail desc */
	dirty = (atomic_inc_return(&shared_ring->dirty) - 1)
		% SHARED_RING_SIZE;
	desc = &shared_ring->desc[dirty];

	desc->port = dest_port;
	desc->cnk_len = cnk_len;
	if (cnk_len < len)
		desc->bitmap = IS_1ST_CNK;
	else
		desc->bitmap = IS_1ST_CNK | IS_LAST_CNK;
	gamma_copy_from_user(desc->buf,data1,len1);

DEBUG(printg(
"gamma_send_2p_intranode(%ld): dp %ld, cnk_len %ld, head %u (%u), dirty %u (%u), room %d\n"
"par_pid %ld, procnum %ld, port %ld, len %ld\n",
my_procnum,dp,desc->cnk_len,
shared_ring->head,shared_ring->head % SHARED_RING_SIZE,
atomic_read(&shared_ring->dirty),dirty,
atomic_read(&shared_ring->room),
dest_par_pid,dp,desc->port,len
));

	gamma_copy_from_user((u8 *)(desc->buf)+len1,data2,cnk_len-len1);

/*  commit this chunk on current shared ring  */
/*
 * It is not sufficient to forward the tail:  In case of multiple senders
 * and a single receiver, this lock-free scheme may lead to a temporary
 * inconsistency between "tail" and "dirty".  Correct ordering is enforced
 * by setting (then testing, in the rx() thread) the desc's IN_USE flag.
 */
	mb();
	desc->bitmap |= IN_USE;
	atomic_inc(&shared_ring->tail);

/*  next shared ring...  */
	}

	len -= cnk_len;
	if (!len) {

DEBUG(printg("gamma_send_2p_intranode(%ld):  Done!\n",my_procnum));

		return 0;
	}

/*  Next chunks...  */
	cnk_len -= len1;

	for (;;) {

		if (shared_ring_room(dest_par_pid,min_procnum,max_procnum,
		excl_procnum) <= SHARED_RING_LOW)
			goto ring_full;

		data2 = (u8 *)data2+cnk_len;
		cnk_len = gmin(len,PAGE_SIZE);

/*  for all destination shared rings.  */
		for (dp = min_procnum; dp <= max_procnum; dp++) {

/*  omit myself if in destination group.  */
		if (dp == excl_procnum) continue;

		shared_ring = &par_proc_table[dest_par_pid]->inst_table[dp]
			.shared_ring;

		dirty = (atomic_inc_return(&shared_ring->dirty) - 1)
			% SHARED_RING_SIZE;
		desc = &shared_ring->desc[dirty];

		desc->port = dest_port;
		desc->cnk_len = cnk_len;
		if (cnk_len < len)
			desc->bitmap = 0;
		else
			desc->bitmap = IS_LAST_CNK;

DEBUG(printg(
"gamma_send_2p_intranode(%ld): dp %ld, cnk_len %ld, head %u (%u), dirty %u (%u), room %d\n"
"par_pid %ld, procnum %ld, port %ld, len %ld\n",
my_procnum,dp,desc->cnk_len,
shared_ring->head,shared_ring->head % SHARED_RING_SIZE,
atomic_read(&shared_ring->dirty),dirty,
atomic_read(&shared_ring->room),
dest_par_pid,dp,desc->port,len
));

		gamma_copy_from_user(desc->buf,data2,cnk_len);

/*  commit this chunk on current shared ring  */
		mb();
		desc->bitmap |= IN_USE;
		atomic_inc(&shared_ring->tail); /* enforces program order */

/*  next shared ring...  */
		}

		len -= cnk_len;
		if (!len) break;
/*  next chunk...  */

	}

DEBUG(printg("gamma_send_2p_intranode(%ld):  Done!\n",my_procnum));

	return 0;

send_bcast:

/*  How many receivers on this node?  */
	if (out_p->partner_par_pid == my_par_pid) {
/*  Sending to same group:  all members but sender.  */
		num_locdest = par_proc_table[my_par_pid]->num_locinst - 1;
		excl_procnum = my_procnum;
	}
	else {
/*  Sending to different group:  all members of dest group.  */
		num_locdest = par_proc_table[out_p->partner_par_pid]
				->num_locinst;
		excl_procnum = -1;
	}
	min_procnum = 0;
	max_procnum = par_proc_table[my_par_pid]->num_locinst - 1;

/*  Do we have local receivers actually?  */
	if (num_locdest) {

DEBUG(printg(
"gamma_send_2p_intranode(%ld): enter: bcast! num_locdest %ld, len1 %ld, len2 %ld\n",
my_procnum,num_locdest,len1,len2
));

		goto send_bcast_back;
	}
/*  No local receivers:  exit. */
	return 0;

ring_full:

DEBUG_L(printg(
"gamma_send_2p_intranode(%ld): pid %d:  no flow ctl --> ring full.  Failure.\n",
my_procnum,current->pid));

no_partner:

	return -GAMMA_ERR_GENERIC;

}


static int gamma_send_2p_flowctl_intranode
(
	unsigned long my_par_pid,
	unsigned long my_procnum,
	unsigned long out_port,
	void *data1, unsigned long len1,
	void *data2, unsigned long len2,
	unsigned long is_bcast,
	unsigned long was_susp
)
{

	port_entry			*out_p, *dest_p;
	unsigned long			dest_par_pid;
	unsigned long			min_procnum;
	unsigned long			max_procnum;
	long				excl_procnum; /* -1 == no exclusion */
	unsigned long			dest_port;
	unsigned long			len;
	unsigned long			cnk_len;
	long				dp;
	shared_ring_t			*shared_ring;
	shared_ring_entry		*desc;
	task_t				*receiver;
	unsigned int			dirty;
	unsigned long			num_locdest = 1;


	out_p = &par_proc_table[my_par_pid]->inst_table[my_procnum]
						.port_table[out_port];
	dest_par_pid = out_p->partner_par_pid;
	min_procnum = out_p->partner_procnum;
	max_procnum = out_p->partner_procnum;
	excl_procnum = -1;
	dest_port = out_p->partner_port;

	if (is_bcast) goto send_bcast;

DEBUG_E(printg(
"gamma_send_2p_flowctl_intranode(%ld): enter: len1 %ld, len2 %ld, my_procnum %ld, num_locdest %ld\n",
my_procnum,len1,len2,my_procnum,num_locdest
));

send_bcast_back:

	if (shared_ring_room(dest_par_pid,min_procnum,max_procnum,excl_procnum)
	<= SHARED_RING_LOW)
		goto ring_full;

/*  First chunk.  */
	len = len1+len2;
	cnk_len = gmin(len,PAGE_SIZE);

/*  for all destination shared rings.  */
	for (dp = min_procnum; dp <= max_procnum; dp++) {

/*  omit myself if in destination group.  */
	if (dp == excl_procnum) continue;

	dest_p = &par_proc_table[dest_par_pid]->inst_table[dp]
						.port_table[dest_port];
	receiver = dest_p->task;
	if (receiver == NULL) goto no_partner;

	shared_ring = &par_proc_table[dest_par_pid]->inst_table[dp]
		.shared_ring;

/* next avail desc */
	dirty = (atomic_inc_return(&shared_ring->dirty) - 1)
		% SHARED_RING_SIZE;
	desc = &shared_ring->desc[dirty];

	desc->port = dest_port;
	desc->cnk_len = cnk_len;
	if (was_susp) {
		if (cnk_len < len)
			desc->bitmap = 0;
		else
			desc->bitmap = IS_LAST_CNK;
	} else {
		if (cnk_len < len)
			desc->bitmap = IS_1ST_CNK;
		else
			desc->bitmap = IS_1ST_CNK | IS_LAST_CNK;
	}
	gamma_copy_from_user(desc->buf,data1,len1);

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): dp %ld, cnk_len %ld, head %u (%u), dirty %u (%u), room %d\n"
"par_pid %ld, procnum %ld, port %ld, len %ld\n",
my_procnum,dp,desc->cnk_len,
shared_ring->head,shared_ring->head % SHARED_RING_SIZE,
atomic_read(&shared_ring->dirty),dirty,
atomic_read(&shared_ring->room),
dest_par_pid,dp,desc->port,len
));

	gamma_copy_from_user((u8 *)(desc->buf)+len1,data2,cnk_len-len1);

/*  commit this chunk on current shared ring  */
	mb();
	desc->bitmap |= IN_USE;
	atomic_inc(&shared_ring->tail); /* enforces program order */

/*  next shared ring...  */
	}

	len -= cnk_len;
	if (!len) {

DEBUG(printg("gamma_send_2p_flowctl_intranode(%ld):  Done!\n",my_procnum));

		return 0;
	}

/*  Next chunks...  */
	cnk_len -= len1;

	for (;;) {

		if (shared_ring_room(dest_par_pid,min_procnum,max_procnum,
		excl_procnum) <= SHARED_RING_LOW)
			goto wait_room;

		data2 = (u8 *)data2+cnk_len;
		cnk_len = gmin(len,PAGE_SIZE);

/*  for all destination shared rings.  */
		for (dp = min_procnum; dp <= max_procnum; dp++) {

/*  omit myself if in destination group.  */
		if (dp == excl_procnum) continue;

		shared_ring = &par_proc_table[dest_par_pid]->inst_table[dp]
			.shared_ring;

		dirty = (atomic_inc_return(&shared_ring->dirty) - 1)
			% SHARED_RING_SIZE;
		desc = &shared_ring->desc[dirty];

		desc->port = dest_port;
		desc->cnk_len = cnk_len;
		if (cnk_len < len)
			desc->bitmap = 0;
		else
			desc->bitmap = IS_LAST_CNK;

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): dp %ld, cnk_len %ld, head %u (%u), dirty %u (%u), room %d\n"
"par_pid %ld, procnum %ld, port %ld, len %ld\n",
my_procnum,dp,desc->cnk_len,
shared_ring->head,shared_ring->head % SHARED_RING_SIZE,
atomic_read(&shared_ring->dirty),dirty,
atomic_read(&shared_ring->room),
dest_par_pid,dp,desc->port,len
));

		gamma_copy_from_user(desc->buf,data2,cnk_len);

/*  commit this chunk on current shared ring  */
		mb();
		desc->bitmap |= IN_USE;
		atomic_inc(&shared_ring->tail); /* enforces program order */

/*  next shared ring...  */
		}

		len -= cnk_len;
		if (!len) break;
/*  next chunk...  */
		continue;

wait_room:

		break;

	}

	if (!len) {
/*  Transmission complete.  Exit  */

DEBUG(printg("gamma_send_2p_flowctl_intranode(%ld):  Done!\n",my_procnum));

		return 0;
	}
/*
 *  Transmission suspended due to ring full.
 */
	if (is_bcast) {
/*
 *  Take note that the operation was suspended while bcasting in the
 *  same node.
 */
		out_p->bcast_was_susp = 1;

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): suspend bcast.\n",my_procnum
));

	}

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): No room.  Suspend (residual len %ld)\n",
my_procnum,len
));

/*  Return how many bytes are still to be transmitted.  */
	return len;

send_bcast:

	if (out_p->bcast_was_susp) {
/*  Bcast in the same node had been suspended due to ring full.  */
		out_p->bcast_was_susp = 0;

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): resume bcast.\n",my_procnum
));

	}

/*  How many receivers on this node?  */
	if (out_p->partner_par_pid == my_par_pid) {
/*  Sending to same group:  all members but sender.  */
		num_locdest = par_proc_table[my_par_pid]->num_locinst-1;
		excl_procnum = my_procnum;
	}
	else {
/*  Sending to different group:  all members of dest group.  */
		num_locdest = par_proc_table[out_p->partner_par_pid]
				->num_locinst;
		excl_procnum = -1;
	}
	min_procnum = 0;
	max_procnum = par_proc_table[my_par_pid]->num_locinst - 1;

/*  Do we have local receivers actually?  */
	if (num_locdest) {

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): enter: bcast! num_locdest %ld, len1 %ld, len2 %ld\n",
my_procnum,num_locdest,len1,len2
));

		goto send_bcast_back;
	}
/*  No local receivers:  exit. */
	return 0;

ring_full:

	if (is_bcast) {
/*
 *  Take note that the operation was suspended while bcasting in the
 *  same node.
 */
		out_p->bcast_was_susp = 1;

DEBUG(printg(
"gamma_send_2p_flowctl_intranode(%ld): suspend bcast.\n",my_procnum
));

	}

DEBUG_E(printg(
"gamma_send_2p_flowctl_intranode(%ld):  No room.  Try again!\n", my_procnum
));

	return -GAMMA_ERR_TRYAGAIN;

no_partner:

	return -GAMMA_ERR_GENERIC;

}


/*
 *  Returns:
 *
 *  P | IS_INTRA_NODE (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK | IS_INTRA_NODE if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_INCOMPLETE if no complete msg and no handler to run, or
 *  the invocation was bogus (no messages at all)
 *
 */
static int rx_gamma_intranode
(unsigned long par_pid, unsigned long procnum)
{

    shared_ring_t			*shared_ring;
    int					res;
    unsigned int			head;
    shared_ring_entry			*desc;
    unsigned long			port,cnk_len,bitmap,first_piece_size;
    port_entry				*pe;
    task_t				*receiver;
    void				*rx_buffer;


    res = -GAMMA_RX_INCOMPLETE;

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
    if ((par_proc_table[par_pid] == NULL)
        || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
        || (par_pid == 0)
    )
        goto bogus_inv;

    shared_ring = &par_proc_table[par_pid]->inst_table[procnum].shared_ring;

    if (shared_ring == NULL) goto bogus_inv;

    if (par_proc_table[par_pid]->inst_table[procnum].reenter)
        goto rec_reentering;

DEBUG_E(printg(
"rx_gamma_intranode(%ld), pid %u: entering...\n",procnum,current->pid
));

    head = shared_ring->head % SHARED_RING_SIZE;
    desc = &shared_ring->desc[head];
    if (!(desc->bitmap & IN_USE)) goto bogus_inv;

    for (;;) {

DEBUG(printg(
"rx_gamma_intranode(%ld): loop: head %u (%u), tail %u (%u), room %d, bitmap %ld\n",
procnum,shared_ring->head,head,atomic_read(&shared_ring->tail),
atomic_read(&shared_ring->tail) % SHARED_RING_SIZE,
atomic_read(&shared_ring->room),desc->bitmap
));

        port = desc->port;
        cnk_len = desc->cnk_len;
        bitmap = desc->bitmap;

DEBUG(printg(
"rx_gamma_intranode(%ld): par_pid %ld, procnum %ld, port %ld, cnk_len %ld\n",
procnum,par_pid,procnum,port,cnk_len
));

        pe = &par_proc_table[par_pid]->inst_table[procnum].port_table[port];

/*  Check whether port is in use as an input one.  */
        receiver = pe->task;
        if (receiver == NULL) goto discard_pkt;

/*  Empty messages:  no buffer, no frame numbering.  */
        if (!cnk_len) goto end_recv_intranode;

/*  Begin of payload transfer.  */

        pe->current_rx_msglen += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

        first_piece_size = 0;

        if (bitmap & IS_1ST_CNK) {
            first_piece_size = pe->boundary_1p;
            if ((first_piece_size > 0) && (first_piece_size <= cnk_len)) {
/*  Payload is to be scattered into two pieces.  Download 1st piece.  */

DEBUG(printg(
"rx_gamma_intranode(%ld):  Msg to be scattered, size of first piece is %ld\n",
procnum,first_piece_size
));

                get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);
                gamma_copy_to_user(rx_buffer,desc->buf,first_piece_size);

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "intra-node msg not complete but must run handler"
 */
                res = port | IS_FIRST_CHUNK | IS_INTRA_NODE;
                set_msglen(par_pid,procnum,first_piece_size);

/*  Will reenter later, to download remaining part of payload  */
		par_proc_table[par_pid]->inst_table[procnum].reenter = 1;

DEBUG(printg(
"[%ld] Quit to run rec handler for port %ld (1st piece)\n",procnum,port
));

                return res;
            }
            else
                first_piece_size = 0;
        }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
        update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

/*  Copy the payload (2nd piece) into user space  */
        recv_payload_2nd_chunk(rx_buffer,desc->buf,0,first_piece_size,cnk_len);

end_recv_intranode:

/*  Empty messages skip a lot of things.  */
        if (!desc->cnk_len) goto notif;

        if (!(bitmap & IS_LAST_CNK)) goto discard_pkt;

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
        switch (pe->buffer_kind) {
            case GO_BACK:
                reset_current_ptr(par_pid,port,procnum);
                break;
            case GO_AHEAD:
                check_room_possibly_reset_current_ptr(par_pid,port,procnum);
                break;
        }

notif:

/*
 *  Prepare to quit, in order to notify msg arrival and run the recv handler
 *  (if a handler is present and did not run already)
 */
        res = port;

DEBUG_E(printg(
"[%ld] Setting \"gamma_msglen\" to %d\n",procnum,
par_proc_table[par_pid]->inst_table[procnum].port_table[port].current_rx_msglen
));

        set_msglen(par_pid,procnum,pe->current_rx_msglen);

/*  Next msg starts from len = 0  */
        pe->current_rx_msglen = 0;

        desc->bitmap = IS_EMPTY;
        shared_ring->head++;
        atomic_inc(&shared_ring->room);
        break;

discard_pkt:

DEBUG_E(printg(
"rx_gamma_intranode(%ld): discarding pkt, head %u (%u), tail %u (%u)\n",
procnum,shared_ring->head,head,atomic_read(&shared_ring->tail),
atomic_read(&shared_ring->tail) % SHARED_RING_SIZE
));

        desc->bitmap = IS_EMPTY;
        shared_ring->head++;
        atomic_inc(&shared_ring->room);

/*  Quit if no more pkts available.  */
        head = shared_ring->head % SHARED_RING_SIZE;
        desc = &shared_ring->desc[head];
        if (!(desc->bitmap & IN_USE)) break;

DEBUG(printg("rx_gamma_intranode(%ld): continuing loop...\n",procnum));

    }

DEBUG(printg(
"rx_gamma_intranode(%ld): exiting, head %u (%u), tail %u (%u), room %d, res %d\n",
procnum,shared_ring->head,head,atomic_read(&shared_ring->tail),
atomic_read(&shared_ring->tail) % SHARED_RING_SIZE,
atomic_read(&shared_ring->room),res
));

    return res;

/*  Here all the unusual branches.  */

bogus_inv:

DEBUG_E(printg("rx_gamma_intranode(%ld): bogus invocation\n",procnum));

    return res;

rec_reentering:

/*  The receiver is reentering after running a rec handler (1sh piece).  */

DEBUG_E(printg("rx_gamma_intranode(%ld): reentering...\n",procnum));

    par_proc_table[par_pid]->inst_table[procnum].reenter = 0;
    head = shared_ring->head % SHARED_RING_SIZE;
    desc = &shared_ring->desc[head];

DEBUG(printg(
"rx_gamma_intranode(%ld): loop: head %u (%u), tail %u (%u), bitmap %ld\n",
procnum,shared_ring->head,head,atomic_read(&shared_ring->tail),
atomic_read(&shared_ring->tail) % SHARED_RING_SIZE,desc->bitmap
));

    port = desc->port;
    cnk_len = desc->cnk_len;
    bitmap = desc->bitmap;
    pe = &par_proc_table[par_pid]->inst_table[procnum].port_table[port];
    receiver = pe->task;
    first_piece_size = pe->boundary_1p;
    cnk_len -= first_piece_size;

DEBUG(printg(
"rx_gamma_intranode(%ld), reentering: par_pid %ld, procnum %ld, port %ld, cnk_len %ld\n",
procnum,par_pid,procnum,port,cnk_len
));

    goto get_remaining_payload;

}
#endif


/************************************************************************
 *	Blocking send routines, best-effort				*
 ************************************************************************/


asmlinkage long sys_gamma_send
(
	unsigned long out_port,
	unsigned long my_par_pid,
	unsigned long my_procnum,
	void *data,
	unsigned long len
)
{

	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void				*ddata;
	unsigned long			llen;
#endif
	l2s				*h;
	unsigned long			entry;
	unsigned long			cnk_len;
	unsigned long			frame_size;
	int				i, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif
#ifdef MEASURE_OUT_TH_MEM
	time_586			ima,imb;
#endif


	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_send(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata = data;
	llen = len;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
/*  copy the precomputed header into the tx buffer  */
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(h,
			cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

	for (;;) {
#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: la numerazione del frame e la stampa di debug sono effettuate
 * prima di controllare che il ring sia full, per semplificare il percorso
 * di esecuzione. Questo potrebbe causare dei problemi in caso di fallimenti (che
 * potrebbero essere anche solo di alcune schede...) */
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;
			
room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		len -= cnk_len;
		if (likely(!len)) break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
/*  copy previous header into new tx buffer  */
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif
		continue;

ring_full:

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif
	}

DEBUG(printg("sys_gamma_send(): Done!\n"));

	res = 0;
	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
							NULL,0,ddata,llen,1);
#endif
	return res;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST)
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
							NULL,0,data,len,0);
	else
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
							NULL,0,data,len,1);
	return res;
#endif

no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_send_flowctl
(
	unsigned long out_port_and_was_susp,
	unsigned long my_par_pid,
	unsigned long my_procnum,
	void *data,
	unsigned long len
)
{

	unsigned long			out_port;
	unsigned long			was_susp;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
	unsigned long			credit;
	unsigned long			curr_credit;
	l2s				*h;
	unsigned long			entry;
	unsigned long			cnk_len;
	unsigned long			frame_size;
	int				i, done, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif
#ifdef MEASURE_OUT_TH_MEM
time_586 ima,imb;
#endif


	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;

	out_port = out_port_and_was_susp & ~0x80000000;
	was_susp = out_port_and_was_susp & 0x80000000;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
	if (pe->bcast_was_susp) goto intra_node_bcast_susp;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_send_flowctl(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data;
		pe->current_tx_len1 = len;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;
	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK
						| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit)) {
			set_send_header_creq(h);
			dec_credit(phys_dest,credit - curr_credit);
			credit = curr_credit;

DEBUG_F({
 int i;
 printg("send frame_no %d to node %lu with credit req bit: curr credit %ld: ",
  win->tx_seq,phys_dest,curr_credit);
 for (i=0; i<total_num_nodes; i++)
  printg("(%d) ",atomic_read(&local_credit[i]));
 printg("\n");
})

		}

#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: la numerazione del frame e la stampa di debug sono effettuate
 * prima di controllare che il ring sia full, per semplificare il percorso
 * di esecuzione. Questo potrebbe causare dei problemi in caso di fallimenti (che
 * potrebbero essere anche solo di alcune schede...) */
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		curr_credit--;
		len -= cnk_len;
		if (!len) break;
		if (!curr_credit) goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(ima);
#endif
		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data,cnk_len);
#ifdef MEASURE_OUT_TH_MEM
gamma_time(imb);
out_counter_mem++;
out_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
out_total_amount_mem_bytes += cnk_len;
#endif

		continue;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done) goto credit_back;

DEBUG_F(printg(
"sys_gamma_send_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		spin_unlock(&gamma_tx_lock);
		if (!len) return -GAMMA_ERR_TRYAGAIN;
		return len;

ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg("sys_gamma_send_flowctl(): Done!\n"));

	res = 0;
	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_flowctl_intranode(
				my_par_pid,my_procnum,out_port,NULL,0,
				pe->current_tx_data1,pe->current_tx_len1,1,0);
#endif
	return res;

no_credit:

DEBUG_E(printg("sys_gamma_send_flowctl():  No credit.  Try again!\n"));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_TRYAGAIN;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_flowctl_intranode(
				my_par_pid,my_procnum,out_port,NULL,0,
				data,len,0,was_susp);
		return res;
	}

intra_node_bcast_susp:

	res = gamma_send_2p_flowctl_intranode(
			my_par_pid,my_procnum,out_port,NULL,0,
			data,len,1,was_susp);
	return res;
#endif

no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_send_2p
(
	unsigned long out_port_and_len1,
	unsigned long my_par_pid_and_procnum,
	void *data1,
	void *data2,
	unsigned long len2
)
{

	unsigned long			my_par_pid;
	unsigned long			my_procnum;
	unsigned long			out_port;
	unsigned long			len1;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void				*ddata2;
	unsigned long			llen1;
#endif
	l2s				*h;
	unsigned long			entry;
	unsigned long			len;
	unsigned long			frame_size;
	unsigned long			cnk_len;
	int				i, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	my_procnum = my_par_pid_and_procnum >> 16;
	out_port = out_port_and_len1>>8;
	len1 = (u8)out_port_and_len1;

	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_send_2p(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata2 = data2;
	llen1 = len1;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	len = len1 + len2;
	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE+len1,data2,cnk_len-len1);

	for (;;) {
#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: la numerazione del frame e la stampa di debug sono effettuate
 * prima di controllare che il ring sia full, per semplificare il percorso
 * di esecuzione. Questo potrebbe causare dei problemi in caso di fallimenti (che
 * potrebbero essere anche solo di alcune schede...) */
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		len -= cnk_len;
		if (!len) break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data2,cnk_len);

		continue;

ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif
	}

DEBUG(printg("sys_gamma_send_2p(): Done!\n"));

	res = 0;
	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
						data1,llen1,ddata2,len2,1);
#endif
	return res;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_2p(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST)
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
						data1,len1,data2,len2,0);
	else
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
						data1,len1,data2,len2,1);
	return res;
#endif

first_too_long:
no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_send_2p_flowctl
(
	unsigned long out_port_and_len1_and_was_susp,
	unsigned long my_par_pid_and_procnum,
	void *data2,
	unsigned long len2,
	void *data1
)
{

	unsigned long			my_par_pid;
	unsigned long			my_procnum;
	unsigned long			out_port;
	unsigned long			len1;
	unsigned long			was_susp;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
	unsigned long			credit;
	unsigned long			curr_credit;
	l2s				*h;
	unsigned long			entry;
	unsigned long			len;
	unsigned long			cnk_len;
	unsigned long			frame_size;
	int				i, done, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	int				j, nic;
#endif


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	my_procnum = my_par_pid_and_procnum >> 16;
	out_port = (out_port_and_len1_and_was_susp & ~0x80000000) >> 8;
	len1 = (u8)out_port_and_len1_and_was_susp;
	was_susp = out_port_and_len1_and_was_susp & 0x80000000;

	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	nic = pe->nic_idx;
	if (nic != BCAST_ALL_NICS) /* normal path, only one nic is used */
		tx_common_init(nic);
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
	if (pe->bcast_was_susp) goto intra_node_bcast_susp;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_send_2p_flowctl(): Entering\n"));

#ifdef MORE_NICS_SAMENODE_ROUTING
DEBUG_E(
if (nic != BCAST_ALL_NICS)
	printg("     Send to phys node %lu through NIC %d\n",phys_dest,nic);
else
	printg("     Broadcast. Send message through all NICs.\n");
);
#endif

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data1;
		pe->current_tx_len1 = len1;
		pe->current_tx_data2 = data2;
		pe->current_tx_len2 = len2;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;
	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	len = len1 + len2;
	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK
						| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE+len1,data2,cnk_len-len1);

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit)) {
			set_send_header_creq(h);
			dec_credit(phys_dest,credit - curr_credit);
			credit = curr_credit;

DEBUG_F({
 int i;
 printg("send frame_no %d to node %lu with credit req bit: curr credit %ld: ",
  win->tx_seq,phys_dest,curr_credit);
 for (i=0; i<total_num_nodes; i++)
  printg("(%d) ",atomic_read(&local_credit[i]));
 printg("\n");
})

		}

#ifdef MORE_NICS_SAMENODE_ROUTING
		if (unlikely(nic == BCAST_ALL_NICS)) {
/* broadcast path (used only when multiple nic support is enabled) */

#ifdef SEQCHECK
			set_send_header_frameno(h,win->tx_seq);
			win->tx_seq++;
#endif
/* luigi - warning: la numerazione del frame e la stampa di debug sono effettuate
 * prima di controllare che il ring sia full, per semplificare il percorso
 * di esecuzione. Questo potrebbe causare dei problemi in caso di fallimenti (che
 * potrebbero essere anche solo di alcune schede...) */
DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));
 
			for (j=0; j < num_gamma_devices; j++) {
				tx_common_init(j);
				
				if (tx_ring_is_full()) goto bc_ring_full;

			bc_room_in_tx_ring:
				
				tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
				tx_start();
				tx_forw_tail();
			}
		}
		else {
/* point-to-point path */
#endif
		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt1(h,HEADER_SIZE+cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef MORE_NICS_SAMENODE_ROUTING
		}
#endif

#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = HEADER_SIZE+cnk_len;

		curr_credit--;
		len -= cnk_len;
		if (!len) break;
		if (!curr_credit) goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		gamma_copy_from_user((u8 *)h+HEADER_SIZE,data2,cnk_len);

		continue;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done) goto credit_back;

DEBUG_F(printg(
"sys_gamma_send_2p_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

		spin_unlock(&gamma_tx_lock);
		if (!len) return -GAMMA_ERR_TRYAGAIN;
		return len;

ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

#ifdef MORE_NICS_SAMENODE_ROUTING
bc_ring_full:
		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto bc_room_in_tx_ring;
#endif

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg("sys_gamma_send_2p_flowctl(): Done!\n"));

	res = 0;
	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST)
		res = gamma_send_2p_flowctl_intranode(
				my_par_pid,my_procnum,out_port,
				pe->current_tx_data1,pe->current_tx_len1,
				pe->current_tx_data2,pe->current_tx_len2,1,0);
#endif
	return res;

no_credit:

DEBUG_E(printg("sys_gamma_send_2p_flowctl():  No credit.  Try again!\n"));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_TRYAGAIN;

tx_failure:

DEBUG_L(printg(
"sys_gamma_send_2p_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_flowctl_intranode(
				my_par_pid,my_procnum,out_port,data1,len1,
				data2,len2,0,was_susp);
		return res;
	}

intra_node_bcast_susp:

	res = gamma_send_2p_flowctl_intranode(
			my_par_pid,my_procnum,out_port,data1,len1,
			data2,len2,1,was_susp);
	return res;
#endif

first_too_long:
no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


/************************************************************************
 *	Non-blocking send routines, best-effort				*
 ************************************************************************/


asmlinkage long sys_gamma_isend
(
	unsigned long out_port,
	unsigned long my_par_pid,
	unsigned long my_procnum,
	void *data,
	unsigned long len
)
{

	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void				*ddata;
	unsigned long			llen;
#endif
	l2s				*h;
	unsigned long			entry;
	unsigned long			frame_size;
	unsigned long			cnk_len;
	void				*paddr;
	void				*next_paddr;
	int				i, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long			res_nic_idx;
#endif


	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];
#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;	
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_isend(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata = data;
	llen = len;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data);
	if (cnk_len >= PAGE_SIZE-((unsigned long)paddr & 0x00000fff))
		goto next_page1;
	next_paddr = (u8 *)paddr+cnk_len;

next_page_back1:

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

	for (;;) {

		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE,paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE;

		len -= cnk_len;
		if (!len) break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= PAGE_SIZE-((unsigned long)paddr & 0x00000fff))
			goto next_page;
		next_paddr = (u8 *)next_paddr+cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
		goto next_page_back;

ring_full:

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

DEBUG(printg("sys_gamma_isend(): Done!  isend_handle %d\n",tx_ring_tail));

	res = tx_ring_tail;
	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST) {
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
						NULL,0,ddata,llen,1);
		if (!res) res = tx_ring_tail;
	}
#endif

#ifdef MORE_NICS_SAMENODE_ROUTING
	/* assuming 26 bit is enough for tx_ring_tail */
	/* luigi warning: stiamo facendo l'OR anche quando res contiene un codice di errore? */
	/* luigi: l'impacchettamento qui sopra va portato su tutti i punti di uscita.
	 * per il caso broadcast non e' ancora sufficiente (per ogni handle bisogna 
	 * memorizzzare l'indice sui ring di tutte le NIC). */
	return res | res_nic_idx;
#else
	return res;
#endif

next_page1:

	cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
	goto next_page_back1;

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST)
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
							NULL,0,data,len,0);
	else
		res = gamma_send_2p_intranode(my_par_pid,my_procnum,out_port,
							NULL,0,data,len,1);
	if (!res) return -GAMMA_ISEND_IS_DONE;
	return res;
#endif

no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_isend_flowctl
(
	unsigned long out_port_and_was_susp,
	unsigned long my_par_pid,
	unsigned long my_procnum,
	void *data,
	unsigned long len
)
{

	unsigned long			out_port;
	unsigned long			was_susp;
	port_entry			*pe;
	unsigned long			phys_dest;
	tx_tmp_ring_t			*win;
	unsigned long			credit;
	unsigned long			curr_credit;
	l2s				*h;
	unsigned long			entry;
	unsigned long			frame_size;
	unsigned long			cnk_len;
	void				*paddr;
	void				*next_paddr;
	int				i, done, res;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long			res_nic_idx;
#endif


	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;

	out_port = out_port_and_was_susp & ~0x80000000;
	was_susp = out_port_and_was_susp & 0x80000000;
	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;
	/* luigi: l'impacchettamento qui sopra va portato su tutti i punti di uscita.
	 * per il caso broadcast non e' ancora sufficiente (per ogni handle bisogna 
	 * memorizzzare l'indice sui ring di tutte le NIC). */

#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
	if (pe->bcast_was_susp) goto intra_node_bcast_susp;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_isend_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data;
		pe->current_tx_len1 = len;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;
	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data);
	if (cnk_len >= (PAGE_SIZE-((unsigned long)paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = (u8 *)paddr+cnk_len;

next_page_back1:

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(h,
						cnk_len | MSG_1ST_FRAME_MASK
						| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit))
			set_send_header_creq(h);
			dec_credit(phys_dest,credit - curr_credit);
			credit = curr_credit;

		if (tx_ring_is_full()) goto ring_full;

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE,paddr,cnk_len);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE;

		curr_credit--;
		len -= cnk_len;
		if (!len) break;
		if (!curr_credit) goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data = (u8 *)data+cnk_len;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-((unsigned long)paddr & 0x00000fff)))
			goto next_page;
		next_paddr = (u8 *)next_paddr+cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
		goto next_page_back;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done) goto credit_back;

DEBUG_F(printg(
"sys_gamma_isend_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

/*  Return the total lenght of the not yet sent part of msg  */
		spin_unlock(&gamma_tx_lock);
		return -(GAMMA_ERR_MAX+len);

ring_full:

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg(
"sys_gamma_isend_flowctl(): Done!  isend_handle %d\n",tx_ring_tail
));

	res = tx_ring_tail;
	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest == BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_intranode(
				my_par_pid,my_procnum,out_port,NULL,0,
				pe->current_tx_data1,pe->current_tx_len1,1,0);
		if (!res)
			res = tx_ring_tail;
		else {
/*
 *  Store the inter-node isend handle: will be returned to the caller as soon
 *  as the intra-node Tx terminates.
 */
			pe->suspended_isend_handle = tx_ring_tail;
			if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
				res = -(GAMMA_ERR_MAX+len);
			else if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
				res = -(GAMMA_ERR_MAX+res);
		}
	}
#endif
	return res;

next_page1:

	cnk_len = PAGE_SIZE-((unsigned long)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data+cnk_len);
	goto next_page_back1;

no_credit:

DEBUG_E(printg("sys_gamma_isend_flowctl():  No credit.  Try again!\n"));

/*  Return the total lenght of the not yet sent part of msg  */
	spin_unlock(&gamma_tx_lock);
	return -(GAMMA_ERR_MAX+len);

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_intranode(
			my_par_pid,my_procnum,out_port,NULL,0,data,len,
			0,was_susp);
		if (!res)
			res = -GAMMA_ISEND_IS_DONE;
		else if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
			res = -(GAMMA_ERR_MAX+len);
		else if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
			res = -(GAMMA_ERR_MAX+res);
		return res;
	}

intra_node_bcast_susp:

/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_intranode(
		my_par_pid,my_procnum,out_port,NULL,0,data,len,
		1,was_susp);
	if (!res) {
/*
 *  The intra-node Tx has finished.
 *  Recover the inter-node isend handle, return it to the caller.
 */
		res = pe->suspended_isend_handle;
		pe->suspended_isend_handle = -GAMMA_ISEND_IS_DONE;
	} else if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		res = -(GAMMA_ERR_MAX+len);
	else if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		res = -(GAMMA_ERR_MAX+res);
	return res;
#endif

no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_isend_2p
(
	unsigned long out_port_and_len1,
	unsigned long my_par_pid_and_procnum,
	void *data1,
	void *data2,
	unsigned long len2
)
{

	unsigned long my_par_pid;
	unsigned long my_procnum;
	unsigned long out_port;
	unsigned long len1;
	port_entry *pe;
	unsigned long phys_dest;
	tx_tmp_ring_t *win;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	void *ddata2;
	unsigned long llen1;
#endif
	l2s *h;
	unsigned long entry;
	unsigned long len;
	unsigned long frame_size;
	unsigned long cnk_len;
	void *paddr;
	void *next_paddr;
	int i;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long res_nic_idx;
#endif
	int res;


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	my_procnum = my_par_pid_and_procnum >> 16;
	out_port = out_port_and_len1 >> 8;
	len1 = (u8)out_port_and_len1;

	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;
	/* luigi: l'impacchettamento qui sopra va portato su tutti i punti di uscita.
	 * per il caso broadcast non e' ancora sufficiente (per ogni handle bisogna 
	 * memorizzzare l'indice sui ring di tutte le NIC). */
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_isend_2p(): Entering\n"));

#ifdef MORE_INSTANCES_ON_SAME_NODE
	ddata2 = data2;
	llen1 = len1;
#endif

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	len = len1+len2;
	frame_size = tx_optimal_frag(len1+len2);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data2);
	if (cnk_len >= len1+(PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len-len1;

next_page_back1:

	if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len | MSG_1ST_FRAME_MASK);
	else
		fill_send_header_cnklen(
			h, cnk_len | MSG_1ST_FRAME_MASK | MSG_COMPLETE_MASK);

	for (;;) {

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE+len1,paddr,cnk_len-len1);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE+len1;

		len -= cnk_len;
		if (!len)
			break;

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len);
		goto next_page_back;

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

DEBUG(printg("sys_gamma_isend_2p(): Done!  isend_handle %d\n",tx_ring_tail));

	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST)
		return tx_ring_tail;
	res = gamma_send_2p_intranode(
		my_par_pid,my_procnum,out_port,data1,llen1,ddata2,len2,1
	);
	if (!res)
		res = tx_ring_tail;
#else
	res = tx_ring_tail;
#endif
	return res;

next_page1:

	cnk_len = len1+PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len-len1);
	goto next_page_back1;

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_2p(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST) {
		res = gamma_send_2p_intranode(
			my_par_pid,my_procnum,out_port,data1,len1,data2,len2,0
		);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		return res;
	}
	res = gamma_send_2p_intranode(
		my_par_pid,my_procnum,out_port,data1,len1,data2,len2,1
	);
	if (!res)
		return -GAMMA_ISEND_IS_DONE;
	return res;
#endif

first_too_long:
no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_isend_2p_flowctl
(
	unsigned long out_port_and_len1_and_was_susp,
	unsigned long my_par_pid_and_procnum,
	void *data2,
	unsigned long len2,
	void *data1
)
{

	unsigned long my_par_pid;
	unsigned long my_procnum;
	unsigned long out_port;
	unsigned long len1;
	unsigned long len;
	unsigned long was_susp;
        port_entry *pe;
	unsigned long phys_dest;
	tx_tmp_ring_t *win;
	unsigned long credit;
	unsigned long curr_credit;
	l2s *h;
	unsigned long entry;
	unsigned long frame_size;
	unsigned long cnk_len;
	void *paddr;
	void *next_paddr;
	int i, done;
	tx_common_decl;
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned long res_nic_idx;
#endif
#ifdef MORE_INSTANCES_ON_SAME_NODE
	int res;
#endif


	my_par_pid = (unsigned short)my_par_pid_and_procnum;
	my_procnum = my_par_pid_and_procnum >> 16;
	out_port = (out_port_and_len1_and_was_susp & ~0x80000000) >> 8;
	len1 = (u8)out_port_and_len1_and_was_susp;
	len = len1+len2;
	was_susp = out_port_and_len1_and_was_susp & 0x80000000;

	if (par_proc_table[my_par_pid] == NULL) goto no_par_pid;
	if (par_proc_table[my_par_pid]->inst_table[my_procnum].pid < 0)
		goto no_par_pid;
	if (len1 > MAX_1ST_PIECE_SIZE) goto first_too_long;

	pe = &par_proc_table[my_par_pid]->inst_table[my_procnum]
							.port_table[out_port];

#ifdef MORE_NICS_SAMENODE_ROUTING
	res_nic_idx = pe->nic_idx;
	tx_common_init(res_nic_idx);
	/* nic_idx is at most 5 bit long */
	res_nic_idx = res_nic_idx << 26;
	/* luigi: l'impacchettamento qui sopra va portato su tutti i punti di uscita.
	 * per il caso broadcast non e' ancora sufficiente (per ogni handle bisogna 
	 * memorizzzare l'indice sui ring di tutte le NIC). */
#else
	tx_common_init(0);
#endif

	phys_dest = pe->phys_partner_node;

	if (phys_dest != BROADCAST)
		win = &tx_tmp_ring[phys_dest];
	else
		win = &tx_tmp_ring[phys_whoami];

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (pe->partner_is_intra_node) goto intra_node;
	if (pe->bcast_was_susp) goto intra_node_bcast_susp;
#endif

	if (!spin_trylock(&gamma_tx_lock)) goto locked;

DEBUG_E(printg("sys_gamma_isend_2p_flowctl(): Entering\n"));

	if ((phys_dest == BROADCAST) && !was_susp) {
/*
 *  Take note of addr and size of message, for resumption after possible
 *  suspension (shortage of credit).
 */
		pe->current_tx_data1 = data1;
		pe->current_tx_len1 = len1;
		pe->current_tx_data2 = data2;
		pe->current_tx_len2 = len2;
	}

	read_credit(phys_dest,credit);
	curr_credit = credit;

	len = len1+len2;

	if (!curr_credit) goto no_credit;

	entry = win->tail;
	incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
	h = win->buf[entry].b;
	memcpy(h,pe->header,HEADER_SIZE);

	gamma_copy_from_user((u8 *)h+HEADER_SIZE,data1,len1);

	frame_size = tx_optimal_frag(len);

	cnk_len = gmin(len,frame_size-HEADER_SIZE);

	paddr = user_space_to_phys(current,data2);
	if (cnk_len >= len1+(PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
		goto next_page1;
	next_paddr = paddr+cnk_len-len1;

next_page_back1:

	if (!was_susp) {
		if (cnk_len < len)
			fill_send_header_cnklen(
				h, cnk_len | MSG_1ST_FRAME_MASK);
		else
			fill_send_header_cnklen(
				h,
				cnk_len | MSG_1ST_FRAME_MASK
				| MSG_COMPLETE_MASK);
	} else if (cnk_len < len)
		fill_send_header_cnklen(h, cnk_len);
	else
		fill_send_header_cnklen(h, cnk_len | MSG_COMPLETE_MASK);

	for (;;) {

		if (is_low_watermark(phys_dest,credit,curr_credit))
			set_send_header_creq(h);
			dec_credit(phys_dest,credit - curr_credit);
			credit = curr_credit;

		if (!tx_ring_is_full()) {

room_in_tx_ring:

#ifdef SEQCHECK
		set_send_header_frameno(h,win->tx_seq);
		win->tx_seq++;
#endif

DEBUG(printg(
"sending through port %ld, frame_no %d\n",out_port,read_send_header_frameno(h)
));

		tx_enq_pkt(h,HEADER_SIZE+len1,paddr,cnk_len-len1);
		tx_start();
		tx_forw_tail();
#ifdef SAFE_TX
		if (tx_ring_occupancy >= STEPS_THEN_STOP)
			tx_flush_ring(tx_ring_tail,MAX_FLUSH_WAIT,i);
#endif
		win->buf[entry].l = -(1 + tx_ring_tail);
		win->buf[entry].hlen = HEADER_SIZE+len1;

		curr_credit--;
		len -= cnk_len;
		if (!len)
			break;

		if (!curr_credit)
			goto credit_exhausted;

credit_back:

/*  Prepare next frame  */
		entry = win->tail;
		incr_modulo(win->tail,GAMMA_TX_WIN_SIZE);
		memcpy(win->buf[entry].b,h,HEADER_SIZE);
		h = win->buf[entry].b;
		data2 = (u8 *)data2+cnk_len-len1;
		len1 = 0;
		cnk_len = gmin(len,frame_size - HEADER_SIZE);
		paddr = next_paddr;
		if (cnk_len >= (PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff)))
			goto next_page;
		next_paddr += cnk_len;

next_page_back:

		if (cnk_len < len)
			fill_send_header_cnklen(h,cnk_len);
		else
			fill_send_header_cnklen(h,cnk_len | MSG_COMPLETE_MASK);

		continue;

next_page:

		cnk_len = PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
		next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len);
		goto next_page_back;

credit_exhausted:

		dec_credit(phys_dest,credit);
		refill_credit(done,phys_dest,credit,curr_credit);
		if (done)
			goto credit_back;

DEBUG_F(printg(
"sys_gamma_isend_2p_flowctl():  No credit.  Suspend (residual len %ld)\n",len
));

/*  Return the total lenght of the not yet sent part of msg  */
		spin_unlock(&gamma_tx_lock);
		return -(GAMMA_ERR_MAX+len);

		}

		tx_wait_ring_room(MAX_FLUSH_WAIT,i);
		if (!i) goto tx_failure;
		goto room_in_tx_ring;

	}

	dec_credit(phys_dest,credit-curr_credit);

DEBUG(printg(
"sys_gamma_isend_2p_flowctl(): Done!  isend_handle %d\n",tx_ring_tail
));

	spin_unlock(&gamma_tx_lock);
#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (phys_dest != BROADCAST)
		return tx_ring_tail;
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_intranode(
		my_par_pid,my_procnum,out_port,
		pe->current_tx_data1,pe->current_tx_len1,
		pe->current_tx_data2,pe->current_tx_len2,1,0);
	if (!res)
		return tx_ring_tail;
/*
 *  Store the inter-node isend handle: will be returned to the caller as soon
 *  as the intra-node Tx terminates.
 */
	pe->suspended_isend_handle = tx_ring_tail;

	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;
#else
	return tx_ring_tail;
#endif

next_page1:

	cnk_len = len1+PAGE_SIZE-((dma_addr_t)paddr & 0x00000fff);
	next_paddr = user_space_to_phys(current,(u8 *)data2+cnk_len-len1);
	goto next_page_back1;

no_credit:

DEBUG_E(printg("sys_gamma_isend_2p_flowctl():  No credit.  Try again!\n"));

/*  Return the total lenght of the not yet sent part of msg  */
	spin_unlock(&gamma_tx_lock);
	return -(GAMMA_ERR_MAX+len);

tx_failure:

DEBUG_L(printg(
"sys_gamma_isend_2p_flowctl(): par_pid %ld, out_port %ld, frame_no %d: Failure (ring pos %d)\n",
my_par_pid,out_port,rec_header_frame_no(h),tx_ring_head
));

	dec_credit(phys_dest,credit-curr_credit);

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
intra_node:

	if (phys_dest != BROADCAST) {
/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
		res = gamma_send_2p_flowctl_intranode(
			my_par_pid,my_procnum,out_port,data1,len1,data2,len2,0,
			was_susp
		);
		if (!res)
			return -GAMMA_ISEND_IS_DONE;
		if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+len);
		if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
			return -(GAMMA_ERR_MAX+res);
		return res;
	}

intra_node_bcast_susp:

/*
 *  Here I would like a non-blocking intra-node send(), but this would need
 *  additional hardware...
 */
	res = gamma_send_2p_flowctl_intranode(
		my_par_pid,my_procnum,out_port,data1,len1,data2,len2,1,
		was_susp);
	if (!res) {
/*
 *  The intra-node Tx has finished.
 *  Recover the inter-node isend handle, return it to the caller.
 */
		res = pe->suspended_isend_handle;
		pe->suspended_isend_handle = -GAMMA_ISEND_IS_DONE;
		return res;
	}
	if (res == -GAMMA_ERR_TRYAGAIN)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+len);
	if (res > 0)
/*  Return the total lenght of the not yet sent part of msg  */
		return -(GAMMA_ERR_MAX+res);
	return res;
#endif

first_too_long:
no_par_pid:

	return -GAMMA_ERR_GENERIC;

locked:

	return -GAMMA_ERR_TRYAGAIN;

}


asmlinkage long sys_gamma_wsend
(unsigned long isend_handle)
{

	int i;
	tx_common_decl;


DEBUG(printg("sys_gamma_wsend(): Entering\n"));

	spin_lock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	tx_common_init( routing_table[ (isend_handle & 0x7c000000) >> 26 ] );
	isend_handle &= 0x3ffffff;
#else
	tx_common_init(0);
#endif

	if (isend_handle > tx_ring_tail)
		goto ring_flushed;
	if ((tx_ring_head < tx_ring_tail) && (isend_handle <= tx_ring_head))
		goto ring_flushed;

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG(printg("sys_gamma_wsend(): Done\n"));

		spin_unlock(&gamma_tx_lock);
		return 0;
	}

#if 0
	tx_start();

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG_L(printg("sys_gamma_wsend(): Done, but had to wake up the NIC\n"));

		spin_unlock(&gamma_tx_lock);
		return 0;
	}
#endif

DEBUG_L(printg(
"sys_gamma_wsend(): Failure (isend_handle %ld, entry %d)\n",
isend_handle,tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

ring_flushed:

DEBUG(printg("sys_gamma_wsend(): Done (tx ring flushed)\n"));

	spin_unlock(&gamma_tx_lock);
	return 0;

}


asmlinkage long sys_gamma_tsend
(unsigned long isend_handle)
{

	int i;
	tx_common_decl;


DEBUG(printg("sys_gamma_tsend(): Entering\n"));

	spin_lock(&gamma_tx_lock);

#ifdef MORE_NICS_SAMENODE_ROUTING
	tx_common_init( routing_table[ (isend_handle & 0x7c000000) >> 26 ] );
	isend_handle &= 0x3ffffff;
#else
	tx_common_init(0);
#endif

	if (isend_handle > tx_ring_tail)
		goto ring_flushed;
	if ((tx_ring_head < tx_ring_tail) && (isend_handle <= tx_ring_head))
		goto ring_flushed;

/* peppe da fare (test invece di flush) */
	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG(printg("sys_gamma_tsend(): Done\n"));

		spin_unlock(&gamma_tx_lock);
		return 1;
	}

#if 0
	tx_start();

	tx_flush_ring(isend_handle,MAX_FLUSH_WAIT,i);

	if (i) {

DEBUG_L(printg("sys_gamma_tsend(): Done, but had to wake up the NIC\n"));

		spin_unlock(&gamma_tx_lock);
		return 1;
	}
#endif

DEBUG_L(printg(
"sys_gamma_tsend(): Failure (isend_handle %ld, entry %d)\n",
isend_handle,tx_ring_head
));

	spin_unlock(&gamma_tx_lock);
	return -GAMMA_ERR_SENDFAILURE;

ring_flushed:

DEBUG(printg("sys_gamma_tsend(): Done (tx ring flushed)\n"));

	spin_unlock(&gamma_tx_lock);
	return 1;

}


/************************************************************************
 *	Receive routines						*
 ************************************************************************/


/*
 *  Returns:
 *  P (0 <= P <= MAX_PORT_NUMBER) if msg complete for port P
 *  (user-level caller will perform notification on sem #P and run rec handler
 *  if present; msg lenght is in user-level variable "gamma_msglen")
 *
 *  P | IS_FIRST_CHUNK if msg not complete for port P but
 *  rec handler must run to manage the arrival of 1st piece of msg
 *  (user-level caller will run the handler, 1st piece lenght is in user-level
 *  variable "gamma_msglen")
 *
 *  -GAMMA_RX_NODATA if no data pkt arrived (no messages at all; only
 *  GAMMA_KERNEL pkts arrived)
 *
 *  -GAMMA_RX_INCOMPLETE if data pkts arrived but no complete msg and no
 *  handler to run
 *
 *  -GAMMA_ERR_MUSTYIELD if there is a pending pkt for a process other than the
 *  invoker, or another process is expected to re-enter this routine soon.
 *
 */
int rx_gamma
(unsigned int dev_idx)
{

struct net_device *dev;
l2s *h;
unsigned long cnk_len;
unsigned long frame_no;
unsigned long phys_sender;
unsigned long par_pid;
unsigned long port;
unsigned long procnum;
port_entry *pe;
task_t *receiver;
int res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long bcast_sequence;
#endif
unsigned long first_piece_size;
void *rx_buffer;
#ifdef MORE_INSTANCES_ON_SAME_NODE
unsigned long bcast_phys_sender = 0;
unsigned long head = 0;
unsigned long bcast_frame_no = 0;
#endif
#ifdef SEQCHECK
u32 err_msg[4];
#endif
rx_common_decl;
#ifdef MEASURE_IN_TH
unsigned long cumul_len;
time_586 a,b;
#endif
#ifdef MEASURE_IN_TH_MEM
time_586 ima,imb;
#endif


dev = gamma_devices[dev_idx];

#ifdef MEASURE_IN_TH
cumul_len = 0;
gamma_time(a);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
bcast_sequence = 0;
#endif

if (gamma_run_handler.task != NULL) goto rec_reentering;

rx_common_init(dev);

rx_read_pkt_info();

res = -GAMMA_RX_NODATA;

if (rx_err_pktnotavail()) goto bogus_inv;

for (;;) {

DEBUG_E(printg(
"rx_gamma(): loop: rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

 rx_header(h);

DEBUG_E(printg(
"rx_gamma(): %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll, h[6].ll, h[7].ll, h[8].ll
));

/*  Interpret the header  */

 cnk_len = rec_header_cnk_len(h);
 frame_no = rec_header_frame_no(h);
 phys_sender = rec_header_phys_sender(h);
 rec_header_parpid_port(h,par_pid,port,procnum);

 if (rec_header_frmtype(h) != ETH_GAMMA) goto pkt_is_not_gamma;

DEBUG(printg(
"rx_gamma(), pid %u: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h) != 0, par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

#ifdef SEQCHECK
/****************************************************************************/
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender]) goto end_rx;
/****************************************************************************/
#endif

accept_pkt:

/*  At least one data pkt arrived.  */
res = -GAMMA_RX_INCOMPLETE;

#ifdef MEASURE_IN_TH
 cumul_len += cnk_len;
#endif

#ifdef SEQCHECK
/****************************************************************************/
/*  Check for correct frame sequence numbering.  */
 if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
  if (frame_no != rx_seq_bcast[phys_sender]) goto err_frame_no;
  rx_seq_bcast[phys_sender]++;
 } else {
  if (frame_no != rx_seq[phys_sender]) goto err_frame_no;
  rx_seq[phys_sender]++;
 }
/****************************************************************************/
#endif

/*  GAMMA pkts not allowed if parallel job not started  */
 if (!gamma_susp_irq_count) goto not_assigned;

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
 if (
  (par_proc_table[par_pid] == NULL)
  || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
 ) {

DEBUG_E(printg("rx_gamma(): par_pid not existing or not assigned. par_pid %s: %d\n",
	       par_proc_table[par_pid] == NULL? "not exists" : "is",
	       par_proc_table[par_pid] == NULL? 0 : par_proc_table[par_pid]->inst_table[procnum].pid));

	 goto not_assigned;
 }

 pe = &par_proc_table[par_pid]->inst_table[procnum].port_table[port];

/*  Check whether port is in use as an input one.  */
 receiver = pe->task;
 if (receiver == NULL) {

DEBUG_E(printg("rx_gamma(): port not in use\n"));

	 goto not_assigned;
 }

/*  Check whether receiver is (thread-equivalent to) the caller.  */
 if (receiver->mm != current->mm) goto cannot_deliver_yet;

/*  Flow Control:  manage the request for credit renewal, if present.  */
 if (rec_header_creq(h)) goto credit_request;

credit_request_back:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/* Check whether a seq of bcast pkts is beginning or has ended */
 if (rec_is_bcast(h) && (!bcast_sequence)) goto found_bcast_pkt;
 if (bcast_sequence &&
  ((!rec_is_bcast(h)) || (phys_sender != bcast_phys_sender))) {
/*
 *  A seq. of bcast pkts has ended.
 *  Quit the routine, to allow other receivers to get the bcast data
 *  before considering the current pkt.
 */

DEBUG(printg("rx_gamma(): end of bcast pkt seq\n"));

/*
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
  goto exit;
 }

found_bcast_pkt_back:
#endif

/*  Frame is OK and we get it now.  */

/*  Empty messages:  nothing to download.  */
 if (!cnk_len) goto notif;

/*  Download the payload.  */

 pe->current_rx_msglen += cnk_len;

/*
 *  Is this packet to be scattered into two pieces?
 *  If present, 1st piece is carried by 1st packet of the msg.
 */

 first_piece_size = 0;

 if (rec_header_1st_frame(h)) {
  first_piece_size = pe->boundary_1p;
  if ((first_piece_size > 0) && (first_piece_size <= cnk_len)) {
/*  Payload is to be scattered into two pieces.  */

DEBUG(printg(
"rx_gamma():  Msg to be scattered, size of first piece is %ld\n",
first_piece_size
));

   get_ptr_to_user_recv_buffer(par_pid,port,procnum,rx_buffer);

DEBUG_E(printg("rx_gamma():  Copying 1st chunk to addr %p\n",rx_buffer));

   recv_payload_1st_chunk(rx_buffer,h,HEADER_SIZE,first_piece_size);
   cnk_len -= first_piece_size;

/*
 *  Prepare to quit, to run the receiver handler of the port.
 *  Since this message is going to be scattered, we can assume
 *  that there is a handler to run.
 */

/*
 *  Tell the caller which handler to run, and len of 1st chunk of data:
 *  "inter-node msg not complete but must run handler"
 */
   res = port | IS_FIRST_CHUNK;

   set_msglen(par_pid,procnum,first_piece_size);

/*
 *  Current process only is allowed to reenter later,
 *  to download remaining part of payload.
 */
   gamma_run_handler.task = current;
   gamma_run_handler.dev = dev;

DEBUG(printg("Quit to run rec handler for port %ld (1st piece)\n",port));

   return res;

  } else
   first_piece_size = 0;
 }

get_remaining_payload:

/*
 *  Shift user-space rec buffer's current ptr forward.
 *  Check for room availability.
 *  If not enough room in rec buffer, truncate the msg chunk
 *  by setting cnk_len equal to the available room.
 *  Decrement the available room by cnk_len.
 *  The current msg chunk will be stored starting
 *  from address "rx_buffer".
 */
 update_ptr_to_user_recv_buffer(par_pid,port,procnum,cnk_len,rx_buffer);

#ifdef MEASURE_IN_TH_MEM
gamma_time(ima);
#endif

/*  Copy the payload (2nd piece) into user space  */
#ifdef USE_DAMNICK

DEBUG_E(printg(
"rx_gamma():  Start copying payload to addr %p through DAMNICK\n",
rx_buffer)
);

 if (!rec_header_msg_complete(h))
  damnick_icopy_k2u(
   rx_buffer,
   (void *)((u8 *)h+HEADER_SIZE+first_piece_size),
   cnk_len
  );
 else {
  unsigned long pre_notif_val;
  pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
  damnick_icopy_k2u_nu(
   rx_buffer,
   (void *)((u8 *)h+HEADER_SIZE+first_piece_size),
   cnk_len,
   (unsigned long *)(par_proc_table[par_pid]->inst_table[procnum].sem1_table)
   + pe->semaphore,
   pre_notif_val
  );
 }
#else

DEBUG_E(
printg("rx_gamma():  Copying payload to addr %p, cnk_len %lu\n",
rx_buffer,cnk_len)
);

 recv_payload_2nd_chunk(rx_buffer,h,HEADER_SIZE,first_piece_size,cnk_len);
#endif

#ifdef MEASURE_IN_TH_MEM
gamma_time(imb);
in_counter_mem++;
in_total_amount_mem_delay += (unsigned long)gamma_time_diff_ticks(imb,ima);
in_total_amount_mem_bytes += cnk_len;
#endif

 if (!rec_header_msg_complete(h)) goto end_rx;

DEBUG(printg("Message complete\n"));

/*
 *  Do one of the following, depending on the kind of buffer attached
 *  to the recv port:
 *  case GO_BACK:
 *  reset the current position ptr of the recv buffer to the begin of
 *  the recv buffer itself.  Next msg will be stored at the begin
 *  of the buffer.
 *  case GO_AHEAD:
 *  If there is still room available in the attached buffer,
 *  leave the current position ptr untouched.
 *  Otherwise, reset the current position ptr of the recv buffer to
 *  the begin of the recv buffer itself; in this case, next msg will
 *  be stored at the begin of the buffer.
 */
 switch (pe->buffer_kind) {
  case GO_BACK:
   reset_current_ptr(par_pid,port,procnum);
   break;
  case GO_AHEAD:
   check_room_possibly_reset_current_ptr(par_pid,port,procnum);
   break;
 }

#ifdef USE_DAMNICK
/*
 *  Skip CPU-operated notification of arrival: it will be carried out by the
 *  DAMNICK device, because of the invocation of damnick_icopy_k2u_nu() .
 */
 goto set_msglen;
#endif

notif:

#ifdef USE_DAMNICK

DEBUG(printg("Notifying arrival at port %ld\n",port));

 {
 unsigned long pre_notif_val;
 pre_notify_arrival(par_pid,port,procnum,pre_notif_val);
 damnick_writel_u(
  (unsigned long *)
  (par_proc_table[par_pid]->inst_table[procnum].sem1_table) + pe->semaphore,
  pre_notif_val
 );
 }

set_msglen:

#endif

/*
 *  Prepare to quit, to notify msg arrival and run the receiver handler
 *  (if a handler is present and it did not run already)
 */
 res = port;

DEBUG_E(printg("Setting \"gamma_msglen\" to %d\n", pe->current_rx_msglen));

 set_msglen(par_pid,procnum,pe->current_rx_msglen);

/*  Next msg starts from len = 0  */
 pe->current_rx_msglen = 0;

/*
 *  Tell the NIC this packet has been consumed, then quit rx_gamma()
 *  because a new message is now complete.
 *  Bcast pkts cannot be removed from Rx ring until all other receivers have
 *  read them.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (bcast_sequence) goto quit_rx_loop_bcastseq;
#endif
 discard_pkt();
 break;

#ifdef MORE_INSTANCES_ON_SAME_NODE
quit_rx_loop_bcastseq:

 rec_header_set_procnum(h,procnum+1);
 forward_rx_ring_head();
/*
 *  NOTE: variable "gamma_must_yield", indicating which proc instance should
 *  run next in the bcast recv, has been set up when the bcast seq begun.
 */
 break;
#endif

end_rx:

/*
 *  Tell the NIC this packet has been consumed.
 *  Bcast pkts cannot be removed from Rx ring until all other receivers have
 *  read them.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (!bcast_sequence)
  discard_pkt();
 else {
  rec_header_set_procnum(h,procnum+1);
  forward_rx_ring_head();
 }
#else
 discard_pkt();
#endif

end_rx_with_updated_ptrs:

 rx_read_pkt_info();

/*  If no further pending frames, quit rx_gamma()  */
 if (rx_err_pktnotavail()) break;

/*  There is at least another pending frame.  */

DEBUG_E(printg("rx_gamma(): continuing loop...\n"));

}

exit:

#ifdef MEASURE_IN_TH
gamma_time(b);
#ifdef MEASURE_IN_TH_PRINTALL
printg("in: %ld usec, %ld bytes, %ld MByte/s\n",
	((unsigned long)gamma_time_diff_ticks(b,a))/CLOCK,cumul_len,
	(cumul_len*CLOCK)/(unsigned long)gamma_time_diff_ticks(b,a)
);
#endif
in_counter++;
in_total_amount += (cumul_len*CLOCK)/(unsigned long)gamma_time_diff_ticks(b,a);
#endif

#ifdef MORE_INSTANCES_ON_SAME_NODE
if (bcast_sequence) goto quit_rx_gamma_bcastseq;
#endif

/*  Quit.  Possibly yield the CPU.  */
if (gamma_must_yield.task != NULL && res < 0)
 res = -GAMMA_ERR_MUSTYIELD;

#ifdef MORE_INSTANCES_ON_SAME_NODE
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d (%ld), rx_ring_tail %d, res %d\n",
 rx_ring_head,head,rx_ring_tail,res
));
#else
DEBUG_E(printg(
"rx_gamma(): exiting: rx_ring_head %d, rx_ring_tail %d, res %d\n",
 rx_ring_head,rx_ring_tail,res
));
#endif

return res;

#ifdef MORE_INSTANCES_ON_SAME_NODE
quit_rx_gamma_bcastseq:

/*
 *  Next local receiver must re-read bcast pkts.
 *  Set ring head back, and roll-back frame numbering sequence.
 */
write_rx_ring_head(head);
rx_seq_bcast[bcast_phys_sender] = bcast_frame_no;
/*
 *  Quit.  Yield the CPU.
 *  "gamma_must_yield" already set to next bcast receiver.
 */
if (res < 0) res = -GAMMA_ERR_MUSTYIELD;

DEBUG_E(printg(
"rx_gamma(): exiting (bcast): rx_ring_head %d (%ld), rx_ring_tail %d, res %d\n",
 rx_ring_head,head,rx_ring_tail,res
));

return res;
#endif


/*  Here all the unusual branches.  */

#ifdef SEQCHECK
err_frame_no:

/*
 *  There are two possibilities, namely:
 *  a)  A previous frame was lost.
 *      In this case, a retransmission must be asked to sender.
 *  b)  This frame is a useless duplicate of a previous one,
 *      maybe caused by a late collision on a shared LAN.
 *      In this case the previous one should have been discarded
 *      and replaced by this one, but this is no longer possible
 *      (a receiver handler could have been issued).  Since no
 *      error was detected on the previous frame, we hope that
 *      everything is OK and this duplicate may be discarded.
 *      However late collisions should never occur if the LAN is
 *      properly cabled.
 */

if (!rec_is_bcast(h)) {
 if (frame_no == (u8)(rx_seq[phys_sender]-1)) {
/*  Case b).  Discard...  */

DEBUG_L(printg(
"rx_gamma():  par_pid %ld, port %ld:  discarding duplicated frame #%ld, rx_status %8.8x\n",
par_pid,port,frame_no,rx_status()
));

  goto end_rx;
 }
} else
 if (frame_no == (u8)(rx_seq_bcast[phys_sender]-1)) {
/*  Case b).  Discard...  */

DEBUG(printg(
"rx_gamma():  par_pid %ld, port %ld:  discarding duplicated bcast frame #%ld, rx_status %8.8x\n",
par_pid,port,frame_no,rx_status()
));

  goto end_rx;
 }

/*  Case a).  */

DEBUG_L(
if (!rec_is_bcast(h)) {
 printg("rx_gamma(): frame #%ld, expected #%d, phys sender %ld\n",
 frame_no,rx_seq[phys_sender],phys_sender);
} else {
 printg("rx_gamma(): bcast frame #%ld, expected #%d, phys sender %ld\n",
 frame_no,rx_seq_bcast[phys_sender],phys_sender);
}
);

if (!rec_is_bcast(h)) {
 if (((frame_no-rx_seq[phys_sender]+256)%256) > GAMMA_TX_WIN_SIZE) {

DEBUG_L(printg("Out of order frame?  Discard...\n"));

  goto end_rx;
 }
} else {
 if (((frame_no-rx_seq_bcast[phys_sender]+256)%256) > GAMMA_TX_WIN_SIZE) {

DEBUG_L(printg("Out of order bcast frame?  Discard...\n"));

  goto end_rx;
 }
}

/*
 *  The interval between consecutive checks for correct pkt numbering
 *  is restored to the base value (it was growing as long as no pkt losses were
 *  observed).
 */
max_polls = base_max_polls;

/*  A missing pkt could indicate a congestion.  Ask everybody to slow down.  */
gamma_kernel_send(BROADCAST,DONT_MIND,CONGESTION,NULL,0,1);
/*
 *  Send a NACK to the origin of this pkt, to enforce retransmission of pkts
 *  starting from actual count (go-back-N).
 */
if (!rec_is_bcast(h)) {
 err_msg[0] = rx_seq[phys_sender];
 err_msg[1] = par_pid;
 err_msg[2] = port;
 err_msg[3] = procnum;
 gamma_kernel_send(phys_sender,DONT_MIND,NACK,err_msg,4*sizeof(u32),1);
} else {
 err_msg[0] = rx_seq_bcast[phys_sender];
 err_msg[1] = par_pid;
 err_msg[2] = port;
 err_msg[3] = procnum;
 gamma_kernel_send(phys_sender,DONT_MIND,NACKB,err_msg,4*sizeof(u32),1);
}
/*  We refuse other pkts from phys_sender until we get the missing ones.  */
rx_discard[phys_sender] = 1;

goto end_rx;
#endif

not_assigned:

/*
 *  Virtual GAMMA not activated or application not yet registered or port not
 *  in use or no job running.
 *  Before discarding the pkt, check for a request for a flow ctl
 *  credit renewal...
 */
if (rec_header_creq(h)) {

DEBUG_F(printg("credit request from node %ld, frame_no %ld (par_pid unassigned)\n",
phys_sender,frame_no))

 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}

DEBUG(
printg("Par_pid or port not assigned, or no job running. Discarding frame (rx_status %8.8x)...\n",
rx_status())
);

goto end_rx;

pkt_is_not_gamma:

if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL) {
/*
 *  GAMMA kernel msg.
 *  In this case ``procnum'' is the kind of kernel msg.
 */
#ifdef SEQCHECK
/****************************************************************************/
 if (procnum == FILL_CREDIT || procnum == CRED_REQ) {
/*  Are we refusing regular/flowctl GAMMA pkts from phys_sender?  */
  if (rx_discard[phys_sender])
   goto end_rx;
/*  Check for correct frame sequence numbering.  */
  if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
   if (frame_no != rx_seq_bcast[phys_sender])
    goto err_frame_no;
   rx_seq_bcast[phys_sender]++;
  } else {
   if (frame_no != rx_seq[phys_sender])
    goto err_frame_no;
   rx_seq[phys_sender]++;
  }
 }
/****************************************************************************/
#endif

DEBUG(printg(
"rx_gamma(): kernel! complete %d, parpid %ld, type %ld, port %ld, cnk_len %ld, phys_sender %ld\n",
rec_header_msg_complete(h) != 0, par_pid, procnum, port, cnk_len, phys_sender
));

DEBUG_F(
if (procnum == FILL_CREDIT)
 printg("FILL_CREDIT from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
if (procnum == CRED_REQ)
 printg("CRED_REQ from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
)

 gamma_kernel_rec((u8 *)h+HEADER_SIZE,par_pid,cnk_len,procnum,phys_sender);

#if 0
 if (procnum == CRED_REQ || procnum == FILL_CREDIT || procnum == RETX_START
     || procnum == RETX_END || procnum == VM_WISH)
#else
 if (procnum == CRED_REQ || procnum == RETX_START || procnum == RETX_END
     || procnum == VM_WISH)
#endif
  goto end_rx;
/*
 *  For the other ETH_GAMMA_KERNEL pkts, we quit rx_gamma() to avoid a possible
 *  monopolization by this thread.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
 if (!bcast_sequence)
  discard_pkt();
 else {
  rec_header_set_procnum(h,procnum+1);
  forward_rx_ring_head();
 }
#else
 discard_pkt();
#endif
 goto exit;
}
#ifdef SEQCHECK
if (rec_header_frmtype(h) == ETH_GAMMA_RETX) {
/*
 *  Regular GAMMA data pkt, simply retransmitted.
 *  Can only arrive after a ETH_GAMMA_KERNEL pkt marked as "START_RETX"
 *  (start of retransmission sequence) has arrived from node "phys_sender",
 *  so we should find rx_discard[phys_sender] == 2 (accept retrasmission
 *  sequence from node "phys_sender").
 */
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg(
"rx_gamma():  refusing ETH_GAMMA_RETX pkt, frame_no %ld, sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
 }
 if (rx_discard[phys_sender] == 2) {
/*  Treat this pkt as it were a regular one  */

/*  GAMMA pkts not allowed if parallel job not started  */
  if (!gamma_susp_irq_count) goto end_rx;

DEBUG_F(
printg(
"rx_gamma(), pid %u: RETX! frame %d, %8.8x %8.8x %8.8x %8.8x %8.8x %8.8x\n",
current->pid, read_send_header_frameno(h),
h[0].ll, h[1].ll, h[2].ll, h[3].ll, h[4].ll, h[5].ll);
printg(
"complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
rec_header_msg_complete(h) != 0, par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h))
);

  goto accept_pkt;
 }
/*
 *  This retransmitted pkt is unexpected.  Maybe because
 *  it is a retransmitted bcast (in this case we must discard it);
 *  or, a RETX_START pkt has been lost
 *  (and we should rather ask the sender to restart its retransmission in
 *  this case; however, the protocol will eventually recover even if we
 *  discard this pkt).
 */

DEBUG_F(printg(
"rx_gamma(): unexpected ETH_GAMMA_RETX pkt, frame_no %ld, phys_sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
}
if (rec_header_frmtype(h) == ETH_GAMMA_KERNEL_RETX) {
/*
 *  GAMMA kernel msg, retransmitted.
 *  In this case ``procnum'' is the kind of kernel msg.
 *  Possible pkt kinds are:  FILL_CREDIT,CRED_REQ.
 */
/*  Are we refusing regular GAMMA pkts from phys_sender?  */
 if (rx_discard[phys_sender] == 1) {

DEBUG_F(printg(
"rx_gamma():  refusing ETH_GAMMA_KERNEL_RETX pkt, frame_no %ld, sender %ld\n",
frame_no,phys_sender
));

  goto end_rx;
 }
 if (rx_discard[phys_sender] == 2) {
/*  Treat this pkt as it were a regular one  */
/*  Check for correct frame sequence numbering.  */
  if (rec_is_bcast(h)) {
/*  Bcast pkts have distinct numbering.  */
   if (frame_no != rx_seq_bcast[phys_sender])
    goto err_frame_no;
   rx_seq_bcast[phys_sender]++;
  } else {
   if (frame_no != rx_seq[phys_sender])
    goto err_frame_no;
   rx_seq[phys_sender]++;
  }

DEBUG_F(printg(
"rx_gamma(): retx kernel! complete %d, parpid %ld, type %ld, port %ld, cnk_len %ld, phys_sender %ld\n",
rec_header_msg_complete(h) != 0, par_pid, procnum, port, cnk_len, phys_sender
));

DEBUG_F(
if (procnum == FILL_CREDIT)
 printg("FILL_CREDIT (retx) from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
if (procnum == CRED_REQ)
 printg("CRED_REQ (retx) from node %ld, frame_no %ld\n",
  phys_sender,frame_no);
)

  gamma_kernel_rec((u8 *)h+HEADER_SIZE,par_pid,cnk_len,procnum,phys_sender);

#if 0
  if (procnum == CRED_REQ || procnum == FILL_CREDIT || procnum == RETX_START
      || procnum == RETX_END)
#else
  if (procnum == CRED_REQ || procnum == RETX_START || procnum == RETX_END)
#endif
   goto end_rx;
/*
 *  For the other ETH_GAMMA_KERNEL pkts, we quit rx_gamma() to avoid a possible
 *  monopolization of this thread.
 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
  if (!bcast_sequence)
   discard_pkt();
  else {
   rec_header_set_procnum(h,procnum+1);
   forward_rx_ring_head();
  }
#else
  discard_pkt();
#endif
  goto exit;
 }
/*
 *  This retransmitted pkt is unexpected.  Maybe because
 *  it is a retransmitted bcast (in this case we must discard it);
 *  or, a RETX_START pkt has been lost
 *  (and we should rather ask the sender to restart its retransmission in
 *  this case; however, the protocol will eventually recover even if we
 *  discard this pkt).
 */

DEBUG_F(printg(
"rx_gamma(): unexpected ETH_GAMMA_KERNEL_RETX pkt, frame_no %ld, phys_sender %ld\n",
frame_no,phys_sender
));

 goto end_rx;
}
#endif
/*  Non-GAMMA frame.  "h" contains its first HEADER_SIZE bytes.  */

DEBUG_E(printg("rx_gamma(): non-GAMMA frame.  Discarding...\n"));

discard_pkt();
goto end_rx_with_updated_ptrs;

#ifdef MORE_INSTANCES_ON_SAME_NODE
found_bcast_pkt:

/*
 *  Found a bcast pkt.  We suppose it is the first of a sequence.
 *  Such pkts will not be removed from the Rx ring until all receivers
 *  have copied them into user space.  Instead, remove all the other
 *  pkts received so far (by forwarding the Rx ring head to the current pos).
 *  Also, pre-compute the next local receiver, to yield the CPU to it.
 */
/*  Is there any receiver next?  */
if (
 (procnum < MAX_NUM_LOCAL_INST-1) &&
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
) {
/*  Yes, there is!  */

DEBUG(printg("rx_gamma():  start of bcast seq\n"));

 bcast_sequence = 1;
 bcast_phys_sender = phys_sender;
/*
 *  Take note of the current Rx ring head and frame numbering:
 *  next local receiver will have to re-read pkts from there on.
 */
 read_rx_ring_head(head);
 bcast_frame_no = frame_no;

/*  Prepare to yield CPU to next receiver in the bcast group.  */
 gamma_must_yield.task = par_proc_table[par_pid]->inst_table[procnum+1].task;
 gamma_must_yield.pid = par_proc_table[par_pid]->inst_table[procnum+1].pid;
 gamma_time(gamma_must_yield.time);
#ifdef MORE_NICS_SAMENODE_ROUTING
 /* with multiple nics, the process we yield to must check the correct NIC
  * before any other to prevent possible race conditions. */
 gamma_must_yield.dev_idx = dev_idx;
#endif
}
/*  If no next local receiver, treat bcast as it were unicast.  */
goto found_bcast_pkt_back;
#endif

cannot_deliver_yet:

{
 task_t *p;
/*  Is the receiver process alive?  */
 if (pe->task->pid != pe->pid) goto dead_process;
 p = find_task_by_pid(pe->pid);
 if (p == NULL) goto dead_process;
 if (receiver->state & (EXIT_ZOMBIE | EXIT_DEAD | TASK_DEAD))
  goto dead_process;

DEBUG(printg(
"rx_gamma(), pid %u (%p):  delaying recv of pkt #%lu, receiver is %u (%p), state is %ld\n",
current->pid,current->mm,frame_no,receiver->pid,receiver->mm,receiver->state
));

#ifdef SEQCHECK
/****************************************************************************/
/*  Rollback the frame sequence numbering.  */
 if (rec_is_bcast(h))
/*  Bcast pkts have distinct numbering.  */
  rx_seq_bcast[phys_sender]--;
 else
  rx_seq[phys_sender]--;
/****************************************************************************/
#endif
/*  Quit, leave pkt in queue. Yield the CPU.  */
 gamma_must_yield.task = receiver;
 gamma_must_yield.pid = receiver->pid;
 gamma_time(gamma_must_yield.time);
#ifdef MORE_NICS_SAMENODE_ROUTING
 /* with multiple nics, the process we yield to must check the correct NIC
  * before any other to prevent possible race conditions. */
 gamma_must_yield.dev_idx = dev_idx;
#endif
 goto yield;

}

dead_process:

/*  receiver is dead.  discard packet.  */

DEBUG(printg("rx_gamma(): recv proc does not exist.  Discarding packet...\n"));

goto end_rx;

credit_request:

DEBUG_F(printg(
"credit request from node %ld, frame_no %ld (is_bcast %d)\n",
phys_sender,frame_no,rec_is_bcast(h)
))

/*  This packet carries a credit request. Refill credit to sender.  */
#ifdef MORE_INSTANCES_ON_SAME_NODE
/*
 *  Exception: if this is a bcast, refill credit
 *  only if this receiver is the last one of the bcast.
 */
if (!(rec_is_bcast(h))) {
#endif
/*  Refill credit and clean the credit request.  */
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
 goto credit_request_back;
#ifdef MORE_INSTANCES_ON_SAME_NODE
}
/*
 *  Bcast packet.
 *  Refill (and clean the request) iff no other receivers (excluding this one).
 */
if (
 (procnum >= MAX_NUM_LOCAL_INST-1) ||
 (par_proc_table[par_pid]->inst_table[procnum+1].whoami < 0)
) {
 gamma_kernel_send(phys_sender,DONT_MIND,FILL_CREDIT,NULL,0,1);
 rec_header_clean_creq(h);
}
goto credit_request_back;
#endif

bogus_inv:

DEBUG_E(printg(
"rx_gamma(): bogus invocation (rx_status %8.8x), pid %u\n",
rx_status(),current->pid
));

return res;

rec_reentering:

if (gamma_run_handler.task != current)
/*
 *  A receiver has temporarily quit to run the rec handler, but will
 *  reenter soon.
 *  Prevent other processes to enter now.
 */
 goto yield;

/*  The receiver is reentering after running a rec handler (1sh piece).  */

dev = gamma_run_handler.dev;
rx_common_init(dev);

DEBUG_E(printg("rx_gamma(), pid %u: reentering...\n",current->pid));

gamma_run_handler.task = NULL;
rx_read_pkt_info();
res = -GAMMA_RX_INCOMPLETE;

DEBUG_E(printg(
"rx_gamma(): rx_ring_head %d, rx_ring_tail %d, rx_status %4.4x\n",
rx_ring_head,rx_ring_tail,rx_status()
));

rx_header(h);

cnk_len = rec_header_cnk_len(h);
frame_no = rec_header_frame_no(h);
phys_sender = rec_header_phys_sender(h);
rec_header_parpid_port(h,par_pid,port,procnum);

DEBUG(printg(
"rx_gamma(), pid %u:  reentering: complete %d, par_pid %ld, procnum %ld, port %ld, cnk_len %ld, frame_no %ld, phys_sender %ld, creq %d\n",
current->pid,
rec_header_msg_complete(h) != 0, par_pid, procnum, port, cnk_len, frame_no,
phys_sender, rec_header_creq(h)
));

/*  Check whether virtual GAMMA #par_pid exists and is assigned.  */
if (
 (par_proc_table[par_pid] == NULL)
 || (par_proc_table[par_pid]->inst_table[procnum].pid == -1)
)
 goto not_assigned;

pe = &par_proc_table[par_pid]->inst_table[procnum].port_table[port];
receiver = pe->task;

#ifdef MORE_INSTANCES_ON_SAME_NODE
if (rec_is_bcast(h)) {
 if (
  (procnum < MAX_NUM_LOCAL_INST-1) &&
  (par_proc_table[par_pid]->inst_table[procnum+1].whoami >= 0)
 ) {
  bcast_sequence = 1;
  read_rx_ring_head(head);
  bcast_frame_no = frame_no;
  gamma_must_yield.task = par_proc_table[par_pid]->inst_table[procnum+1].task;
  gamma_must_yield.pid = par_proc_table[par_pid]->inst_table[procnum+1].pid;
  gamma_time(gamma_must_yield.time);
#ifdef MORE_NICS_SAMENODE_ROUTING
  /* with multiple nics, the process we yield to must check the correct NIC
   * before any other to prevent possible race conditions. */
  gamma_must_yield.dev_idx = dev_idx;
#endif
 }
}
#endif

first_piece_size = pe->boundary_1p;
cnk_len -= first_piece_size;
goto get_remaining_payload;

yield:

return -GAMMA_ERR_MUSTYIELD;

}


asmlinkage long sys_gamma_poll
(unsigned long my_par_pid, unsigned long my_procnum)
{

	int res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t *shared_ring = NULL;
	int res2;
#endif
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned int gamma_dev_idx;
	int i;
	poll_common_decl; 
#else
	poll_common_decl_and_init;
#endif
	task_t *p;
	yield_t *y;
	time_586 a, b, cur_time;


	res  = -GAMMA_RX_INCOMPLETE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (likely(my_par_pid))
		shared_ring =
			&par_proc_table[my_par_pid]->inst_table[my_procnum]
			.shared_ring;
//if (!my_par_pid) printg("ciao: %p %p\n",par_proc_table[my_par_pid],shared_ring);
	res2 = -GAMMA_RX_INCOMPLETE;
#endif
	gamma_time(a);
	
	for (;;) {
/*  Check out the NIC rx ring (inter-node msgs)  */

		if (gamma_must_yield.task != NULL) goto chk_yield;

		if (!spin_trylock(&gamma_rx_lock)) goto skip_inter;

#ifdef MORE_NICS_SAMENODE_ROUTING
		gamma_dev_idx = 0;
#endif

ok_enter:

#ifdef MORE_NICS_SAMENODE_ROUTING
/* scan all NICS sequentially and return as soon as a packet is found  */
		for (i=gamma_dev_idx; i < num_gamma_devices; i++) {
			rx_read_pkt_info_light_arg(i);
			if (!(rx_err_pktnotavail())) {
				res = rx_gamma(i);

#else
		{
			rx_read_pkt_info_light();
			if (!(rx_err_pktnotavail())) {
				res = rx_gamma(0);
#endif
				clear_rx_irqs();
/*
 * res can be:
 * >= 0, -GAMMA_RX_INCOMPLETE, -GAMMA_ERR_MUSTYIELD, -GAMMA_RX_NODATA
 */
#ifdef SEQCHECK
				if (res >= 0 || res == -GAMMA_RX_INCOMPLETE)
					how_many_polls = 0;
#endif
				if (likely(res != -GAMMA_RX_INCOMPLETE)) {
					spin_unlock(&gamma_rx_lock);
					return res;
				}
#ifdef MORE_NICS_SAMENODE_ROUTING
				break;
#endif
			}
		}

		spin_unlock(&gamma_rx_lock);

skip_inter:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  Check for intra-node messages  */

		if (unlikely(!my_par_pid)) goto skip_intra;

		read_lock(&gamma_rxintranode_lock);
		if (shared_ring->head != atomic_read(&shared_ring->tail)) {
			res2 = rx_gamma_intranode(my_par_pid,my_procnum);
/*  res2 can be: >=0, -GAMMA_RX_INCOMPLETE  */
#ifdef SEQCHECK
			how_many_polls = 0;
#endif
			if (likely(res2 != -GAMMA_RX_INCOMPLETE)) {
				read_unlock(&gamma_rxintranode_lock);
				return res2;
			}
		}
		read_unlock(&gamma_rxintranode_lock);

skip_intra:

#endif

		gamma_time(b);
		if (gamma_time_diff_ticks(b,a) >= REPEAT_POLL*CLOCK) break;

		continue;

chk_yield:

		y = &gamma_must_yield;

		if (y->task == current) {
			spin_lock(&gamma_rx_lock);
#ifdef MORE_NICS_SAMENODE_ROUTING
/*
 * make sure we get the message that requested the cpu to be yielded 
 * to this process before any other
 */
			gamma_dev_idx = y->dev_idx;
#endif
			y->task = NULL;
			goto ok_enter;
		}

		gamma_time(cur_time);
		if (gamma_time_diff_ticks(cur_time,y->time) < SIGURG_TIMEOUT)
			return -GAMMA_ERR_MUSTYIELD;

/*
 * Too long time without entering NIC Rx queue.  Stimulate the
 * other process on same node to take its own pkts away.
 */
		if (!spin_trylock(&gamma_rx_lock))
			return -GAMMA_ERR_MUSTYIELD;
		if (y->task == NULL) {
/* The other process on same node had taken away its own pkts meanwhile. */
			spin_unlock(&gamma_rx_lock);
			continue;
		}
		if (y->task->pid != y->pid) goto dead;
		p = find_task_by_pid(y->pid);
		if (p == NULL) goto dead;
		if (y->task->state & (EXIT_ZOMBIE | EXIT_DEAD | TASK_DEAD))
			goto dead;

DEBUG_L(printg("sys_gamma_poll(): too long waiting to enter rx().\n"));
DEBUG(printg("A pkt sits in the NIC Rx queue and the dest process does not read it.\n"));

		if (!(y->task->state &
			(TASK_UNINTERRUPTIBLE | TASK_STOPPED | TASK_TRACED))) {

DEBUG(printg("SIGURG to destination process (pid %d)\n",y->task->pid));

			force_sig(SIGURG,(task_t *)y->task);
		}
		gamma_time(y->time);
		spin_unlock(&gamma_rx_lock);
		continue;  /* will yield at next round */

dead:

/*  Receiver dead.  Go ahead, will remove dangling pkt later if any.  */
		y->task = NULL;
		spin_unlock(&gamma_rx_lock);
		continue;

	}

#ifdef SEQCHECK
	if (++how_many_polls >= max_polls) {
		how_many_polls = 0;

DEBUG_L(printg(
"sys_gamma_poll():  no pkts for a long time.  Testing frame seq #s...\n"
));

{
DEBUG_F(
 int i;
 for (i=0;i<total_num_nodes;i++)
	printg("%d ",atomic_read(&local_credit[i]));
 printg("\n");
 for (i=0;i<total_num_nodes;i++)
	 printg("%d ",rx_seq[i]);
 printg("\n");
)
}

		spin_lock(&gamma_tx_lock);

/*  Ask everybody to spread their frame numbering tables.  */
		gamma_kernel_send(BROADCAST,DONT_MIND,TEST,NULL,0,0);
#if 0
/*  Transmit the local frame numbering table (both unicast and bcast).  */
 for (
  i = (phys_whoami + 1) % total_num_nodes;
  i != phys_whoami;
  i = (i + 1) % total_num_nodes
 ) {
  unsigned long frame_no;

  frame_no = tx_tmp_ring[i].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACK,&frame_no,sizeof(frame_no),0);
  frame_no = tx_tmp_ring[phys_whoami].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACKB,&frame_no,sizeof(frame_no),0);
 }
#endif

		spin_unlock(&gamma_tx_lock);
	}
#endif

	return -GAMMA_RX_INCOMPLETE;

}


asmlinkage long sys_gamma_poll_intranode
(unsigned long my_par_pid, unsigned long my_procnum)
#ifdef MORE_INSTANCES_ON_SAME_NODE
{

	shared_ring_t *shared_ring;
	int res = -GAMMA_RX_INCOMPLETE;


	if (unlikely(!my_par_pid)) return res;

/*  Check for intra-node messages  */

	read_lock(&gamma_rxintranode_lock);
	shared_ring = &par_proc_table[my_par_pid]->inst_table[my_procnum]
			.shared_ring;
	if (shared_ring->head != atomic_read(&shared_ring->tail))
		res = rx_gamma_intranode(my_par_pid,my_procnum);
	read_unlock(&gamma_rxintranode_lock);
	return res;

}
#else
{
	return 0;
}
#endif


asmlinkage long sys_gamma_poll_nowait
(unsigned long my_par_pid, unsigned long my_procnum)
{

	int res;
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t *shared_ring = NULL;
	int res2;
#endif
#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned int gamma_dev_idx;
	int i;
	poll_common_decl; 
#else
	poll_common_decl_and_init;
#endif
	task_t *p;
	yield_t *y;
	time_586 cur_time;


	res  = -GAMMA_RX_INCOMPLETE;

#ifdef MORE_INSTANCES_ON_SAME_NODE
	if (likely(my_par_pid))
		shared_ring =
			&par_proc_table[my_par_pid]->inst_table[my_procnum]
			.shared_ring;
	res2 = -GAMMA_RX_INCOMPLETE;
#endif

	for (;;) {

 /*  Check out the NIC rx ring (inter-node msgs)  */

		if (gamma_must_yield.task != NULL) goto chk_yield;

		if (!spin_trylock(&gamma_rx_lock)) goto skip_inter;

#ifdef MORE_NICS_SAMENODE_ROUTING
		gamma_dev_idx = 0;
#endif

ok_enter:

#ifdef MORE_NICS_SAMENODE_ROUTING
		for (i=gamma_dev_idx; i < num_gamma_devices; i++) {
			rx_read_pkt_info_light_arg(i);
			if (!(rx_err_pktnotavail())) {
				res = rx_gamma(i);

#else
		{
			rx_read_pkt_info_light();
			if (!rx_err_pktnotavail()) {
				res = rx_gamma(0);
#endif
				clear_rx_irqs();
/*
 * res can be:
 * >= 0, -GAMMA_RX_INCOMPLETE, -GAMMA_ERR_MUSTYIELD, -GAMMA_RX_NODATA
 */
#ifdef SEQCHECK
				if (res >= 0 || res == -GAMMA_RX_INCOMPLETE)
					how_many_polls = 0;
#endif
				if (likely(res != -GAMMA_RX_INCOMPLETE)) {
					spin_unlock(&gamma_rx_lock);
					return res;
				}
#ifdef MORE_NICS_SAMENODE_ROUTING
				break;
#endif
			}
		}

		spin_unlock(&gamma_rx_lock);

skip_inter:

#ifdef MORE_INSTANCES_ON_SAME_NODE
/*  Check for intra-node messages  */

		if (unlikely(!my_par_pid)) goto skip_intra;

		read_lock(&gamma_rxintranode_lock);
		if (shared_ring->head != atomic_read(&shared_ring->tail)) {
			res2 = rx_gamma_intranode(my_par_pid,my_procnum);
/*  res2 can be: >=0, -GAMMA_RX_INCOMPLETE  */
#ifdef SEQCHECK
			how_many_polls = 0;
#endif
			if (likely(res2 != -GAMMA_RX_INCOMPLETE)) {
				read_unlock(&gamma_rxintranode_lock);
				return res2;
			}
		}
		read_unlock(&gamma_rxintranode_lock);

skip_intra:

#endif

#ifdef SEQCHECK
		if (++how_many_polls >= (max_polls*2000)) {
			/*
			 *  "2000":  Adjust waiting time to take into account
			 *  that this routine does not spin over Rx (it is a
			 *  "one time only" check) and so returns quickly.
			 */
			how_many_polls = 0;

DEBUG_L(printg(
"sys_gamma_poll_nowait():  no pkts for a long time.  Testing frame seq #s...\n"
));

			spin_lock(&gamma_tx_lock);

/*  Ask everybody to spread their frame numbering tables.  */
			gamma_kernel_send(BROADCAST,DONT_MIND,TEST,NULL,0,0);
#if 0
/*  Transmit the local frame numbering table (both unicast and bcast).  */
 for (
  i = (phys_whoami + 1) % total_num_nodes;
  i != phys_whoami;
  i = (i + 1) % total_num_nodes
 ) {
  unsigned long frame_no;

  frame_no = tx_tmp_ring[i].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACK,&frame_no,sizeof(frame_no),0);
  frame_no = tx_tmp_ring[phys_whoami].tx_seq;
  gamma_kernel_send(i,DONT_MIND,ACKB,&frame_no,sizeof(frame_no),0);
 }
#endif

			spin_unlock(&gamma_tx_lock);
		}
#endif

		return -GAMMA_RX_INCOMPLETE;

chk_yield:

		y = &gamma_must_yield;

		if (y->task == current) {
			spin_lock(&gamma_rx_lock);
#ifdef MORE_NICS_SAMENODE_ROUTING
/*
 * make sure we get the message that requested the cpu to be yielded 
 * to this process before any other
 */
			gamma_dev_idx = y->dev_idx;
#endif
			y->task = NULL;
			goto ok_enter;
		}

		gamma_time(cur_time);
		if (gamma_time_diff_ticks(cur_time,y->time) < SIGURG_TIMEOUT)
			return -GAMMA_ERR_MUSTYIELD;

/*
 * Too long time without entering NIC Rx queue.  Stimulate the
 * other process on same node to take its own pkts away.
 */
		if (!spin_trylock(&gamma_rx_lock))
			return -GAMMA_ERR_MUSTYIELD;
		if (y->task == NULL) {
/* The other process on same node had taken away its own pkts meanwhile. */
			spin_unlock(&gamma_rx_lock);
			continue;
		}
		if (y->task->pid != y->pid) goto dead;
		p = find_task_by_pid(y->pid);
		if (p == NULL) goto dead;
		if (y->task->state & (EXIT_ZOMBIE | EXIT_DEAD | TASK_DEAD))
			goto dead;

DEBUG_L(printg("sys_gamma_poll_nowait(): too long waiting to enter rx().\n"));
DEBUG(printg("A pkt sits in the NIC Rx queue and the dest process does not read it.\n"));

		if (!(y->task->state &
			(TASK_UNINTERRUPTIBLE | TASK_STOPPED | TASK_TRACED))) {

DEBUG(printg("SIGURG to destination process (pid %d)\n",y->task->pid));

			force_sig(SIGURG,(task_t *)y->task);
		}
		gamma_time(y->time);
		spin_unlock(&gamma_rx_lock);
		continue;  /* will yield at next round */

dead:

/*  Receiver dead.  Go ahead, will remove dangling pkt later if any.  */
		y->task = NULL;
		spin_unlock(&gamma_rx_lock);
		continue;

	}

}


/************************************************************************
 *	Miscellaneous routines						*
 ************************************************************************/


/*  Disable IRQ-driven message delivery  */
asmlinkage long sys_gamma_suspend_irq
(void)
{
	spin_lock(&gamma_susp_irq_count_lock);
	if (!(gamma_susp_irq_count++))
		rx_suspend_irq();
	spin_unlock(&gamma_susp_irq_count_lock);
	return 0;
}


/*  Enable IRQ-driven message delivery  */
asmlinkage long sys_gamma_resume_irq
(void)
{
	spin_lock(&gamma_susp_irq_count_lock);
	if (gamma_susp_irq_count)
		if (!(--gamma_susp_irq_count))
			rx_resume_irq();
	spin_unlock(&gamma_susp_irq_count_lock);
	return 0;
}


/*
 *  Print an integer value to "dmesg".
 */
asmlinkage long sys_gamma_printk
(long val)
{
	printg("sys_gamma_printk(): process %s (pid %d): %ld (0x%016lx)\n",
		current->comm,current->pid,val,val
	);
	return 0;
}


/*
 *  Set, reset, get the initial credit size.
 */
asmlinkage long sys_gamma_setcredit
(unsigned long size)
{
	int i;

	credit_size = size;
	for (i=0;i<total_num_nodes;i++) {
		initial_credit[i] = size;
		credit_ceil[i] = size;
		credit_growth_cnt[i] = 0;
		atomic_set(&local_credit[i],initial_credit[i]);
	}
	return 0;
}

asmlinkage long sys_gamma_resetcredit
(void)
{
	int i;

	credit_size = MAX_CREDIT_SIZE;
	for (i=0;i<total_num_nodes;i++) {
		initial_credit[i] = MAX_CREDIT_SIZE;
		credit_ceil[i] = MAX_CREDIT_SIZE;
		credit_growth_cnt[i] = 0;
		atomic_set(&local_credit[i],initial_credit[i]);
	}
	return 0;
}

asmlinkage long sys_gamma_getcredit
(void)
{
	int i;

	printg("sys_gamma_getcredit(): ");
	for (i=0;i<total_num_nodes;i++) {
		printg("-->%d: %ld",i,initial_credit[i]);
		if (i<total_num_nodes-1)
			printg(", ");
	}
	printg("\n");
/*
 *  We only return the credit size towards node 0.  The complete picture has
 *  been printed by the statements above.
 */
	return initial_credit[0];
}


/*
 *  Set and reset the "max_polls" variables (threshold which triggers the
 *  collective check of frame numbering and credits.
 */
asmlinkage long sys_gamma_setmaxpolls
(unsigned long n)
{
	max_polls = n;
	base_max_polls = n;
	return 0;
}

asmlinkage long sys_gamma_resetmaxpolls
(void)
{
	max_polls = MAX_POLLS;
	base_max_polls = MAX_POLLS;
	return 0;
}

asmlinkage long sys_gamma_getmaxpolls
(void)
{
	return base_max_polls;
}


/************************************************************************
 *	Local routines							*
 ************************************************************************/


static inline int is_low_watermark
(unsigned long node, unsigned long credit, unsigned long curr_credit)
{
	int i;

	if (node != BROADCAST)
		return (curr_credit == low_watermark_pt2pt(node));
	for (i=0; i<total_num_nodes; i++)
		if (i != phys_whoami)
			if (
				(atomic_read(&local_credit[i])
				 - credit + curr_credit)
				== low_watermark_pt2pt(i)
			)
				return 1;
	return 0;
}


/*
 *  Other support routines.
 */
#include "sys_gamma_connect_port.c"
#include "sys_gamma_vm.c"
#include "sys_gamma_mem.c"
#include "sys_gamma_timers.c"
