/***************************************************************************
	Read COPYRIGHT file in this directory
****************************************************************************/


typedef union {
	u32 ll;
	u16 ss[2];
} l2s;

typedef struct {
	u_int32_t key;
	int32_t pid;  /* sign is important: -1 == not taken */
} wish_t;

typedef struct {
	unsigned short partner_instnum;
	unsigned short partner_port;
	unsigned char partner_par_pid;
	unsigned char partner_procnum;
	unsigned short phys_partner_node;
	unsigned char buffer_kind;
	unsigned char boundary_1p;
	unsigned short semaphore;
	void *in_buffer;

	unsigned int current_rx_msglen;
	unsigned int in_size;
	void *in_progress;
	unsigned int in_room;

	struct task_struct *task;    /* thread who uses this port for input */
	int pid;                     /* pid of above                        */
	l2s *header;                 /*
                                      *  Header for frames, plus room for a
                                      *  possible small payload (1st pkt only)
                                      */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	unsigned char partner_is_intra_node;
	unsigned char bcast_was_susp;
#endif
	short suspended_isend_handle;  /* Signed:  Default value is < 0       */

	void *current_tx_data1;
	unsigned int current_tx_len1;
	void *current_tx_data2;
	unsigned int current_tx_len2;

#ifdef MORE_NICS_SAMENODE_ROUTING
	unsigned int nic_idx;  /* the nic on which the remote partner is   */
#endif

} port_entry;

#ifdef MORE_INSTANCES_ON_SAME_NODE

/*
 *  Rings of (ptrs to) temp buffers for intra-node message exchange.
 *  Init and alloc by sys_gamma_register()
 */

#define IS_1ST_CNK 0x1
#define IS_LAST_CNK 0x2
#define IS_EMPTY 0x4
#define IN_USE 0x8

typedef struct {
	unsigned long port;
	unsigned long cnk_len;
	volatile unsigned long bitmap;
	void *buf;
} shared_ring_entry;

typedef struct {
	volatile unsigned int head;
	atomic_t dirty;
	atomic_t tail;
	atomic_t room;
	char dummy[8*sizeof(void *)-4*sizeof(int)]; /* padding */
	shared_ring_entry desc[SHARED_RING_SIZE];
} shared_ring_t;

#endif

typedef struct {
	port_entry port_table[MAX_PORT_NUMBER+1];

	/*  Is the process re-entering Rx after running handler?  */
	int reenter;

	/*
	 *  Pointers to user-space variable "gamma_msglen", holding
	 *  the len of the last arrived msg.
	 */
	void *msglen_ptr;

#ifdef USE_DAMNICK
	void *sem1_table;
	void *pre_sem_table;
#endif

	volatile int pid;
/*
 *  The following is also present in the "port_entry" data struct,
 *  to allow multiple threads: different threads can use different
 *  input ports.
 */
	struct task_struct *task;
	int whoami;
		/*
		 *  "whoami" wrt the virtual GAMMA corresponding to this entry
		 *  Not unsigned because -1 == "not involved in this GAMMA vm"
		 */
#ifdef MORE_INSTANCES_ON_SAME_NODE
	shared_ring_t shared_ring;
#endif
} instance_entry;

typedef struct {
	u_int32_t random_key;		/*  To distinguish between two       */
					/*  consecutive virtual GAMMAs       */
					/*  who happen to have same par_pid  */
	u_int32_t num_instances;
	unsigned short node_list[MAX_NUM_NODES*MAX_NUM_LOCAL_INST];
} vm_descriptor;

typedef struct {
	unsigned int num_locinst;
	unsigned int quitting;
		/*  process instances of this group that run on local node   */
	instance_entry inst_table[MAX_NUM_LOCAL_INST];
	vm_descriptor vm;
	int32_t wish_pid[MAX_NUM_NODES*MAX_NUM_LOCAL_INST];
} par_entry;
